#include <graph/graph.h>
#include <graph/components.h>
#include <graph/dc.h>
#include <util/input_generator.h>
#include <util/stopwatch.h>
#include <util/assert.h>
#include <cstdio>
#include <typeinfo>
#include <map>
#include <string>

using namespace Memory;
using namespace Allocator;
using namespace std;

namespace {

class Mark : public Graph::Vertex {
	bool enabled;
public:
	Mark() :
		enabled(false)
	{ }
	void setMark(bool enabled) { this->enabled = enabled; }
	bool isMarkSet() const { return enabled; }
};

size_t count(10);

template<class Graph, class Pool, class Function> Pointer<Core, size_t> singleTest(Graph &graph, const char *name, size_t size, Pool &allocator, Function function, Pointer<Core, size_t> expected = Pointer<Core, size_t>()) {
	Util::Stopwatch watch(name, graph.name, Graph::StorageType::name(), "void", size);
	typename Pool::State state(allocator);
	Pointer<Core, size_t> result;
	watch.start();
	for (size_t i(0); i < count; ++ i) {
		allocator.restore(state);
		result = function(graph, allocator);
	}
	watch.stop();
	if (expected) {
		typedef map<size_t, size_t> Map;
		Map map;
		for (size_t i(0); i < graph.n; ++ i) {
			Map::const_iterator it(map.find(expected[i]));
			if (it != map.end()) {
				ASSERT_EXPL(it->second == result[i], ("Fault with " + string(name)).c_str());
			} else {
				map[expected[i]] = result[i];
			}
		}
		allocator.restore(state);
	}
	return result;
}

typedef Graph::Graph<Core, Mark> G;

void test(Pool<Core> &allocator, size_t size, G::Generator generator, int mult = 0) {
	Core::allowThreads = false;
	typename Pool<Core>::State state(allocator);
	G graph(G::generate(allocator, generator, size, mult));
	Pointer<Core, size_t> expected(singleTest(graph, "componentsDfs", size, allocator, Graph::componentsDfs<G>));
	Core::allowThreads = false;
	singleTest(graph, "componentsBfs", size, allocator, Graph::componentsBfs<G>, expected);
	singleTest(graph, "componentsDfsBRT", size, allocator, Graph::componentsDfsBRT<G>, expected);
	singleTest(graph, "componentsBfsBRT", size, allocator, Graph::componentsBfsBRT<G>, expected);
	singleTest(graph, "componentsBfsSort", size, allocator, Graph::componentsBfsSort<G>, expected);
	singleTest(graph, "componentsDC", size, allocator, Graph::DivideConquer::components<G>, expected);
	Core::allowThreads = true;
	singleTest(graph, "componentsBfsSortThreaded", size, allocator, Graph::componentsBfsSort<G>, expected);
	singleTest(graph, "componentsDCThreaded", size, allocator, Graph::DivideConquer::components<G>, expected);
	allocator.restore(state);
}

}

int main() {
	Pool<Core> allocator;
	size_t size(TEST_SIZE);
	size /= 3;
	for (double s(1); s * s < size; s *= 1.5)
		test(allocator, s, G::Fixed);
	size /= 160;
	for (double s(1); s < size; s *= 1.5)
		test(allocator, s, G::Count, 10);
	size *= 5;
	for (double s(1); s < size; s *= 1.5) {
		test(allocator, s, G::Count, 1);
		test(allocator, s, G::Grid);
		test(allocator, s, G::Triangulate, 1);
	}
	return 0;
}
