#include <graph/graph.h>
#include <graph/classic.h>
#include <graph/treeish.h>
#include <graph/sort-bfs.h>
#include <util/input_generator.h>
#include <util/stopwatch.h>
#include <util/assert.h>
#include <typeinfo>
#include <cmath>

using namespace Memory;
using namespace Memory::Allocator;
using namespace Util;
using namespace std;

namespace {

class Mark : public Graph::Vertex {
	bool enabled;
public:
	Mark() :
		enabled(false)
	{ }
	void setMark(bool enabled) { this->enabled = enabled; }
	bool isMarkSet() const { return enabled; }
};

size_t count(10);

template<class Graph, class Pool, class Function> Pointer<Core, size_t> singleTest(Graph &graph, const char *name, size_t size, Pool &allocator, Function function, Pointer<Core, size_t> expected = Pointer<Core, size_t>()) {
	Stopwatch watch(name, graph.name, Graph::StorageType::name(), "void", size);
	typename Pool::State state(allocator);
	Pointer<Core, size_t> result;
	watch.start();
	for (size_t i(0); i < count; ++ i) {
		allocator.restore(state);
		result = function(graph, allocator, 0);
	}
	watch.stop();
	if (expected)
		for (size_t i(0); i < graph.n; ++ i) {
			ASSERT(result[i] == expected[i]);
			if (expected[i] == graph.n + 1)
				break;
		}
	return result;
}

typedef Graph::Graph<Core, Mark> G;

void test(Pool<Core> &allocator, size_t size, G::Generator generator, int mult = 0) {
	typedef Pool<Core>::State State;
	State state(allocator);
	G graph(G::generate(allocator, generator, size, mult));
	State state2(allocator);
	Pointer<Core, size_t> cdfs(singleTest(graph, "dfsClassic", size, allocator, Graph::Classic::dfs<G>));
	singleTest(graph, "dfsBRT", size, allocator, Graph::Treeish::dfs<G>, cdfs);
	allocator.restore(state2);
	Pointer<Core, size_t> cbfs(singleTest(graph, "bfsClassic", size, allocator, Graph::Classic::bfs<G>));
	singleTest(graph, "bfsBRT", size, allocator, Graph::Treeish::bfs<G>, cbfs);
	Pointer<Core, size_t> sbfs(singleTest(graph, "bfsSortThreaded", size, allocator, Graph::bfsSort<G>));
	Core::allowThreads = false;
	singleTest(graph, "bfsSort", size, allocator, Graph::bfsSort<G>, sbfs);
	Core::allowThreads = true;
	for (size_t i(0); i < size; ++ i)
		if (cbfs[i] == graph.n + 1) {
			size = i;
			break;
		}
	Algorithm::Sort::sort(cbfs, size);
	Algorithm::Sort::sort(sbfs, size);
	for (size_t i(0); i < size; ++ i)
		ASSERT(cbfs[i] == sbfs[i]);
	allocator.restore(state);
}

}

int main() {
	Pool<Core> allocator;
	size_t size(TEST_SIZE);
	size /= 3;
	for (double s(1); s * s < size; s *= 1.5)
		test(allocator, s, G::Fixed);
	size /= 160;
	for (double s(1); s < size; s *= 1.5)
		test(allocator, s, G::Count, 10);
	size *= 5;
	for (double s(1); s < size; s *= 1.5) {
		test(allocator, s, G::Count, 1);
		test(allocator, s, G::Grid);
		test(allocator, s, G::Triangulate, 1);
	}
	return 0;
}
