#include <graph/graph.h>
#include <graph/components.h>
#include <graph/dc.h>
#include <util/input_generator.h>
#include <util/stopwatch.h>
#include <util/assert.h>
#include <memory/cache.h>
#include <cstdio>
#include <typeinfo>
#include <map>
#include <string>

using namespace Memory;
using namespace Allocator;
using namespace std;

namespace {

typedef Disk<DiskCache<8192, 4096> > DCache;

class Mark : public Graph::Vertex {
	bool enabled;
public:
	Mark() :
		enabled(false)
	{ }
	void setMark(bool enabled) { this->enabled = enabled; }
	bool isMarkSet() const { return enabled; }
};

template<class Graph, class Pool, class Function> Pointer<DCache, size_t> singleTest(Graph &graph, const char *name, size_t size, Pool &allocator, Function function, Pointer<DCache, size_t> expected = Pointer<DCache, size_t>()) {
	Util::Stopwatch watch(name, graph.name, Graph::StorageType::name(), "void", size);
	watch.start();
	typename Pool::State state(allocator);
	Pointer<DCache, size_t> result(function(graph, allocator));
	watch.stop();
	if (expected) {
		typedef map<size_t, size_t> Map;
		Map map;
		for (size_t i(0); i < graph.n; ++ i) {
			Map::const_iterator it(map.find(expected[i]));
			if (it != map.end()) {
				ASSERT_EXPL(it->second == result[i], ("Fault with " + string(name)).c_str());
			} else {
				map[expected[i]] = result[i];
			}
		}
		allocator.restore(state);
	}
	return result;
}

typedef Graph::Graph<DCache, Mark> G;

void test(Pool<DCache> &allocator, size_t size, G::Generator generator, int mult = 0) {
	Pool<DCache>::State state(allocator);
	DCache::resize(524288); // Allow more memory for faster graph generation
	G graph(G::generate(allocator, generator, size, mult));
	DCache::restoreSize(); // But the tests should not have it
	if (size < 11057332) {
		singleTest(graph, "componentsBfs", size, allocator, Graph::componentsBfs<G>);
	}
	singleTest(graph, "componentsBfsBRT", size, allocator, Graph::componentsBfsBRT<G>);
	if (size < 11057332) {
		singleTest(graph, "componentsBfsSort", size, allocator, Graph::componentsBfsSort<G>);
	}
	singleTest(graph, "componentsDC", size, allocator, Graph::DivideConquer::components<G>);
	allocator.restore(state);
}

}

int main() {
	Pool<DCache> allocator;
	size_t size(TEST_SIZE);
	for (double s(1); s < size; s *= 1.5) {
		test(allocator, s, G::Count, 3);
	}
	return 0;
}
