#include <matrix/classic.h>
#include <matrix/divide.h>
#include <matrix/z.h>
#include <matrix/strassen.h>
#include <util/input_generator.h>
#include <util/stopwatch.h>
#include <cstdio>
#include <typeinfo>
#include <vector>
#include <cmath>

using namespace Memory;
using namespace Memory::Allocator;
using namespace Matrix;
using namespace Util;
using namespace std;

namespace {

template<class Matrix> void print(Matrix m) {
	printf("%zu %zu:\n", m.height(), m.width());
	for (size_t i(0); i < m.height(); ++ i) {
		for (size_t j(0); j < m.width(); ++ j)
			printf("%u ", (unsigned) m[i][j]);
		printf("\n");
	}
}

template<class Matrix, class Pool, class Function> void singleTest(Matrix l, Matrix r, Matrix correct, const char *name, size_t size, Pool *pool, Function function) {
	typename Pool::State state(*pool);
	Stopwatch watch(name, "matrix", Matrix::StorageType::name(), typeid(typename Matrix::Type).name(), size);
	watch.start();
	Matrix result(size, size, pool);
	function(result, l, r, 0, pool);
	watch.stop();
	for (size_t i(0); i < size; ++ i)
		for (size_t j(0); j < size; ++ j)
			ASSERT(result[i][j] == correct[i][j]);
	pool->restore(state);
}

template<class Matrix> void test(Pool<typename Matrix::StorageType> *allocator, size_t size) {
	typename Pool<typename Matrix::StorageType>::State state(*allocator);
	size_t width(size);
	size_t height(size);
	size_t length(size);
	Matrix m1(height, length, allocator);
	Generator::random(m1);
	Matrix m2(length, width, allocator);
	Generator::random(m2);
	Stopwatch watchC("classic", "matrix", Matrix::StorageType::name(), typeid(typename Matrix::Type).name(), size);
	watchC.start();
	Matrix resultC(height, width, allocator);
	classic(resultC, m1, m2, 0, allocator);
	watchC.stop();
	singleTest(m1, m2, resultC, "inverse", size, allocator, inverse<Matrix>);
	singleTest(m1, m2, resultC, "divide", size, allocator, divide<Matrix>);
	singleTest(m1, m2, resultC, "divideHybrid", size, allocator, divideHybrid<Matrix>);
	singleTest(m1, m2, resultC, "square", size, allocator, square<Matrix>);
	singleTest(m1, m2, resultC, "squareHybrid", size, allocator, squareHybrid<Matrix>);
	singleTest(m1, m2, resultC, "squareZ", size, allocator, divideZ<Matrix>);
	singleTest(m1, m2, resultC, "squareZHybrid", size, allocator, divideZHybrid<Matrix>);
	singleTest(m1, m2, resultC, "strassen", size, allocator, strassen<Matrix>);
	singleTest(m1, m2, resultC, "strassenHybrid", size, allocator, strassenHybrid<Matrix>);
	singleTest(m1, m2, resultC, "strassenZ", size, allocator, strassenZ<Matrix>);
	singleTest(m1, m2, resultC, "strassenZHybrid", size, allocator, strassenZHybrid<Matrix>);
	allocator->restore(state);
}

}

int main() {
	Pool<Core> allocator;
	typedef Classic<Core, unsigned> Matrix;
	size_t size(16384);
	for (size_t s(2); s <= size; s *= 2)
		test<Matrix>(&allocator, s);
	return 0;
}
