#include <graph/graph.h>
#include <graph/maxmatch.h>
#include <graph/dc.h>
#include <util/input_generator.h>
#include <util/stopwatch.h>
#include <util/assert.h>
#include <typeinfo>

using namespace Memory;
using namespace Allocator;
using namespace std;

namespace {

class Mark : public Graph::Vertex {
	bool enabled;
public:
	Mark() :
		enabled(false)
	{ }
	void setMark(bool enabled) { this->enabled = enabled; }
	bool isMarkSet() const { return enabled; }
};

size_t count(10);

template<class Graph, class Pool, class Function> void singleTest(Graph &graph, const char *name, size_t size, Pool &allocator, Function function) {
	Util::Stopwatch watch(name, graph.name, Graph::StorageType::name(), "void", size);
	typename Pool::State state(allocator);
	Pointer<Core, ::Graph::Edge> result;
	watch.start();
	for (size_t i(0); i < count; ++ i) {
		allocator.restore(state);
		result = function(graph, allocator);
	}
	watch.stop();
	Pointer<Core, bool> vertices(allocator[graph.n]);
	for (size_t i(0); i < graph.n; ++ i)
		vertices[i] = false;
	for (size_t i(0); i < graph.m; ++ i) {
		if (result[i].vertices[0] == result[i].vertices[1])
			break;
		for (size_t j(0); j < 2; ++ j) {
			ASSERT(!vertices[result[i].vertices[j]]);
			vertices[result[i].vertices[j]] = true;
		}
	}
	for (size_t i(0); i < graph.m * 2; ++ i)
		ASSERT(vertices[graph.edges[i].vertices[0]] || vertices[graph.edges[i].vertices[1]]);
	allocator.restore(state);
}

typedef Graph::Graph<Core, Mark> G;

void test(Pool<Core> &allocator, size_t size, G::Generator generator, int mult = 0) {
	Core::allowThreads = false;
	typename Pool<Core>::State state(allocator);
	G graph(G::generate(allocator, generator, size, mult));
	singleTest(graph, "maxMatchClassic", size, allocator, Graph::maxMatchClassic<G>);
	singleTest(graph, "maxMatchBRT", size, allocator, Graph::maxMatchBRT<G>);
	singleTest(graph, "maxMatchDC", size, allocator, Graph::maxMatchDC<G>);
	allocator.restore(state);
}

}

int main() {
	Pool<Core> allocator;
	size_t size(TEST_SIZE);
	size /= 3;
	for (double s(1); s * s < size; s *= 1.5)
		test(allocator, s, G::Fixed);
	size /= 160;
	for (double s(1); s < size; s *= 1.5)
		test(allocator, s, G::Count, 10);
	size *= 5;
	for (double s(1); s < size; s *= 1.5) {
		test(allocator, s, G::Count, 1);
		test(allocator, s, G::Grid);
		test(allocator, s, G::Triangulate, 1);
	}
	return 0;
}
