#ifndef GRAPH_BRT_H
#define GRAPH_BRT_H

#include <utility>
#include <memory/pointer.h>
#include <memory/allocator.h>

namespace Graph {

template<class Storage> class BRT {
public:
	typedef std::pair<size_t, size_t> Pair;
private:
	struct Block {
		Memory::Pointer<Storage, Block> next;
		Memory::Pointer<Storage, Pair> pairs;
		size_t count;
		Block() :
			count(0)
		{ }
	};
	size_t mask;
	size_t limit;
	struct Node {
		Memory::Pointer<Storage, Block> blocks;
		size_t nextSize;
		Node() :
			nextSize(0)
		{ }
	};
	Memory::Pointer<Storage, Node> tree;
	Memory::Mapping<Storage> rootBuffer;
	Memory::Allocator::ArrayBorrower<Storage, Block> listBorrower;
	Memory::Allocator::ArrayBorrower<Storage, Pair> pairBorrower;
	size_t rootSize, rootAlloc;
	static size_t getMask(size_t size) {
		size_t level(0);
		if (size)
			size --;
		while (size) {
			size /= 2;
			level ++;
		}
		if (level)
			return 1 << level - 1;
		else
			return 0;
	}
	static size_t getSize(size_t mask, size_t size) {
		size_t result(size + 1);
		while (mask) {
			result += mask;
			mask /= 2;
		}
		return result;
	}
	void prepare(size_t index) {
		if (index < limit && tree[index].nextSize) {
			Memory::Pointer<Storage, Block> block(listBorrower.get(1));
			block->next = tree[index].blocks;
			block->pairs = pairBorrower.get(tree[index].nextSize);
			tree[index].nextSize = 0;
			tree[index].blocks = block;
		}
	}
public:
	BRT(size_t size, Memory::Allocator::Pool<Storage> &allocator) :
		mask(getMask(size)),
		limit(getSize(mask, size)),
		tree(allocator[limit]), // This get's already allocated there
		rootBuffer(sizeof(Pair), "BRT root buffer"),
		listBorrower(allocator),
		pairBorrower(allocator),
		rootSize(0), rootAlloc(1)
	{
		tree[0].blocks = listBorrower.get(1);
		tree[0].blocks->pairs = Memory::Pointer<Storage, Pair>(rootBuffer);
	}
	void insert(size_t key, size_t value) {
		if (rootSize == rootAlloc) {
			rootBuffer.resize((rootAlloc *= 2) * sizeof (Pair));
			rootAlloc = rootBuffer.size() / sizeof (Pair);
			tree[0].blocks->pairs = Memory::Pointer<Storage, Pair>(rootBuffer);
		}
		tree[0].blocks->pairs[rootSize ++] = Pair(key, value);
		if (key & mask)
			tree[2].nextSize ++;
		else
			tree[1].nextSize ++;
	}
	class Result {
	private:
		size_t mSize;
		size_t pos;
		Memory::Pointer<Storage, Block> block;
		BRT<Storage> &owner;
		friend class BRT<Storage>;
		Result(size_t size, const Memory::Pointer<Storage, Block> &b, BRT<Storage> &o) :
			mSize(size),
			pos(0),
			block(b),
			owner(o)
		{ }
	public:
		size_t size() { return mSize; }
		bool empty() {
			return block;
		}
		size_t next() {
			size_t result(block->pairs[pos ++].second);
			if (pos == block->count) {
				Memory::Pointer<Storage, Block> b(block);
				block = b->next;
				owner.pairBorrower.put(b->pairs, b->count);
				owner.listBorrower.put(b, 1);
				pos = 0;
			};
			return result;
		}
	};
	Result get(size_t key) {
		size_t index(0);
		size_t mask(this->mask);
		tree[0].blocks->count = rootSize;
		rootSize = 0;
		while (mask) {
			size_t newMask(mask >> 1);
			size_t newIndex((index + 1) * 2 - 1);
			prepare(newIndex);
			prepare(newIndex + 1);
			Memory::Pointer<Storage, Block> b(tree[index].blocks);
			if (index)
				tree[index].blocks = Memory::Pointer<Storage, Block>();
			while (b) {
				for (size_t i(0); i < b->count; ++ i) {
					size_t where(!!(b->pairs[i].first & mask));
					tree[newIndex + where].blocks->pairs[tree[newIndex + where].blocks->count ++] = b->pairs[i];
					if (newMask) {
						size_t sub((newIndex + where + 1) * 2 - 1);
						sub += !!(b->pairs[i].first & newMask);
						tree[sub].nextSize ++;
					}
				}
				Memory::Pointer<Storage, Block> next(b->next);
				if (index) {
					pairBorrower.put(b->pairs, b->count);
					listBorrower.put(b, 1);
				}
				b = next;
			}
			index = newIndex + !!(key & mask);
			mask = newMask;
		}
		size_t size(0);
		for (Memory::Pointer<Storage, Block> b(tree[index].blocks); b; b = b->next)
			size += b->count;
		Result result(size, tree[index].blocks, *this);
		if (index)
			tree[index].blocks = Memory::Pointer<Storage, Block>();
		tree[index].nextSize = 0;
		return result;
	}
};

template<class Storage> class BPT {
public:
	struct Buffer {
		Memory::Pointer<Storage, Buffer> next;
		Memory::Pointer<Storage, size_t> data;
		size_t size;
		Buffer() :
			size(0)
		{ }
	};
private:
	struct Node {
		size_t key, expect, rest;
		Memory::Pointer<Storage, Buffer> buffers;
		Node() :
			expect(0)
		{ }
	};
	void prepare(size_t index) {
		if (index < limit && tree[index].expect) {
			Memory::Pointer<Storage, Buffer> buffer(bufferBorrower.get(1));
			buffer->next = tree[index].buffers;
			buffer->data = dataBorrower.get(tree[index].expect);
			tree[index].expect = 0;
			tree[index].buffers = buffer;
		}
	}
	void cleanup(size_t index) {
		while (tree[index].buffers) {
			Memory::Pointer<Storage, Buffer> buffer(tree[index].buffers);
			tree[index].buffers = buffer->next;
			dataBorrower.put(buffer->data, buffer->size);
			bufferBorrower.put(buffer, 1);
		}
	}
	size_t levels, start, limit;
	Memory::Pointer<Storage, Node> tree;
	Memory::Allocator::ArrayBorrower<Storage, size_t> &dataBorrower;
	Memory::Allocator::ArrayBorrower<Storage, Buffer> &bufferBorrower;
	static size_t getLevels(size_t size) {
		size_t result(0);
		size --;
		while (size) {
			size /= 2;
			result ++;
		}
		return result;
	}
	static size_t getStart(size_t levels) {
		return (1 << levels) - 1;
	}
	template<class Array> std::pair<size_t, size_t> fill(size_t index, Array items) {
		if (index >= limit)
			return std::pair<size_t, size_t>(-1, 0);
		else if (index >= start) {
			tree[index].rest = 1;
			tree[index].key = items[index - start];
			return std::pair<size_t, size_t>(items[index - start], 1);
		} else {
			size_t newIndex((index + 1) * 2 - 1);
			std::pair<size_t, size_t> left(fill(newIndex, items)), right(fill(newIndex + 1, items));
			tree[index].key = left.first;
			tree[index].rest = left.second + right.second;
			return std::pair<size_t, size_t>(right.first, tree[index].rest);
		}
	}
public:
	template<class Array> BPT(size_t size, Array items, Memory::Allocator::Pool<Storage> &allocator, Memory::Allocator::ArrayBorrower<Storage, size_t> &dataBorrower_, Memory::Allocator::ArrayBorrower<Storage, Buffer> &bufferBorrower_) :
		levels(getLevels(size)),
		start(getStart(levels)),
		limit(start + size),
		tree(allocator[limit]),
		dataBorrower(dataBorrower_),
		bufferBorrower(bufferBorrower_)
	{
		if (size) {
			fill(0, items);
			tree[0].buffers = bufferBorrower_.get(1);
		}
	}
	size_t getNext(size_t count, Memory::Pointer<Storage, size_t> data) {
		if (!limit)
			return size_t(-1);
		bool release(false);
		tree[0].rest -= count;
		if (!tree[0].rest)
			return size_t(-1);
		tree[0].buffers->size = count;
		tree[0].buffers->data = data;
		for (size_t i(0); i < count; ++ i) {
			size_t index(1 + (data[i] > tree[0].key));
			tree[index].expect ++;
			tree[index].rest --;
		}
		size_t index(0);
		for (size_t level(0); level < levels; ++ level) {
			// We'll remove the result from here
			tree[index].rest --;
			size_t newIndex((index + 1) * 2 - 1);
			prepare(newIndex);
			prepare(newIndex + 1);
			Memory::Pointer<Storage, Buffer> buffer(tree[index].buffers);
			if (release)
				tree[index].buffers = Memory::Pointer<Storage, Buffer>();
			while (buffer) {
				for (size_t i(0); i < buffer->size; ++ i) {
					size_t where(newIndex + (buffer->data[i] > tree[index].key));
					tree[where].buffers->data[tree[where].buffers->size ++] = buffer->data[i];
					size_t sub((where + 1) * 2 - 1);
					sub += buffer->data[i] > tree[where].key;
					if (sub < limit) {
						tree[sub].expect ++;
						tree[sub].rest --;
					}
				}
				if (release) {
					Memory::Pointer<Storage, Buffer> old(buffer);
					buffer = buffer->next;
					dataBorrower.put(old->data, old->size);
					bufferBorrower.put(old, 1);
				} else
					buffer = Memory::Pointer<Storage, Buffer>();
			}
			release = true;
			if (tree[newIndex].rest) {
				index = newIndex;
			} else {
				cleanup(newIndex);
				index = newIndex + 1;
			}
		}
		if (release)
			cleanup(index);
		tree[index].rest --;
		return tree[index].key;
	}
};

};

#endif
