#ifndef GRAPH_DC2_H
#define GRAPH_DC2_H

#include <memory/pointer.h>
#include <memory/allocator.h>

#include <util/stack.h>
#include <util/assert.h>
#include <sort/sort.h>
#include "graph.h"

namespace Graph {
namespace DivideConquer {

namespace Internal {

// Both to hold edges and the results
struct BaseItem {
	size_t values[2];
	char hints;
};

template<size_t primaryIndex = 0> struct Item : public BaseItem {
	bool operator <(const Item<primaryIndex> &other) {
		return values[primaryIndex] < other.values[primaryIndex] ||
			(values[primaryIndex] == other.values[primaryIndex] &&
			 values[!primaryIndex] < other.values[!primaryIndex]);
	}
	bool operator >(const Item<primaryIndex> &other) {
		return values[primaryIndex] > other.values[primaryIndex] ||
			(values[primaryIndex] == other.values[primaryIndex] &&
			 values[!primaryIndex] > other.values[!primaryIndex]);
	}
	bool operator !=(const Item<primaryIndex> &other) {
		return (values[0] != other.values[0] || values[1] != other.values[1]);
	}
};

template<> struct Item<2> : public BaseItem {
	bool operator <(const Item<2> &other) {
		return hints < other.hints;
	}
	bool operator >(const Item<2> &other) {
		return hints > other.hints;
	}
};

template<class Storage> struct Result {
	Memory::Pointer<Storage, Item<> > assignments;
	size_t size;
};

template<class Storage> struct Input {
	Memory::Pointer<Storage, Item<> > edges;
	size_t size, depth;
};

enum Phase {
	Split = 'S',
	Contract = 'C',
	Join = 'J',
	Grow = 'G', // Recover from reducing the graph size and repeating the whole algorithm
	Free = 'F'// In between the first two tasks, frees some memory
};

template<class Storage> struct SubProblem {
	Input<Storage> input;
	Result<Storage> result;
};

template<class Storage> struct Level {
	Memory::Pointer<Storage, SubProblem<Storage> > problem, subproblems;
	typename Memory::Allocator::Pool<Storage>::State allocatorState;
	Phase phase;
	Level(const Memory::Pointer<Storage, SubProblem<Storage> > &problem_) :
#ifdef DEBUG_DATA
		id(gid ++),
#endif
		problem(problem_),
		phase(Split)
	{ }
	Level(const typename Memory::Allocator::Pool<Storage>::State &state) :
#ifdef DEBUG_DATA
		id(gid ++),
#endif
		allocatorState(state),
		phase(Free)
	{ }
#ifdef DEBUG_DATA
	size_t id;
	void dumpId(const char *note = "") {
		fprintf(DEBUG_DATA, "DATA: Level %zu phase %c sizes %zu %zu depth %zu %s\n", id, (char) phase, problem ? problem->input.size : 0xffffffff, problem ? problem->result.size : 0xffffffff, problem ? problem->input.depth : 0xffffffff, note);
	}
	void dump(Memory::Pointer<Storage, Item<> > edges, size_t size, char prefix) {
		fprintf(DEBUG_DATA, "DATA: dump %zu %c\n", id, prefix);
		for (size_t i(0); i < size; ++ i)
			fprintf(DEBUG_DATA, "DATA: %c %zu %zu %d\n", prefix, edges[i].values[0], edges[i].values[1], edges[i].hints);
		fprintf(DEBUG_DATA, "DATA: %c dump end\n", prefix);
	}
	void dumpInput() {
		dump(problem->input.edges, problem->input.size, 'I');
	}
	void dumpOutput() {
		dump(problem->result.assignments, problem->result.size, 'O');
	}
private:
	static size_t gid;
};

template<class Storage> size_t Level<Storage>::gid(0);
#else
};
#endif

template<class Storage, class Stack> void sortInverse(Memory::Pointer<Storage, Item<0> > items, size_t count, Stack *stack) {
	Algorithm::Sort::sort(Memory::Pointer<Storage, Item<1> >::casted(items), count, stack);
}

template<class Storage> void removeDupes(Input<Storage> &input) {
	if (!input.size)
		return;
	size_t pos(0);
	for (size_t i(1); i < input.size; ++ i)
		if (input.edges[i].values[0] != input.edges[pos].values[0] || input.edges[i].values[1] != input.edges[pos].values[1])
			input.edges[++ pos] = input.edges[i];
	input.size = pos + 1;
}

template<class Storage> void check(Memory::Pointer<Storage, Item<> > array, size_t size) {
	for (size_t i(0); i < size; ++ i)
		ASSERT(array[i].values[0] != array[i].values[1]);
}

template<class Storage> void contractWalk(Input<Storage> &what, Result<Storage> &by, size_t index) {
	size_t what_pos(0), by_pos(0);
#ifdef DEBUG_DC
	check(what.edges, what.size);
	check(by.assignments, by.size);
#endif
	while (what_pos < what.size && by_pos < by.size) {
		if (what.edges[what_pos].values[index] == by.assignments[by_pos].values[0]) {
			// A match
#ifdef DEBUG_DC
			fprintf(DEBUG_DC, "DEBUG: Changing %zu-%zu to ", what.edges[what_pos].values[0], what.edges[what_pos].values[1]);
#endif
			what.edges[what_pos ++].values[index] = by.assignments[by_pos].values[1];
#ifdef DEBUG_DC
			fprintf(DEBUG_DC, "%zu-%zu by %zu-%zu\n", what.edges[what_pos - 1].values[0], what.edges[what_pos - 1].values[1], by.assignments[by_pos].values[0], by.assignments[by_pos].values[1]);
#endif
		} else if (what.edges[what_pos].values[index] < by.assignments[by_pos].values[0])
			what_pos ++;
		else
			by_pos ++;
	}
#ifdef DEBUG_DC
	check(what.edges, what.size);
	check(by.assignments, by.size);
#endif
}

template<class Storage> void markScan(Memory::Pointer<Storage, Item<> > &edges, Memory::Pointer<Storage, size_t> &numbers, size_t size, size_t index) {
	size_t edgePos(0), numberPos(0);
	while (edgePos < size && numberPos < size * 2) {
		if (edges[edgePos].values[index] == numbers[numberPos])
			edges[edgePos ++].hints = 1;
		else if (edges[edgePos].values[index] < numbers[numberPos])
			++ edgePos;
		else
			++ numberPos;
	}
}

template<class Storage, class Stack> void uncontractCopy(Result<Storage> &half, Result<Storage> &contracted, Result<Storage> &result, Stack *stack) {
#ifdef DEBUG_DC
	check(half.assignments, half.size);
	check(contracted.assignments, contracted.size);
	check(result.assignments, result.size);
#endif
	sortInverse(half.assignments, half.size, stack);
	size_t pos(0);
	for (size_t i(0); i < half.size; ++ i) {
		while (pos < contracted.size - 1 && contracted.assignments[pos].values[0] < half.assignments[i].values[1])
			pos ++;
		// Copy the new assignment
		result.assignments[i + result.size] = half.assignments[i];
		// And correct it in case it got changed in contract
		if (pos < contracted.size && half.assignments[i].values[1] == contracted.assignments[pos].values[0]) {
#ifdef DEBUG_DC
			fprintf(DEBUG_DC, "DEBUG: Uncontracting from %zu-%zu to ", result.assignments[i + result.size].values[0], result.assignments[i + result.size].values[1]);
#endif
			result.assignments[i + result.size].values[1] = contracted.assignments[pos].values[1];
#ifdef DEBUG_DC
			fprintf(DEBUG_DC, "%zu-%zu\n", result.assignments[i + result.size].values[0], result.assignments[i + result.size].values[1]);
#endif
		}
	}
	result.size += half.size;
#ifdef DEBUG_DC
	check(half.assignments, half.size);
	check(contracted.assignments, contracted.size);
	check(result.assignments, result.size);
#endif
}

}

template<class Graph> Memory::Pointer<typename Graph::StorageType, size_t> components(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator) {
	typedef typename Graph::StorageType Storage;
	using Memory::Pointer;
	using namespace Internal;
	typedef Internal::Level<Storage> Level;
	typedef typename Memory::Allocator::Pool<Storage>::State State;
	Pointer<Storage, size_t> result(allocator[graph.n]);
	State resultState(allocator);

	// Prepare the big problem and push it to the stack
	Pointer<Storage, SubProblem<Storage> > problem(allocator);
	problem->input.edges = problem->result.assignments = Pointer<Storage, Item<> >(allocator[graph.m]);
	size_t pos(0);
	for (size_t i(0); i < graph.m * 2; ++ i)
		if (graph.edges[i].vertices[0] < graph.edges[i].vertices[1]) {
			problem->input.edges[pos].values[0] = graph.edges[i].vertices[0];
			problem->input.edges[pos ++].values[1] = graph.edges[i].vertices[1];
		}
	problem->input.size = pos;
	problem->input.depth = 0;
	Util::Stack<Storage, Level> stack;
	stack.push(Level(problem));
#ifdef DEBUG_DATA
	fprintf(DEBUG_DATA, "DATA: Start %zu %zu\n", graph.n, graph.m);
	stack.top().dumpId();
	stack.top().dumpInput();
#endif
	while (!stack.empty()) {
		Level l(stack.pop());
#ifdef DEBUG_DATA
		l.dumpId("activate");
#endif
		switch (l.phase) {
			case Free:
				allocator.restore(l.allocatorState);
#ifdef DEBUG_DATA
				fprintf(DEBUG_DATA, "DATA: done %zu\n", l.id);
#endif
				break;
			case Split: {
#ifdef DEBUG_DATA
				l.dumpInput();
#endif
				size_t mask(1 << l.problem->input.depth);
				// The trivial subproblem
				if (l.problem->input.size <= 1 || mask >= graph.n) {
					// Check if the trivial problem is really trivial or it got destroyed by the contractions
					if (l.problem->input.size > 1) {
						State state(allocator);
						// Are there any vertices with degree >1?
						Pointer<Storage, size_t> numbers(allocator[2 * l.problem->input.size]);
						for (size_t i(0); i < l.problem->input.size; ++ i) {
							for (size_t j(0); j < 2; ++ j)
								numbers[2 * i + j] = l.problem->input.edges[i].values[j];
							l.problem->input.edges[i].hints = 0;
						}
						Algorithm::Sort::sort(numbers, 2 * l.problem->input.size, &stack);
						size_t pos(0);
						size_t last(numbers[0]);
						bool written(false);
						for (size_t i(1); i < 2 * l.problem->input.size; ++ i)
							if (numbers[i] == last) {
								if (!written) {
									numbers[pos ++] = numbers[i];
									written = true;
								}
							} else {
								written = false;
								last = numbers[i];
							}
						if (pos) {
							// Yes, there are, we need to restart the splitting from the beginning
#ifdef DEBUG_DATA
							fprintf(DEBUG_DATA, "DATA: Recursion\n");
#endif
							Algorithm::Sort::sort(l.problem->input.edges, l.problem->input.size, &stack);
							markScan(l.problem->input.edges, numbers, l.problem->input.size, 0);
							sortInverse(l.problem->input.edges, l.problem->input.size, &stack);
							markScan(l.problem->input.edges, numbers, l.problem->input.size, 1);
							allocator.restore(state);
							Algorithm::Sort::sort(Pointer<Storage, Item<2> >::casted(l.problem->input.edges), l.problem->input.size, &stack);
#ifdef DEBUG_DC
							fprintf(DEBUG_DC, "DEBUG: Recursion split\n");
							for (size_t i(0); i < l.problem->input.size; ++ i)
								fprintf(DEBUG_DC, "DEBUG: Edge %zu-%zu (%d)\n", l.problem->input.edges[i].values[0], l.problem->input.edges[i].values[1], l.problem->input.edges[i].hints);
#endif
							size_t gpos(l.problem->input.size);
							for (size_t i(0); i < l.problem->input.size; ++ i)
								if (l.problem->input.edges[i].hints) {
									gpos = i;
									break;
								}
							l.allocatorState = state;
							l.subproblems = Pointer<Storage, SubProblem<Storage> >(allocator[2]);
							l.subproblems[0].input.edges = l.subproblems[0].result.assignments = l.problem->input.edges;
							l.subproblems[0].input.size = gpos;
							l.subproblems[1].input.edges = l.subproblems[1].result.assignments = l.problem->input.edges + gpos;
							l.subproblems[1].input.size = l.problem->input.size - gpos;
							l.subproblems[1].input.depth = 0;
							l.phase = Grow;
							stack.push(l);
							stack.push(Level(l.subproblems + 1));
#ifdef DEBUG_DATA
							stack.top().dumpId();
							stack.top().dumpInput();
#endif
#ifdef DEBUG_DC
							fprintf(DEBUG_DC, "DEBUG: Recursing with graph of %zu edges (of %zu)\n", l.subproblems[1].input.size, l.problem->input.size);
#endif
							break;
						}
						allocator.restore(state);
					}
					l.problem->result.assignments = l.problem->input.edges;
					l.problem->result.size = l.problem->input.size;
					Algorithm::Sort::sort(l.problem->result.assignments, l.problem->result.size, &stack);
#ifdef DEBUG_DATA
					fprintf(DEBUG_DATA, "DATA: Through\n");
					l.dumpOutput();
					fprintf(DEBUG_DATA, "DATA: done %zu\n", l.id);
#endif
#ifdef DEBUG_DC
					fprintf(DEBUG_DC, "DEBUG: Let through\n");
					for (size_t i(1); i < l.problem->result.size; ++ i)
						ASSERT(l.problem->result.assignments[i - 1].values[0] != l.problem->result.assignments[i].values[0]);
#endif
					break;
				}
#ifdef DEBUG_DC
				fprintf(DEBUG_DC, "DEBUG: Start splitting by mask %zu\n", mask);
#endif
				Pointer<Storage, Item<2> > array(Pointer<Storage, Item<2> >::casted(l.problem->input.edges));
				size_t counts[] = {0, 0, 0};
				for (size_t i(0); i < l.problem->input.size; ++ i) {
					bool bits[2];
					for (size_t j(0); j < 2; ++ j)
						bits[j] = array[i].values[j] & mask;
					// They differ -> 2, otherwise the value of the bit.
					size_t target(2 * (bits[0] ^ bits[1]) + (bits[0] & bits[1]));
					array[i].hints = 2 - target;
					counts[target] = 1;
#ifdef DEBUG_DC
					fprintf(DEBUG_DC, "DEBUG: Edge %zu(%d)-%zu(%d) to group %zu\n", array[i].values[0], (int) bits[0], array[i].values[1], (int) bits[1], target);
#endif
					// If the left end is in right part, flip it
					if (bits[0] && target == 2) {
						size_t tmp(array[i].values[0]);
						array[i].values[0] = array[i].values[1];
						array[i].values[1] = tmp;
					}
				}
				if (counts[0] + counts[1] + counts[2] <= 1) {
					// No split happened, don't bother to create the subtasks
					l.problem->input.depth ++;
#ifdef DEBUG_DATA
					fprintf(DEBUG_DATA, "DATA: No-split\n");
#endif
#ifdef DEBUG_DC
					fprintf(DEBUG_DC, "DEBUG: No-split\n");
#endif
					stack.push(l);
					break;
				}
				Algorithm::Sort::sort(array, l.problem->input.size, &stack);
				l.subproblems = Pointer<Storage, SubProblem<Storage> >(allocator[3]);
				l.allocatorState = typename Memory::Allocator::Pool<Storage>::State(allocator);
				for (size_t i(0); i < 3; ++ i) {
					l.subproblems[i].input.edges = Pointer<Storage, Item<> >::casted(array);
					l.subproblems[i].input.size = 0;
					l.subproblems[i].input.depth = l.problem->input.depth + 1;
				}
				for (size_t i(0); i < l.problem->input.size; ++ i) {
					++ l.subproblems[array[i].hints].input.size;
					for (size_t j(array[i].hints + 1); j < 3; ++ j)
						l.subproblems[j].input.edges ++;
				}
				for (size_t i(0); i < 3; ++ i)
					l.subproblems[i].result.assignments = l.subproblems[i].input.edges;
#ifdef DEBUG_DC
				fprintf(DEBUG_DC, "DEBUG: All\n");
				for (size_t i(0); i < l.problem->input.size; ++ i)
					fprintf(DEBUG_DC, "DEBUG: %zu-%zu %d\n", array[i].values[0], array[i].values[1], array[i].hints);
				for (size_t i(0); i < 3; ++ i) {
					fprintf(DEBUG_DC, "DEBUG: Splitted %zu\n", i);
					for (size_t j(0); j < l.subproblems[i].input.size; ++ j)
						fprintf(DEBUG_DC, "DEBUG: %zu-%zu %d\n", l.subproblems[i].input.edges[j].values[0], l.subproblems[i].input.edges[j].values[1], (int) l.subproblems[i].input.edges[j].hints);
				}
#endif
				// Prepare the next tasks. This one will do the contraction and we have two new there.
				// There's one to free some memory in between.
				l.phase = Contract;
				stack.push(l);
				stack.push(Level(l.subproblems + 1));
#ifdef DEBUG_DATA
				fprintf(DEBUG_DATA, "DATA: Split\n");
				stack.top().dumpId();
				stack.top().dumpInput();
#endif
				stack.push(Level(l.allocatorState));
#ifdef DEBUG_DATA
				stack.top().dumpId();
#endif
				stack.push(Level(l.subproblems + 2));
#ifdef DEBUG_DATA
				stack.top().dumpId();
				stack.top().dumpInput();
#endif
				break;
			}
			case Grow: {
				ASSERT(l.subproblems[1].result.assignments == l.problem->result.assignments + l.subproblems[0].input.size);
				l.problem->result.size = l.subproblems[0].input.size + l.subproblems[1].result.size;
				allocator.restore(l.allocatorState);
				Algorithm::Sort::sort(l.problem->result.assignments, l.problem->result.size, &stack);
#ifdef DEBUG_DATA
				l.dumpOutput();
				fprintf(DEBUG_DATA, "DATA: done %zu\n", l.id);
#endif
#ifdef DEBUG_DC
				fprintf(DEBUG_DC, "DEBUG: Grow to %zu edges\n", l.problem->result.size);
#endif
				break;
			}
			case Contract: {
				allocator.restore(l.allocatorState);
				size_t size(l.subproblems[0].input.size);
				// Copy the new labels
#ifdef DEBUG_DC
				fprintf(DEBUG_DC, "DEBUG: Contracting\n");
#endif
				sortInverse(l.subproblems[0].input.edges, size, &stack);
				contractWalk(l.subproblems[0].input, l.subproblems[1].result, 1);
				Algorithm::Sort::sort(l.subproblems[0].input.edges, size, &stack);
				contractWalk(l.subproblems[0].input, l.subproblems[2].result, 0);
				// Remove dupes
				Algorithm::Sort::sort(l.subproblems[0].input.edges, size, &stack);
				removeDupes(l.subproblems[0].input);
				// Now push the last subproblem and uncontraction
				l.phase = Join;
				stack.push(l);
				stack.push(Level(l.subproblems));
#ifdef DEBUG_DATA
				stack.top().dumpId();
				stack.top().dumpInput();
#endif
				break;
			}
			case Join: {
				allocator.restore(l.allocatorState);
				// Copy the part from third subproblem.
				l.problem->result.size = l.subproblems[0].result.size;
				l.problem->result.assignments = l.subproblems[0].result.assignments;
				// Copy and modify the two subproblems to the final result
				uncontractCopy(l.subproblems[1].result, l.subproblems[0].result, l.problem->result, &stack);
				uncontractCopy(l.subproblems[2].result, l.subproblems[0].result, l.problem->result, &stack);
				// Sort it for further use
				Algorithm::Sort::sort(l.problem->result.assignments, l.problem->result.size, &stack);
#ifdef DEBUG_DATA
				l.dumpOutput();
				fprintf(DEBUG_DATA, "DATA: done %zu\n", l.id);
#endif
#ifdef DEBUG_DC
				fprintf(DEBUG_DC, "DEBUG: Joined\n");
				for (size_t i(0); i < l.problem->result.size; ++ i)
					fprintf(DEBUG_DC, "DEBUG: %zu-%zu\n", l.problem->result.assignments[i].values[0], l.problem->result.assignments[i].values[1]);
				for (size_t i(1); i < l.problem->result.size; ++ i)
					ASSERT(l.problem->result.assignments[i - 1].values[0] != l.problem->result.assignments[i].values[0]);
#endif
				break;
			}
		}
	}
#ifdef DEBUG_DATA
	fprintf(DEBUG_DATA, "DATA: Finish\n");
#endif
	pos = 0;
	for (size_t i(0); i < graph.n; ++ i) {
		if (pos < problem->result.size && problem->result.assignments[pos].values[0] == i)
			result[i] = problem->result.assignments[pos ++].values[1];
		else
			result[i] = i;
	}
	allocator.restore(resultState);
	return result;
}

struct FlaggedEdge : public Edge {
	bool active: 1;
	bool operator <(const FlaggedEdge &other) {
		return vertices[0] < other.vertices[0] || (vertices[0] == other.vertices[0] && vertices[1] < other.vertices[1]);
	}
	bool operator ==(const FlaggedEdge &other) {
		return vertices[0] == other.vertices[0] && vertices[1] == vertices[1];
	}
	bool operator >(const FlaggedEdge &other) {
		return (!(*this < other || *this == other));
	}
};

enum MatchPhase {
	Split = 'S',
	LeftDone = 'L',
	RightDone = 'R',
	MiddleDone = 'M'
};

template<class Graph> struct MatchLevel {
	Memory::Pointer<typename Graph::StorageType, size_t> usedResult, usedBuffer;
	Memory::Pointer<typename Graph::StorageType, FlaggedEdge> edges, edgeBuffers[3];
	Memory::Pointer<typename Graph::StorageType, MatchLevel<Graph> > subproblems[3];
	typename Memory::Allocator::Pool<typename Graph::StorageType>::State state;
	size_t size;
	size_t depth;
	MatchPhase phase;
};

template<class Graph> Memory::Pointer<typename Graph::StorageType, Edge> maxMatch(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator) {
	typedef typename Graph::StorageType Storage;
	using Memory::Pointer;
	Pointer<Storage, Edge> result(allocator[graph.n / 2 + 1]);
	size_t pos(0);
	typename Memory::Allocator::Pool<Storage>::State state(allocator);
	Pointer<Storage, size_t> usedResult(allocator[graph.n + 1]), usedBuffer(allocator[graph.n + 1]);
	size_t endMark(graph.n + 1);
	typedef Pointer<Storage, FlaggedEdge> Edges;
	Edges edgeBuffers[4];
	for (size_t i(0); i < 4; ++ i)
		edgeBuffers[i] = Edges(allocator[graph.m * 2]);
	typedef MatchLevel<Graph> Level;
	// They are already sorted in the graph
	for (size_t i(0); i < graph.m * 2; ++ i) {
		edgeBuffers[0][i].vertices[0] = graph.edges[i].vertices[0];
		edgeBuffers[0][i].vertices[1] = graph.edges[i].vertices[1];
		edgeBuffers[0][i].active = true;
	}
	Pointer<Storage, Level> toplevel(allocator);
	toplevel->phase = Split;
	toplevel->depth = 0;
	toplevel->size = graph.m * 2;
	toplevel->edges = edgeBuffers[0];
	for (size_t i(0); i < 3; ++ i)
		toplevel->edgeBuffers[i] = edgeBuffers[i + 1];
	toplevel->usedResult = usedResult;
	toplevel->usedBuffer = usedBuffer;
	Util::Stack<Storage, Pointer<Storage, Level> > stack;
	stack.push(toplevel);
	while (!stack.empty()) {
		Pointer<Storage, Level> task(stack.pop());
		if (task->size == 0)
			continue;
		size_t mask(1 << task->depth);
		// The trivial case
		if (mask >= graph.n) {
#ifdef DEBUG_DC
			fprintf(DEBUG_DC, "DATA: Lowest level\n");
			for (size_t i(0); i < task->size; ++ i)
				fprintf(DEBUG_DC, "DATA: Edge %zu %zu %d\n", task->edges[i].vertices[0], task->edges[i].vertices[1], task->edges[i].active);
#endif
			// Flip the edges so the lower end is first
			for (size_t i(0); i < task->size; ++ i) {
				if (task->edges[i].vertices[0] > task->edges[i].vertices[1]) {
					size_t tmp(task->edges[i].vertices[0]);
					task->edges[i].vertices[0] = task->edges[i].vertices[1];
					task->edges[i].vertices[1] = tmp;
				}
			}
			Algorithm::Sort::sort(task->edges, task->size, &stack);
			size_t osize(0);
			for (size_t i(0); i < task->size; i += 2) {
				if (task->edges[i].active && task->edges[i + 1].active) {
					task->usedResult[osize ++] = result[pos].vertices[0] = task->edges[i].vertices[0];
					task->usedResult[osize ++] = result[pos ++].vertices[1] = task->edges[i].vertices[1];
				}
			}
			Algorithm::Sort::sort(task->usedResult, osize, &stack);
			task->size = osize;
			continue;
		}
		switch (task->phase) {
			case Split: {
#ifdef DEBUG_DC
				fprintf(DEBUG_DC, "DATA: Split level (%zu)\n", task->depth);
				for (size_t i(0); i < task->size; ++ i)
					fprintf(DEBUG_DC, "DATA: Edge %zu %zu %d\n", task->edges[i].vertices[0], task->edges[i].vertices[1], task->edges[i].active);
#endif
				size_t sizes[4] = {0, 0, 0, 0};
				// Split to the three subtasks
				for (size_t i(0); i < task->size; ++ i) {
					bool bits[2];
					for (size_t j(0); j < 2; ++ j)
						bits[j] = task->edges[i].vertices[j] & mask;
					// They differ -> 2, otherwise the value of the bit.
					size_t target(2 * (bits[0] ^ bits[1]) + (bits[0] & bits[1]));
					task->edgeBuffers[target][sizes[target] ++] = task->edges[i];
				}
				// Prepare an with four buffers
				for (size_t i(0); i < 3; ++ i) {
					edgeBuffers[i] = task->edgeBuffers[i];
				}
				// Prepare the subtasks
				task->state = typename Memory::Allocator::Pool<Storage>::State(allocator);
				edgeBuffers[3] = task->edges;
				for (size_t i(0); i < 3; ++ i) {
					Pointer<Storage, Level> sub(allocator);
					task->subproblems[i] = sub;
					sub->usedBuffer = task->usedResult;
					sub->usedResult = task->usedBuffer;
					sub->edges = edgeBuffers[i];
					sub->size = sizes[i];
					sub->depth = task->depth + 1;
					sub->phase = Split;
					size_t buffPos(0);
					for (size_t j(0); j < 4; ++ j)
						if (i != j)
							sub->edgeBuffers[buffPos ++] = edgeBuffers[j] + sizes[j];
				}
				task->phase = LeftDone;
				stack.push(task);
				stack.push(task->subproblems[0]);
#ifdef DEBUG_DC
				fprintf(DEBUG_DC, "DATA: Split stat %zu %zu %zu\n", sizes[0], sizes[1], sizes[2]);
#endif
				break;
			}
			case LeftDone:
#ifdef DEBUG_DC
				fprintf(DEBUG_DC, "DATA: Left done (%zu), returned %zu\n", task->depth, task->subproblems[0]->size);
#endif
				task->subproblems[1]->usedResult = task->subproblems[0]->usedResult + task->subproblems[0]->size;
				task->phase = RightDone;
				stack.push(task);
				stack.push(task->subproblems[1]);
				break;
			case RightDone: {
#ifdef DEBUG_DC
				fprintf(DEBUG_DC, "DATA: Right done (%zu), returned %zu\n", task->depth, task->subproblems[1]->size);
				fprintf(DEBUG_DC, "DATA: Forbidden left\n");
				for (size_t i(0); i < task->subproblems[0]->size; ++ i)
					fprintf(DEBUG_DC, "DATA: %zu\n", task->subproblems[0]->usedResult[i]);
				fprintf(DEBUG_DC, "DATA: Forbidden right\n");
				for (size_t i(0); i < task->subproblems[1]->size; ++ i)
					fprintf(DEBUG_DC, "DATA: %zu\n", task->subproblems[1]->usedResult[i]);
#endif
				task->subproblems[2]->usedResult = task->subproblems[1]->usedResult + task->subproblems[1]->size;
				size_t positions[2] = {0, 0};
				for (size_t i(0); i < task->subproblems[2]->size; ++ i) {
					for (size_t j(0); j < 2; ++ j) {
						while (positions[j] < task->subproblems[j]->size && task->subproblems[j]->usedResult[positions[j]] < task->subproblems[2]->edges[i].vertices[0])
							positions[j] ++;
						if (positions[j] < task->subproblems[j]->size && task->subproblems[2]->edges[i].vertices[0] == task->subproblems[j]->usedResult[positions[j]])
							task->subproblems[2]->edges[i].active = false;
					}
				}
				task->phase = MiddleDone;
				stack.push(task);
				stack.push(task->subproblems[2]);
				break;
			}
			case MiddleDone:
#ifdef DEBUG_DC
				fprintf(DEBUG_DC, "DATA: Middle (%zu) done, returned %zu\n", task->depth, task->subproblems[2]->size);
				fprintf(DEBUG_DC, "DATA: Join stats %zu %zu %zu\n", task->subproblems[0]->size, task->subproblems[1]->size, task->subproblems[2]->size);
#endif
				size_t opos(0);
				size_t positions[3] = {0, 0, 0};
				for (;;) {
					size_t bestPos;
					size_t bestVal(endMark);
					for (size_t i(0); i < 3; ++ i)
						if (positions[i] < task->subproblems[i]->size && task->subproblems[i]->usedResult[positions[i]] < bestVal) {
							bestPos = i;
							bestVal = task->subproblems[i]->usedResult[positions[i]];
						}
					if (bestVal == endMark)
						break;
					task->usedResult[opos ++] = task->subproblems[bestPos]->usedResult[positions[bestPos] ++];
				}
#ifdef DEBUG_DC
				ASSERT(task->subproblems[0]->size + task->subproblems[1]->size + task->subproblems[2]->size == opos);
				for (size_t i(1); i < opos; ++ i)
					ASSERT(task->usedResult[i - 1] < task->usedResult[i]);
#endif
				task->size = opos;
				allocator.restore(task->state);
				break;
		}
	}
	allocator.restore(state);
	result[pos].vertices[0] = result[pos].vertices[1] = 0;
	return result;
}

}

template<class Graph> Memory::Pointer<typename Graph::StorageType, size_t> componentsDC(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator) {
	return DivideConquer::components(graph, allocator);
}

template<class Graph> Memory::Pointer<typename Graph::StorageType, Edge> maxMatchDC(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator) {
	return DivideConquer::maxMatch(graph, allocator);
}

}

#endif
