#ifndef GRAPH_GRAPH_H
#define GRAPH_GRAPH_H

#include <memory/memory.h>
#include <memory/pointer.h>
#include <memory/allocator.h>
#include <memory/mapping.h>
#include <sort/sort.h>
#include <util/assert.h>
#include <cmath>
#include <vector>

namespace Graph {

struct Vertex { size_t firstEdge; };
struct Edge { size_t vertices[2]; };

template<class Storage, class VertexType = ::Graph::Vertex, class EdgeType = ::Graph::Edge> class Graph {
public:
	typedef VertexType Vertex;
	typedef EdgeType Edge;
	Graph(size_t n_, size_t m_, Memory::Allocator::Pool<Storage> &allocator, const char *name_) :
		n(n_),
		m(m_),
		vertices(allocator[n + 1]),
		edges(allocator[m * 2]),
		name(name_)
	{ }
	const size_t n, m;
	Memory::Pointer<Storage, Vertex> vertices;
	Memory::Pointer<Storage, Edge> edges;
	const char *const name;
private:
	template<size_t primaryIndex> struct SortedEdge : public Edge {
		bool operator <(const SortedEdge<primaryIndex> &other) {
			return this->vertices[primaryIndex] < other.vertices[primaryIndex] ||
				(this->vertices[primaryIndex] == other.vertices[primaryIndex] &&
				 this->vertices[!primaryIndex] < other.vertices[!primaryIndex]);
		}
		bool operator >(const SortedEdge<primaryIndex> &other) {
			return this->vertices[primaryIndex] > other.vertices[primaryIndex] ||
				(this->vertices[primaryIndex] == other.vertices[primaryIndex] &&
				 this->vertices[!primaryIndex] > other.vertices[!primaryIndex]);
		}
	};
	static void edgeSort(Memory::Pointer<Storage, Edge> edges, size_t m) {
		Algorithm::Sort::sort(Memory::Pointer<Storage, SortedEdge<0> >::casted(edges), m);
	}
public:
	void index() {
		edgeSort(edges, 2 * m);
		vertices[0].firstEdge = 0;
		size_t lastVertex = 0;
		for (size_t i(0); i < 2 * m; ++ i) {
			while (lastVertex < edges[i].vertices[0])
				vertices[++ lastVertex].firstEdge = i;
		}
		for (size_t i(lastVertex + 1); i <= n; ++ i)
			vertices[i].firstEdge = 2 * m;
	}
	static Graph<Storage, VertexType, EdgeType> generateProbability(size_t n, double p, Memory::Allocator::Pool<Storage> &allocator) {
		size_t alloc;
		if (n >= 100)
			alloc = (n * n - 1) / 2 * p * 10;
		else
			alloc = (n * n - 1) / 2;
		Memory::Mapping<Storage> temporary(alloc * sizeof(Edge), "Temporary edge storage");
		Memory::Pointer<Storage, Edge> edges(temporary);
		size_t epos(0);
		for (size_t i(0); i < n; ++ i)
			for (size_t j(i + 1); j < n; ++ j) {
				if (random() < p * RAND_MAX) {
					edges[epos].vertices[0] = i;
					edges[epos].vertices[1] = j;
					epos ++;
				}
				ASSERT(epos <= alloc);
			}
		Graph<Storage, VertexType, EdgeType> result(n, epos, allocator, "dense");
		for (size_t i(0); i < epos; ++ i) {
			result.edges[2 * i] = edges[i];
			result.edges[2 * i + 1].vertices[0] = edges[i].vertices[1];
			result.edges[2 * i + 1].vertices[1] = edges[i].vertices[0];
		}
		result.index();
		return result;
	}
	static Graph<Storage, VertexType, EdgeType> generateCount(size_t n, size_t m, Memory::Allocator::Pool<Storage> &allocator) {
		if (m > n * (n - 1) / 2)
			m = n * (n - 1) / 2;
		Graph<Storage, VertexType, EdgeType> result(n, m, allocator, n < m ? "sparse" : "treelike");
		size_t count(0);
		// Generate enough edges (eliminate dupes, and if it isn't enough, try again)
		while (count < m) {
			for (size_t i(count) ; i < 2 * m;) {
				result.edges[i].vertices[0] = random() % n;
				result.edges[i].vertices[1] = random() % n;
				if (result.edges[i].vertices[0] < result.edges[i].vertices[1])
					++ i;
			}
			edgeSort(result.edges, 2 * m);
			size_t pos(1);
			for (size_t i(1); i < 2 * m; ++ i)
				if (result.edges[i].vertices[0] != result.edges[pos - 1].vertices[0] ||
					result.edges[i].vertices[1] != result.edges[pos - 1].vertices[1])
					result.edges[pos ++] = result.edges[i];
			count = pos;
		}
		// Drop few edges if there are too many
		while (count > m) {
			result.edges[random() % count] = result.edges[count - 1];
			count --;
		}
		// Make inverse copies
		for (size_t i(0); i < m; ++ i) {
			result.edges[i + m].vertices[0] = result.edges[i].vertices[1];
			result.edges[i + m].vertices[1] = result.edges[i].vertices[0];
		}
		// Create the result
		result.index();
		return result;
	}
	static Graph<Storage, VertexType, EdgeType> generateGrid(size_t n, double p, Memory::Allocator::Pool<Storage> &allocator) {
		Memory::Mapping<Storage> temporary(2 * n * sizeof(Edge), "Temporary edge storage");
		Memory::Pointer<Storage, Edge> edges(temporary);
		size_t cols(sqrt(n));
		size_t rows(1 + n / cols);
		size_t edgePos(0);
		for (size_t i(0); i < cols; ++ i)
			for (size_t j(0); i < rows; ++ i) {
				size_t vertex(i + j * cols);
				if (vertex >= n)
					continue;
				if (i && random() < p * RAND_MAX) {
					edges[edgePos].vertices[0] = vertex;
					edges[edgePos ++].vertices[1] = vertex - 1;
				}
				if (j && random() < p * RAND_MAX) {
					edges[edgePos].vertices[0] = vertex;
					edges[edgePos ++].vertices[1] = vertex - cols;
				}
			}
		Graph<Storage, VertexType, EdgeType> result(n, edgePos, allocator, "grid");
		for (size_t i(0); i < edgePos; ++ i)
			for (size_t j(0); j < 2; ++ j) {
				result.edges[2 * i].vertices[j] = edges[i].vertices[j];
				result.edges[2 * i + 1].vertices[j] = edges[i].vertices[1 - j];
			}
		result.index();
		return result;
	}
	static Graph<Storage, VertexType, EdgeType> generateTriangulation(size_t n, double p, Memory::Allocator::Pool<Storage> &allocator) {
		// WARNING: This uses usual non-simulable pointers/data structures (i.e. stl).
		if (n < 3) {
			Graph<Storage, VertexType, EdgeType> result(n, n / 2, allocator, "triangulation");
			if (n > 1) {
				result.edges[0].vertices[0] = result.edges[1].vertices[1] = 0;
				result.edges[1].vertices[0] = result.edges[0].vertices[1] = 1;
			}
			result.index();
			return result;
		} else {
			struct Triangle {
				size_t v[3];
				Triangle(size_t v1, size_t v2, size_t v3) {
					v[0] = v1;
					v[1] = v2;
					v[2] = v3;
				}
				Triangle(size_t vertices[3]) {
					for (size_t i(0); i < 3; ++ i)
						v[i] = vertices[i];
				}
			};
			struct MaybeEdge : public Edge {
				bool active;
				MaybeEdge(size_t v1, size_t v2, double p) :
					active(random() < p * RAND_MAX)
				{
					this->vertices[0] = v1;
					this->vertices[1] = v2;
				}
			};
			std::vector<Triangle> faces;
			std::vector<MaybeEdge> edges;
			size_t position(3);
			faces.push_back(Triangle(0, 1, 2));
			faces.push_back(Triangle(0, 1, 2));
			edges.push_back(MaybeEdge(0, 1, p));
			edges.push_back(MaybeEdge(1, 2, p));
			edges.push_back(MaybeEdge(2, 0, p));
			while (position < n) {
				size_t face(random() % faces.size());
				for (size_t i(0); i < 3; ++ i) {
					size_t v(faces[face].v[i]);
					edges.push_back(MaybeEdge(v, position, position));
					faces[face].v[i] = position;
					faces.push_back(Triangle(faces[face].v));
					faces[face].v[i] = v;
				}
				faces[face] = faces[faces.size() - 1];
				faces.pop_back();
				++ position;
			}
			size_t ecount(0);
			for (typename std::vector<MaybeEdge>::const_iterator i(edges.begin()); i != edges.end(); ++ i)
				if (i->active)
					++ ecount;
			Graph<Storage, VertexType, EdgeType> result(n, ecount, allocator, "triangulation");
			ecount = 0;
			for (typename std::vector<MaybeEdge>::const_iterator i(edges.begin()); i != edges.end(); ++ i) {
				if (!i->active)
					continue;
				for (size_t j(0); j < 2; ++ j) {
					result.edges[ecount - j].vertices[j] = result.edges[ecount - j + 1].vertices[1 - j] = i->vertices[j];
					ecount ++;
				}
			}
			result.index();
			return result;
		}
	}
	enum Generator {
		Fixed,
		Count,
		Grid,
		Triangulate
	};
	static Graph<Storage, VertexType, EdgeType> generate(Memory::Allocator::Pool<Storage> &allocator, Generator generator, size_t size, int mult = 0) {
		switch (generator) {
			case Fixed:
				return generateProbability(size, 0.1, allocator);
			case Count:
				return generateCount(size, mult * size, allocator);
			case Grid:
				return generateGrid(size, 0.9, allocator);
			case Triangulate:
				return generateTriangulation(size, 0.9, allocator);
		}
	}
	typedef Storage StorageType;
};

}

#endif
