#ifndef GRAPH_MAXMATCH_H
#define GRAPH_MAXMATCH_H

#include "graph.h"
#include "brt.h"
#include <memory/pointer.h>
#include <sort/sort.h>
#include <util/stack.h>

namespace Graph {

template<class Graph> Memory::Pointer<typename Graph::StorageType, Edge> maxMatchClassic(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator) {
	typedef typename Graph::StorageType Storage;
	Memory::Pointer<Storage, Edge> edges(allocator[graph.n / 2 + 1]);
	typename Memory::Allocator::Pool<Storage>::State state(allocator);
	Memory::Pointer<Storage, bool> used(allocator[graph.n]);
	for (size_t i(0); i < graph.n; ++ i)
		used[i] = false;
	size_t pos(0);
	for (size_t i(0); i < graph.m * 2; ++ i)
		if (graph.edges[i].vertices[0] < graph.edges[i].vertices[1] && !used[graph.edges[i].vertices[0]] && !used[graph.edges[i].vertices[1]]) {
			used[graph.edges[i].vertices[0]] = true;
			used[graph.edges[i].vertices[1]] = true;
			edges[pos ++] = graph.edges[i];
		}
	allocator.restore(state);
	edges[pos].vertices[0] = edges[pos].vertices[1] = graph.m + 1;
	return edges;
}

template<class Graph> Memory::Pointer<typename Graph::StorageType, Edge> maxMatchBRT(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator) {
	typedef typename Graph::StorageType Storage;
	Memory::Pointer<Storage, Edge> edges(allocator[graph.n / 2 + 1]);
	typename Memory::Allocator::Pool<Storage>::State state(allocator);
	Memory::Pointer<Storage, size_t> forbidden(allocator[graph.n - 1]);
	Util::Stack<Storage, unsigned char> stack;
	BRT<Storage> brt(graph.n, allocator);
	for (size_t i(0); i < graph.n; ++ i)
		graph.vertices[i].setMark(false);
	size_t pos(0);
	for (size_t i(0); i < graph.n; ++ i) {
		if (graph.vertices[i].isMarkSet())
			continue;
		// Extract the "forbidden" ones from the brt
		typename BRT<Storage>::Result used(brt.get(i));
		size_t size(used.size());
		for (size_t j(0); j < size; ++ j)
			forbidden[j] = used.next();
		Algorithm::Sort::sort(forbidden, size, &stack);
		size_t forbid(0);
		for (size_t neigh(graph.vertices[i].firstEdge); neigh < graph.vertices[i + 1].firstEdge; ++ neigh) {
			size_t neighbor(graph.edges[neigh].vertices[1]);
			if (neighbor < i)
				continue; // We examined this edge already in the other direction
			while (forbid < size && forbidden[forbid] < neighbor)
				forbid ++;
			if (forbid < size && forbidden[forbid] == neighbor)
				continue; // Can't use this edge, the other end is already used
			// The edge is OK now
			edges[pos].vertices[0] = i;
			edges[pos ++].vertices[1] = neighbor;
			graph.vertices[neighbor].setMark(true);
			for (size_t j(graph.vertices[neighbor].firstEdge); j < graph.vertices[neighbor + 1].firstEdge; ++ j)
				brt.insert(graph.edges[j].vertices[1], neighbor);
			break;
		}
	}
	allocator.restore(state);
	edges[pos].vertices[0] = edges[pos].vertices[1] = 0;
	return edges;
}

}

#endif
