#ifndef GRAPH_TREEISH_H
#define GRAPH_TREEISH_H

#include "graph.h"
#include "brt.h"
#include <util/stack.h>
#include <util/queue.h>

namespace Graph {
namespace Treeish {

namespace Internal {

template<class Graph> class Selector {
private:
	Memory::Pointer<typename Graph::StorageType, typename Graph::Edge> edges;
public:
	Selector(Memory::Pointer<typename Graph::StorageType, typename Graph::Edge> edges_) :
		edges(edges_)
	{ }
	size_t operator [](size_t index) { return edges[index].vertices[1]; }
};

template<class Graph> struct WalkState {
	BRT<typename Graph::StorageType> brt;
	Memory::Pointer<typename Graph::StorageType, BPT<typename Graph::StorageType> > bpts;
	Memory::Allocator::ArrayBorrower<typename Graph::StorageType, size_t> dataBorrower;
	Memory::Allocator::ArrayBorrower<typename Graph::StorageType, typename BPT<typename Graph::StorageType>::Buffer> bufferBorrower;
	Memory::Mapping<typename Graph::StorageType> transfBuff;
	size_t tmpSpace;
	WalkState(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator) :
		brt(graph.n, allocator),
		dataBorrower(allocator),
		bufferBorrower(allocator),
		transfBuff(0, "Temporary transfer buffer"),
		tmpSpace(0)
	{
		allocator.provideMemory(bpts, graph.n);
		for (size_t i(0); i < graph.n; ++ i)
			new (bpts + i) BPT<typename Graph::StorageType>(graph.vertices[i + 1].firstEdge - graph.vertices[i].firstEdge, Selector<Graph>(graph.edges + graph.vertices[i].firstEdge), allocator, dataBorrower, bufferBorrower);
	}
};

template<class Graph, class Container> void walkInternal(Graph &graph, Container &container, WalkState<Graph> &state, Memory::Pointer<typename Graph::StorageType, size_t> &result, size_t origin) {
	container.push(origin);
	size_t pos(0);
	result[pos ++] = origin;
	// Do not return to the origin from whereever we go first
	for (size_t i(graph.vertices[origin].firstEdge); i < graph.vertices[origin + 1].firstEdge; ++ i)
		state.brt.insert(graph.edges[i].vertices[1], origin);
	while (!container.empty()) {
		size_t index(container.top());
		// Extract the already visited neighbors
		typename BRT<typename Graph::StorageType>::Result visited(state.brt.get(index));
		if (state.tmpSpace < visited.size()) {
			state.transfBuff.resize((state.tmpSpace = visited.size()) * sizeof(size_t));
			state.tmpSpace = state.transfBuff.size() / sizeof(size_t);
		}
		Memory::Pointer<typename Graph::StorageType, size_t> visitedA(state.transfBuff);
		for (size_t i(0); i < visited.size(); ++ i)
			visitedA[i] = visited.next();
		// Remove them and get the first unvisited
		size_t next(state.bpts[index].getNext(visited.size(), visitedA));
		if (next >= graph.n) {
			// Nothing returned, this vertex is done
			container.pop();
		} else {
			// Schedule this one and number it
			container.push(next);
			result[pos ++] = next;
			// Notify all neighbors this one was already used
			for (size_t i(graph.vertices[next].firstEdge); i < graph.vertices[next + 1].firstEdge; ++ i)
				if (graph.edges[i].vertices[1] != index) // We already removed this one
					state.brt.insert(graph.edges[i].vertices[1], next);
		}
	}
	// Mark the end
	result[pos] = graph.n + 1;
}

template<class Graph, class Container> Memory::Pointer<typename Graph::StorageType, size_t> walk(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator, Container &container, size_t origin) {
	Memory::Pointer<typename Graph::StorageType, size_t> result(allocator[graph.n]);
	typename Memory::Allocator::Pool<typename Graph::StorageType>::State aState(allocator);
	WalkState<Graph> state(graph, allocator);
	walkInternal(graph, container, state, result, origin);
	allocator.restore(aState);
	return result;
}

}

template<class Graph> Memory::Pointer<typename Graph::StorageType, size_t> dfs(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator, size_t origin = 0) {
	Util::Stack<typename Graph::StorageType, size_t> stack;
	return Internal::walk(graph, allocator, stack, origin);
}

template<class Graph> Memory::Pointer<typename Graph::StorageType, size_t> bfs(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator, size_t origin = 0) {
	Util::Queue<typename Graph::StorageType, size_t> queue(graph.n);
	return Internal::walk(graph, allocator, queue, origin);
}

}

template<class Graph> Memory::Pointer<typename Graph::StorageType, size_t> dfsBRT(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator, size_t origin = 0) {
	return Treeish::dfs(graph, allocator, origin);
}

template<class Graph> Memory::Pointer<typename Graph::StorageType, size_t> bfsBRT(Graph &graph, Memory::Allocator::Pool<typename Graph::StorageType> &allocator, size_t origin = 0) {
	return Treeish::bfs(graph, allocator, origin);
}

}

#endif
