#ifndef MATRIX_CLASSIC_H
#define MATRIX_CLASSIC_H

#include <memory/pointer.h>
#include <memory/mapping.h>
#include <memory/allocator.h>
#include <util/assert.h>

namespace Matrix {

template<class Storage, class T> class Classic {
public:
	typedef Storage StorageType;
	typedef T Type;
	Classic() :
		allocator(NULL)
	{}
	Classic(size_t height, size_t width, Memory::Allocator::Pool<Storage> *allocator_ = NULL) :
		mHeight(height),
		mWidth(width),
		rowSize(width)
	{
		if (!allocator_)
			allocator = allocator_ = new Memory::Allocator::Pool<Storage>();
		else
			allocator = NULL;
		data = Memory::Pointer<Storage, T>((*allocator_)[width * height]);
	}
	Classic(const Classic<Storage, T> &other) :
		data(other.data),
		allocator(NULL),
		mHeight(other.height()),
		mWidth(other.width()),
		rowSize(other.rowSize)
	{ }
	Classic<Storage, T> &operator =(const Classic<Storage, T> &other) {
		data = other.data;
		ASSERT(!allocator);
		allocator = NULL;
		mHeight = other.height();
		mWidth = other.width();
		rowSize = other.rowSize;
		return *this;
	}
	~ Classic() {
		delete allocator;
	}
	size_t height() const { return mHeight; }
	size_t width() const { return mWidth; }
	Classic<Storage, T> submatrix(size_t top, size_t left, size_t height, size_t width) {
		return Classic<Storage, T>(height, width, rowSize, data + left + rowSize * top);
	}
	class Row {
	private:
		Memory::Pointer<Storage, T> mRow;
	public:
		Row(const Memory::Pointer<Storage, T> &row) :
			mRow(row)
		{}
		T &operator[] (size_t x) {
			return mRow[x];
		}
	};
	Row operator [](size_t y) {
		return Row(data + rowSize * y);
	}
	void constant(const T &value) {
		for (size_t i(0); i < height(); ++ i)
			for (size_t j(0); j < width(); ++ j)
				(*this)[i][j] = value;
	}
	void add(Classic<Storage, T> other, const T &mult) {
		ASSERT(other.height() == height() && other.width() == width());
		for (size_t i(0); i < height(); ++ i)
			for (size_t j(0); j < width(); ++ j)
				(*this)[i][j] += mult * other[i][j];
	}
	void copy(Classic<Storage, T> other) {
		ASSERT(other.height() == height() && other.width() == width());
		for (size_t i(0); i < height(); ++ i)
			for (size_t j(0); j < width(); ++ j)
				(*this)[i][j] = other[i][j];
	}
	template<class Function> void iterate(Function function) {
		for(size_t i(0); i < height(); ++ i)
			for (size_t j(0); j < width(); ++ j)
				function((*this)[i][j]);
	}
	void format(const Classic<Storage, T> &) { }
private:
	Classic(size_t height, size_t width, size_t rowSize_, const Memory::Pointer<Storage, T> &data_) :
		data(data_),
		allocator(NULL),
		mHeight(height),
		mWidth(width),
		rowSize(rowSize_)
	{ }
	Memory::Pointer<Storage, T> data;
	Memory::Allocator::Pool<Storage> *allocator;
	size_t mHeight, mWidth;
	size_t rowSize;
};

namespace ClassicInternal {

template<class Matrix, bool add> void classicMultiply(Matrix result, Matrix left, Matrix right, const typename Matrix::Type &zero = typename Matrix::Type(), Memory::Allocator::Pool<typename Matrix::StorageType> *pool = NULL) {
	ASSERT(result.height() == left.height());
	ASSERT(result.width() == right.width());
	ASSERT(left.width() == right.height());
	for (size_t i(0); i < result.height(); ++ i)
		for (size_t j(0); j < result.width(); ++ j) {
			typename Matrix::Type sum(zero);
			for (size_t k(0); k < left.width(); ++ k)
				sum += left[i][k] * right[k][j];
			if (add)
				result[i][j] += sum;
			else
				result[i][j] = sum;
		}
}

template<class Matrix, bool add> void inverseClassicMultiply(Matrix result, Matrix left, Matrix right, const typename Matrix::Type &zero = typename Matrix::Type(), Memory::Allocator::Pool<typename Matrix::StorageType> *pool = NULL) {
	ASSERT(result.height() == left.height());
	ASSERT(result.width() == right.width());
	ASSERT(left.width() == right.height());
	for (size_t j(0); j < result.width(); ++ j)
		for (size_t i(0); i < result.height(); ++ i) {
			typename Matrix::Type sum(zero);
			for (size_t k(0); k < left.width(); ++ k)
				sum += left[i][k] * right[k][j];
			if (add)
				result[i][j] += sum;
			else
				result[i][j] = sum;
		}
}

}

template<class Matrix> void classic(Matrix result, Matrix left, Matrix right, const typename Matrix::Type &zero = typename Matrix::Type(), Memory::Allocator::Pool<typename Matrix::StorageType> *pool = NULL) {
	ClassicInternal::classicMultiply<Matrix, false>(result, left, right, zero);
}

template<class Matrix> void inverse(Matrix result, Matrix left, Matrix right, const typename Matrix::Type &zero = typename Matrix::Type(), Memory::Allocator::Pool<typename Matrix::StorageType> *pool = NULL) {
	ClassicInternal::inverseClassicMultiply<Matrix, false>(result, left, right, zero);
}

}

#endif
