#ifndef MATRIX_DIVIDE_H
#define MATRIX_DIVIDE_H

#include "classic.h"
#include <util/stack.h>
#include <util/assert.h>

namespace Matrix {

template<class Storage, class T> class Z;

namespace Divide {

template<class Matrix> struct Task {
	Matrix destination, left, right;
	Task(const Matrix &destination_, const Matrix &left_, const Matrix &right_) :
		destination(destination_),
		left(left_),
		right(right_)
	{ }
};

template<class Matrix, unsigned limit> void divideMultiply(Matrix destination, Matrix left, Matrix right, const typename Matrix::Type &zero = typename Matrix::Type()) {
	ASSERT(destination.height() == left.height());
	ASSERT(destination.width() == right.width());
	ASSERT(left.width() == right.height());
	destination.format(left);
	destination.constant(zero);
	typedef typename Matrix::StorageType Storage;
	typedef Task<Matrix> Task;
	Util::Stack<Storage, Task> stack;
	stack.push(Task(destination, left, right));
	while (!stack.empty()) {
		Task task(stack.pop());
		size_t width(task.destination.width());
		size_t height(task.destination.height());
		size_t length(task.left.width());
		if (height == 1 && width == 1 && length == 1) {
			task.destination[0][0] += task.left[0][0] * task.right[0][0];
		} else if (height <= limit && width <= limit && length <= limit) {
			ClassicInternal::classicMultiply<Matrix, true>(task.destination, task.left, task.right, zero);
		} else if (width > height && width > length) {
			stack.push(Task(task.destination.submatrix(0, width / 2, height, width - width / 2), task.left, task.right.submatrix(0, width / 2, length, width - width / 2)));
			stack.push(Task(task.destination.submatrix(0, 0, height, width / 2), task.left, task.right.submatrix(0, 0, length, width / 2)));
		} else if (height > length) {
			stack.push(Task(task.destination.submatrix(height / 2, 0, height - height / 2, width), task.left.submatrix(height / 2, 0, height - height / 2, length), task.right));
			stack.push(Task(task.destination.submatrix(0, 0, height / 2, width), task.left.submatrix(0, 0, height / 2, length), task.right));
		} else {
			stack.push(Task(task.destination, task.left.submatrix(0, length / 2, height, length - length / 2), task.right.submatrix(length / 2, 0, length - length / 2, width)));
			stack.push(Task(task.destination, task.left.submatrix(0, 0, height, length / 2), task.right.submatrix(0, 0, length / 2, width)));
		}
	}
}

template<class Matrix, unsigned limit> void divideSquareMultiply(Matrix destination, Matrix left, Matrix right, const typename Matrix::Type &zero = typename Matrix::Type()) {
	ASSERT(destination.height() == destination.width());
	ASSERT(destination.height() == left.height());
	ASSERT(destination.height() == right.height());
	ASSERT(destination.height() == left.width());
	ASSERT(destination.height() == right.width());
	destination.format(left);
	destination.constant(zero);
	typedef typename Matrix::StorageType Storage;
	typedef Task<Matrix> Task;
	Util::Stack<Storage, Task> stack;
	stack.push(Task(destination, left, right));
	while (!stack.empty()) {
		Task task(stack.pop());
		size_t size(task.destination.width());
		if (size == 1) {
			task.destination[0][0] += task.left[0][0] * task.right[0][0];
		} else if (size <= limit) {
			ClassicInternal::classicMultiply<Matrix, true>(task.destination, task.left, task.right);
		} else {
			size_t sth(size / 2);
			size_t sbh(size - size / 2);

			Matrix dtl(task.destination.submatrix(0, 0, sth, sth));
			Matrix dtr(task.destination.submatrix(0, sth, sth, sbh));
			Matrix dbl(task.destination.submatrix(sth, 0, sbh, sth));
			Matrix dbr(task.destination.submatrix(sth, sth, sbh, sbh));

			Matrix ltl(task.left.submatrix(0, 0, sth, sth));
			Matrix ltr(task.left.submatrix(0, sth, sth, sbh));
			Matrix lbl(task.left.submatrix(sth, 0, sbh, sth));
			Matrix lbr(task.left.submatrix(sth, sth, sbh, sbh));

			Matrix rtl(task.right.submatrix(0, 0, sth, sth));
			Matrix rtr(task.right.submatrix(0, sth, sth, sbh));
			Matrix rbl(task.right.submatrix(sth, 0, sbh, sth));
			Matrix rbr(task.right.submatrix(sth, sth, sbh, sbh));

			stack.push(Task(dbr, lbr, rbr));
			stack.push(Task(dbr, lbl, rtr));
			stack.push(Task(dbl, lbr, rbl));
			stack.push(Task(dbl, lbl, rtl));
			stack.push(Task(dtr, ltr, rbr));
			stack.push(Task(dtr, ltl, rtr));
			stack.push(Task(dtl, ltr, rbl));
			stack.push(Task(dtl, ltl, rtl));
		}
	}
}

}

template<class Matrix> void divide(Matrix destination, Matrix left, Matrix right, const typename Matrix::Type &zero = typename Matrix::Type(), Memory::Allocator::Pool<typename Matrix::StorageType> *pool = NULL) {
	Divide::divideMultiply<Matrix, 1>(destination, left, right, zero);
}

template<class Matrix> void divideHybrid(Matrix destination, Matrix left, Matrix right, const typename Matrix::Type &zero = typename Matrix::Type(), Memory::Allocator::Pool<typename Matrix::StorageType> *pool = NULL) {
	Divide::divideMultiply<Matrix, 20>(destination, left, right, zero);
}

template<class Matrix> void square(Matrix destination, Matrix left, Matrix right, const typename Matrix::Type &zero = typename Matrix::Type(), Memory::Allocator::Pool<typename Matrix::StorageType> *pool = NULL) {
	Divide::divideSquareMultiply<Matrix, 1>(destination, left, right, zero);
}

template<class Matrix> void squareHybrid(Matrix destination, Matrix left, Matrix right, const typename Matrix::Type &zero = typename Matrix::Type(), Memory::Allocator::Pool<typename Matrix::StorageType> *pool = NULL) {
	Divide::divideSquareMultiply<Matrix, 20>(destination, left, right, zero);
}

}

#endif
