#ifndef MATRIX_STRASSEN_H
#define MATRIX_STRASSEN_H

#include "classic.h"
#include <util/stack.h>
#include <memory>

namespace Matrix {

namespace Strassen {

template<class Matrix> struct Task {
	enum Kind {
		multiply,
		consolidate
	} kind;
	Matrix l, r, d;
	Matrix m1, m2, m3;
	Matrix m1l, m1r, m2l, m3r, m4r, m5l, m6l, m6r, m7l, m7r;
	typename Memory::Allocator::Pool<typename Matrix::StorageType>::State s;
	Task(const Matrix &left, const Matrix &right, const Matrix &destination) :
		kind(multiply),
		l(left),
		r(right),
		d(destination)
	{ }
	Task(const typename Memory::Allocator::Pool<typename Matrix::StorageType>::State &state,
		const Matrix &m1_, const Matrix &m2_, const Matrix &m3_,
		const Matrix &m1l_, const Matrix &m1r_, const Matrix &m2l_, const Matrix &m3r_, const Matrix &m4r_, const Matrix &m5l_, const Matrix &m6l_, const Matrix &m6r_, const Matrix &m7l_, const Matrix &m7r_,
		const Matrix &destination
		):
		kind(consolidate),
		d(destination),
		m1(m1_),
		m2(m2_),
		m3(m3_),
		m1l(m1l_),
		m1r(m1r_),
		m2l(m2l_),
		m3r(m3r_),
		m4r(m4r_),
		m5l(m5l_),
		m6l(m6l_),
		m6r(m6r_),
		m7l(m7l_),
		m7r(m7r_),
		s(state)
	{ }
};

template<class Matrix, size_t limit> void strassen(Matrix result, Matrix left, Matrix right, const typename Matrix::Type &zero = typename Matrix::Type(), Memory::Allocator::Pool<typename Matrix::StorageType> *pool = NULL) {
	ASSERT(result.height() == result.width());
	ASSERT(result.height() == left.height());
	ASSERT(result.height() == right.height());
	ASSERT(result.height() == left.width());
	ASSERT(result.height() == right.width());
	result.format(left);
	typedef typename Matrix::StorageType Storage;
	std::auto_ptr<Memory::Allocator::Pool<Storage> > localPool;
	if (!pool)
		localPool.reset(pool = new Memory::Allocator::Pool<Storage>);
	typedef Strassen::Task<Matrix> Task;
	Util::Stack<Storage, Task> stack;
	stack.push(Task(left, right, result));
	while (!stack.empty()) {
		Task t(stack.pop());
		switch (t.kind) {
			case Task::multiply: {
				if (t.d.width() == 1) {
					t.d[0][0] = t.l[0][0] * t.r[0][0];
				} else if (t.d.width() <= limit) {
					classic(t.d, t.l, t.r, zero, pool);
				} else {
					// Store current allocator state before creating new matrices
					typename Memory::Allocator::Pool<Storage>::State state(*pool);
					// Define the parts of original ones
					size_t s(t.d.width() / 2);
					Matrix ltl(t.l.submatrix(0, 0, s, s));
					Matrix ltr(t.l.submatrix(0, s, s, s));
					Matrix lbl(t.l.submatrix(s, 0, s, s));
					Matrix lbr(t.l.submatrix(s, s, s, s));
					Matrix rtl(t.r.submatrix(0, 0, s, s));
					Matrix rtr(t.r.submatrix(0, s, s, s));
					Matrix rbl(t.r.submatrix(s, 0, s, s));
					Matrix rbr(t.r.submatrix(s, s, s, s));
					Matrix dtl(t.d.submatrix(0, 0, s, s));
					Matrix dtr(t.d.submatrix(0, s, s, s));
					Matrix dbl(t.d.submatrix(s, 0, s, s));
					Matrix dbr(t.d.submatrix(s, s, s, s));
					// Create temporary source ones
					Matrix m1l(s, s, pool);
					m1l.format(t.l);
					m1l.copy(ltl);
					m1l.add(lbr, 1);
					Matrix m1r(s, s, pool);
					m1r.format(t.r);
					m1r.copy(rtl);
					m1r.add(rbr, 1);
					Matrix m2l(s, s, pool);
					m2l.format(t.l);
					m2l.copy(lbl);
					m2l.add(lbr, 1);
					Matrix m3r(s, s, pool);
					m3r.format(t.r);
					m3r.copy(rtr);
					m3r.add(rbr, -1);
					Matrix m4r(s, s, pool);
					m4r.format(t.r);
					m4r.copy(rbl);
					m4r.add(rtl, -1);
					Matrix m5l(s, s, pool);
					m5l.format(t.l);
					m5l.copy(ltl);
					m5l.add(ltr, 1);
					Matrix m6l(s, s, pool);
					m6l.format(t.l);
					m6l.copy(lbl);
					m6l.add(ltl, -1);
					Matrix m6r(s, s, pool);
					m6r.format(t.r);
					m6r.copy(rtl);
					m6r.add(rtr, 1);
					Matrix m7l(s, s, pool);
					m7l.format(t.l);
					m7l.copy(ltr);
					m7l.add(lbr, -1);
					Matrix m7r(s, s, pool);
					m7r.format(t.r);
					m7r.copy(rbl);
					m7r.add(rbr, 1);
					// Now temporary destination ones (3 are enough, the rest fits into the real destination)
					Matrix m1(s, s, pool);
					m1.format(t.d);
					Matrix m2(s, s, pool);
					m2.format(t.d);
					Matrix m3(s, s, pool);
					m3.format(t.d);
					// Plan what we do now
					// We finish by consolidating them together
					stack.push(Task(state, m1, m2, m3, m1l, m1r, m2l, m3r, m4r, m5l, m6l, m6r, m7l, m7r, t.d));
					// But before that, we need to do some multiplications
					stack.push(Task(m1l, m1r, m1));
					stack.push(Task(m2l, rtl, m2));
					stack.push(Task(ltl, m3r, m3));
					stack.push(Task(lbr, m4r, dbl));
					stack.push(Task(m5l, rbr, dtr));
					stack.push(Task(m6l, m6r, dbr));
					stack.push(Task(m7l, m7r, dtl));
				}
				break;
			}
			case Task::consolidate: {
				size_t s(t.d.width() / 2);
				Matrix dtl(t.d.submatrix(0, 0, s, s));
				Matrix dtr(t.d.submatrix(0, s, s, s));
				Matrix dbl(t.d.submatrix(s, 0, s, s));
				Matrix dbr(t.d.submatrix(s, s, s, s));
				dtl.add(dtr, -1);
				dtl.add(dbl, 1);
				dtl.add(t.m1, 1);
				dbr.add(t.m1, 1);
				dbr.add(t.m2, -1);
				dbr.add(t.m3, 1);
				dtr.add(t.m3, 1);
				dbl.add(t.m2, 1);
				pool->restore(t.s);
				break;
			}
		}
	}
}

}

template<class Matrix> void strassen(Matrix result, Matrix left, Matrix right, const typename Matrix::Type &zero = typename Matrix::Type(), Memory::Allocator::Pool<typename Matrix::StorageType> *pool = NULL) {
	Strassen::strassen<Matrix, 1>(result, left, right, zero, pool);
}

template<class Matrix> void strassenHybrid(Matrix result, Matrix left, Matrix right, const typename Matrix::Type &zero = typename Matrix::Type(), Memory::Allocator::Pool<typename Matrix::StorageType> *pool = NULL) {
	Strassen::strassen<Matrix, 20>(result, left, right, zero, pool);
}

}

#endif
