#ifndef MATRIX_Z_H
#define MATRIX_Z_H

#include <memory/pointer.h>
#include <memory/mapping.h>
#include <memory/allocator.h>
#include <util/assert.h>
#include <util/stack.h>
#include "divide.h"
#include "strassen.h"

namespace Matrix {

template<class Storage, class T> class Classic;

template<class Storage, class T> class Z {
public:
	typedef Storage StorageType;
	typedef T Type;
	Z() :
		allocator(NULL)
	{ }
	Z(size_t height, size_t width, Memory::Allocator::Pool<Storage> *allocator_ = NULL) :
		mHeight(height),
		mWidth(width),
		blocksize(1)
	{
		if (!allocator_)
			allocator = allocator_ = new Memory::Allocator::Pool<Storage>();
		else
			allocator = NULL;
		data = Memory::Pointer<Storage, T>((*allocator_)[width * height]);
	}
	Z(const Z<Storage, T> &other) :
		allocator(NULL),
		data(other.data),
		mHeight(other.height()),
		mWidth(other.width()),
		blocksize(other.blocksize)
	{ }
	Z &operator =(const Z<Storage, T> &other) {
		data = other.data;
		ASSERT(!allocator);
		mHeight = other.height();
		mWidth = other.width();
		blocksize = other.blocksize;
		return *this;
	}
	~ Z() {
		delete allocator;
	}
	Z<Storage, T> submatrix(size_t top, size_t left, size_t height, size_t width) {
		// FIXME: This expects the user uses our „native“ division of the matrix
		size_t offset(0);
		offset += this->width() * top;
		offset += height * left;
		return Z<Storage, T>(height, width, data + offset, blocksize);
	}
	size_t height() const { return mHeight; }
	size_t width() const { return mWidth; }
	class Row {
	private:
		size_t row;
		Z<Storage, T> *owner;
	public:
		T &operator [](size_t column) {
			// Hack, ugly hack and even worse hack…
			return owner->data[row * owner->blocksize + column];
		}
		Row(Z<Storage, T> *owner_, size_t row_) :
			row(row_),
			owner(owner_)
		{ }
	};
	friend class Row;
	Row operator [](size_t row) {
		return Row(this, row);
	}
	void constant(const T& value) {
		for (size_t i(0); i < height() * width(); ++ i)
			data[i] = value;
	}
	template<class Function> void iterate(Function function) {
		for (size_t i(0); i < height() * width(); ++ i)
			function(data[i]);
	}
	// For square and power of two only!
	template<class Other> void imp(Other& m, size_t limit) {
		convert(*this, m, limit, impInternal<Other>);
	}
	template<class Other> void exp(Other& m, size_t limit) {
		convert(*this, m, limit, expInternal<Other>);
	}
	void copy(Z<Storage, T> other) {
		ASSERT(width() == other.width());
		ASSERT(height() == other.height());
		for (size_t i(0); i < width() * height(); ++ i)
			data[i] = other.data[i];
	}
	void add(Z<Storage, T> other, const T &multiply) {
		ASSERT(width() == other.width());
		ASSERT(height() == other.height());
		for (size_t i(0); i < width() * height(); ++ i)
			data[i] += multiply * other.data[i];
	}
	size_t blocksize;
	void format(const Z<Storage, T> &other) {
		blocksize = other.blocksize;
	}
private:
	Z(size_t height, size_t width, Memory::Pointer<Storage, T> data_, size_t blocksize_) :
		allocator(NULL),
		data(data_),
		mHeight(height),
		mWidth(width),
		blocksize(blocksize_)
	{ }
	Memory::Allocator::Pool<Storage> *allocator;
	Memory::Pointer<Storage, T> data;
	size_t mHeight, mWidth;
	template<class Other> struct ConvertTask {
		Z<Storage, T> s;
		Other o;
		ConvertTask(Z<Storage, T> self, Other other) :
			s(self),
			o(other)
		{ }
	};
	template<class Other, class Function> void convert(Z<Storage, T> &self, Other& m, size_t limit, Function function) {
		ASSERT(self.height() == self.width());
		ASSERT(self.height() == m.height());
		ASSERT(self.width() == m.width());
		typedef ConvertTask<Other> Task;
		Util::Stack<Storage, Task> stack;
		stack.push(Task(self, m));
		while (!stack.empty()) {
			Task task(stack.pop());
			if (task.s.height() <= limit) {
				function(task.s, task.o);
				blocksize = task.s.height();
			} else {
				size_t s(task.s.height());
				ASSERT(!(s % 2));
				s /= 2;
				stack.push(Task(task.s.submatrix(s, s, s, s), task.o.submatrix(s, s, s, s)));
				stack.push(Task(task.s.submatrix(s, 0, s, s), task.o.submatrix(s, 0, s, s)));
				stack.push(Task(task.s.submatrix(0, s, s, s), task.o.submatrix(0, s, s, s)));
				stack.push(Task(task.s.submatrix(0, 0, s, s), task.o.submatrix(0, 0, s, s)));
			}
		}
	}
	template<class Other> static void impInternal(Z<Storage, T> &self, Other& other) {
		for (size_t i(0); i < self.height(); ++ i)
			for (size_t j(0); j < self.width(); ++ j)
				self.data[i * self.width() + j] = other[i][j];
	}
	template<class Other> static void expInternal(Z<Storage, T> &self, Other& other) {
		for (size_t i(0); i < self.height(); ++ i)
			for (size_t j(0); j < self.width(); ++ j)
				other[i][j] = self.data[i * self.width() + j];
	}
};

template<class Matrix> void divideZ(Matrix destination, Matrix left, Matrix right, const typename Matrix::Type &zero = typename Matrix::Type(), Memory::Allocator::Pool<typename Matrix::StorageType> *pool = NULL) {
	typename Memory::Allocator::Pool<typename Matrix::StorageType>::State state(*pool);
	Z<typename Matrix::StorageType, typename Matrix::Type> d(destination.height(), destination.width(), pool);
	Z<typename Matrix::StorageType, typename Matrix::Type> l(left.height(), left.width(), pool);
	Z<typename Matrix::StorageType, typename Matrix::Type> r(right.height(), right.width(), pool);
	l.imp(left, 1);
	r.imp(right, 1);
	square(d, l, r, zero, pool);
	d.exp(destination, 1);
	pool->restore(state);
}

template<class Matrix> void divideZHybrid(Matrix destination, Matrix left, Matrix right, const typename Matrix::Type &zero = typename Matrix::Type(), Memory::Allocator::Pool<typename Matrix::StorageType> *pool = NULL) {
	typename Memory::Allocator::Pool<typename Matrix::StorageType>::State state(*pool);
	Z<typename Matrix::StorageType, typename Matrix::Type> d(destination.height(), destination.width(), pool);
	Z<typename Matrix::StorageType, typename Matrix::Type> l(left.height(), left.width(), pool);
	Z<typename Matrix::StorageType, typename Matrix::Type> r(right.height(), right.width(), pool);
	l.imp(left, 20);
	r.imp(right, 20);
	squareHybrid(d, l, r, zero, pool);
	d.exp(destination, 20);
	pool->restore(state);
}

template<class Matrix> void strassenZ(Matrix destination, Matrix left, Matrix right, const typename Matrix::Type &zero = typename Matrix::Type(), Memory::Allocator::Pool<typename Matrix::StorageType> *pool = NULL) {
	typename Memory::Allocator::Pool<typename Matrix::StorageType>::State state(*pool);
	Z<typename Matrix::StorageType, typename Matrix::Type> d(destination.height(), destination.width(), pool);
	Z<typename Matrix::StorageType, typename Matrix::Type> l(left.height(), left.width(), pool);
	Z<typename Matrix::StorageType, typename Matrix::Type> r(right.height(), right.width(), pool);
	l.imp(left, 1);
	r.imp(right, 1);
	strassen(d, l, r, zero, pool);
	d.exp(destination, 1);
	pool->restore(state);
}

template<class Matrix> void strassenZHybrid(Matrix destination, Matrix left, Matrix right, const typename Matrix::Type &zero = typename Matrix::Type(), Memory::Allocator::Pool<typename Matrix::StorageType> *pool = NULL) {
	typename Memory::Allocator::Pool<typename Matrix::StorageType>::State state(*pool);
	Z<typename Matrix::StorageType, typename Matrix::Type> d(destination.height(), destination.width(), pool);
	Z<typename Matrix::StorageType, typename Matrix::Type> l(left.height(), left.width(), pool);
	Z<typename Matrix::StorageType, typename Matrix::Type> r(right.height(), right.width(), pool);
	l.imp(left, 20);
	r.imp(right, 20);
	strassenHybrid(d, l, r, zero, pool);
	d.exp(destination, 20);
	pool->restore(state);
}

}

#endif
