#ifndef MEMORY_ALLOCATOR_H
#define MEMORY_ALLOCATOR_H

#include "memory.h"
#include "mapping.h"
#include "pointer.h"

#include <list>
#include <vector>

namespace Memory {
namespace Allocator {

template<class Storage> class Pool {
private:
	Pool(const Pool &other);
	Pool &operator =(const Pool &other);
	static const size_t align = sizeof(void *);
	template<bool, class T> class initArray;
	template<class T> class initArray<true, T> {
	public:
		initArray(Pointer<Storage, T> p, size_t count) {
			for (size_t i(0); i < count; ++ i)
				new (p + i) T;
		}
	};
	template<class T> class initArray<false, T> {
	public:
		initArray(Pointer<Storage, T>, size_t) { }
	};
public:
	Pool() :
		position(0),
		mappingPos(mappings.begin())
	{ }
	~ Pool() {
		for (typename std::list<Mapping<Storage> *>::iterator i(mappings.begin()); i != mappings.end(); ++ i)
			delete *i;
	}
	class State {
			friend class Pool<Storage>;
			size_t position;
			typename std::list<Mapping<Storage> *>::iterator mappingPos;
			bool start;
		public:
			State(const Pool& pool) :
				position(pool.position),
				mappingPos(pool.mappingPos),
				start(mappingPos == pool.mappings.begin())
			{ }
			State() {}
	};
	template<class T, bool init> Pointer<Storage, T> allocate(size_t count = 1) {
		bool move(true);
		if (count == 0) {
			count = 1;
			move = false;
		}
		size_t reqSize(count * sizeof(T));
		size_t curPos(position);
		// Align it nicely
		curPos = ((curPos - 1) / align + 1) * align;
		bool erase(false);
		// Try to find large enough mapping
		while (mappingPos != mappings.end() && curPos + reqSize > (*mappingPos)->size()) {
			if (erase) {
				// Delete small ones
				mappingPos = mappings.erase(mappingPos);
			} else {
				++ mappingPos;
				erase = true;
			}
			curPos = 0;
		}
		// Did it work? Does mappPos show into valid one?
		if (mappingPos == mappings.end()) {
			size_t mapSize(mappings.empty() ? Storage::blocksize() : mappings.back()->size() * 2);
			if (mapSize < reqSize)
				mapSize = ((reqSize - 1) / Storage::blocksize() + 1) * Storage::blocksize();
			mappingPos = mappings.insert(mappingPos, new Mapping<Storage>(mapSize, "pool allocator"));
		}
		Pointer<Storage, unsigned char> bare(**mappingPos);
		bare += curPos;
		if (move)
			position = curPos + reqSize;
		Pointer<Storage, T> result(Pointer<Storage, T>::casted(bare));
		initArray<init, T>(result, count);
		return result;
	}
	template<class T> operator Pointer<Storage, T>() {
		return allocate<T, true>();
	}
	template<class T> void provideMemory(Pointer<Storage, T> &p, size_t count) {
		p = allocate<T, false>(count);
	}
	// To allow allocation by pointer p(pool[size])
	class Count {
			friend class Pool<Storage>;
			Pool<Storage> &pool;
			size_t count;
			Count(Pool<Storage> &pool_, size_t count_) :
				pool(pool_),
				count(count_)
			{ }
		public:
			template<class T> operator Pointer<Storage, T>() {
				return pool.allocate<T, true>(count);
			}
	};
	Count operator[](size_t size) {
		return Count(*this, size);
	};
	void restore(const State &state) {
		position = state.position;
		mappingPos = state.mappingPos;
		if (state.start)
			mappingPos = mappings.begin();
	}
private:
	size_t position;
	std::list<Mapping<Storage> *> mappings;
	typename std::list<Mapping<Storage> *>::iterator mappingPos;
};

template<class Storage, class Type> class ArrayBorrower {
private:
	ArrayBorrower(const ArrayBorrower<Storage, Type> &);
	ArrayBorrower<Storage, Type> &operator =(const ArrayBorrower<Storage, Type> &);
	Pool<Storage> &source;
	size_t bucket(size_t what) {
		size_t result(0);
		size_t value(1);
		while (value < what) {
			value <<= 1;
			result ++;
		}
		return result;
	}
	std::vector<std::vector<Pointer<Storage, Type> > > cache;
public:
	ArrayBorrower(Pool<Storage> &source_) :
		source(source_)
	{ }
	Pointer<Storage, Type> get(size_t size) {
		size_t b(bucket(size));
		while (b >= cache.size())
			cache.push_back(std::vector<Pointer<Storage, Type> >());
		if (cache[b].size()) {
			Pointer<Storage, Type> result(cache[b].back());
			for (size_t i(0); i < size; ++ i)
				new (result + i) Type;
			cache[b].pop_back();
			return result;
		} else {
			return Pointer<Storage, Type>(source[1 << b]);
		}
	}
	void put(Pointer<Storage, Type> pointer, size_t size) {
		size_t b(bucket(size));
		for (size_t i(0); i < size; ++ i)
			pointer[i].~Type();
		cache[b].push_back(pointer);
	}
};

}
}

#endif
