#include "cache.h"
#include <util/assert.h>

#include <signal.h>
#include <cstring>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <cerrno>
#include <thread/lock.h>
#include <pthread.h>
#include <list>
#include <algorithm>
#ifdef DEBUG_PAGESWAP
#include <cstdio>
#endif

using std::vector;
using std::list;
using std::min;

namespace Memory {

namespace {

namespace AIO {

Thread::Lock lock[2];

struct Task {
	virtual void operator ()() = 0;
	virtual ~ Task() {}
};

vector<std::pair<Task *, bool> > tasks[2];
vector<Task *> done[2];

bool working[2] = { false, false };

pthread_cond_t cond[2] = { PTHREAD_COND_INITIALIZER, PTHREAD_COND_INITIALIZER };
pthread_cond_t finished[2] = { PTHREAD_COND_INITIALIZER, PTHREAD_COND_INITIALIZER };

void *thread(void *num) {
	size_t tNum(*(size_t *) num);
	delete (size_t *) num;
	Thread::Scope l(lock[tNum]);
	for (;;) {
		if (tasks[tNum].empty()) {
			int result(pthread_cond_wait(&cond[tNum], lock[tNum].osMutex()));
			ASSERT_EXPL(!result, "Could not wait on condition");
		} else {
			lock[tNum].unlock();
			(*tasks[tNum][0].first)();
			lock[tNum].lock();
			if (tasks[tNum][0].second) {
				done[tNum].push_back(tasks[tNum][0].first);
				int result(pthread_cond_broadcast(&finished[tNum]));
				ASSERT_EXPL(!result, "Could not signal finished");
			} else {
				delete tasks[tNum][0].first;
			}
			tasks[tNum].erase(tasks[tNum].begin());
		}
	}
}

bool threadStarted[2] = { false, false };

void submit(Task *task, bool waitable, size_t line = 0) {
	Thread::Scope l(lock[line]);
	tasks[line].push_back(std::pair<Task *, bool>(task, waitable));
	if (!threadStarted[line]) {
		threadStarted[line] = true;
		pthread_t tid;
		size_t *linePtr = new size_t;
		*linePtr = line;
		int result(pthread_create(&tid, NULL, thread, linePtr));
		ASSERT_EXPL(!result, "Thread didn't start");
	} else {
		int result(pthread_cond_signal(&cond[line]));
		ASSERT_EXPL(!result, "Couldn't signal thread");
	}
}

void wait(Task *task, size_t line = 0) {
	Thread::Scope l(lock[line]);
	for (;;) {
		for (vector<Task *>::iterator it(done[line].begin()); it != done[line].end(); ++ it) {
			if (*it == task) {
				delete task;
				done[line].erase(it);
				return;
			}
		}
		int result(pthread_cond_wait(&finished[line], lock[line].osMutex()));
		ASSERT_EXPL(!result, "Couldn't wait");
	}
}

}

}

namespace Cache {

namespace {

class SigMuxer {
public:
	static SigMuxer *get() {
		if (!singleton)
			singleton = new SigMuxer;
		return singleton;
	}
private:
	static SigMuxer *singleton;
	SigMuxer() {
		struct sigaction action;
		memset(&action, 0, sizeof action);
		action.sa_sigaction = &swapPages;
		action.sa_flags = SA_RESTART | SA_SIGINFO;
		int result(sigaction(SIGSEGV, &action, NULL));
		ASSERT(result != -1);
	}
	static void swapPages(int signum, siginfo_t *info, void *ucontext) {
		get()->swapPages(info->si_addr);
	}
	struct Block {
		void *baseAddress;
		void *endAddress;
		PageLoader *loader;
		void *note;
	};
	vector<Block> blocks;
	void swapPages(void *address) {
#ifdef DEBUG_PAGESWAP
		fprintf(DEBUG_PAGESWAP, "DEBUG: Looking for a page for address %p\n", address);
#endif
		for (vector<Block>::iterator it(blocks.begin()); it != blocks.end(); ++ it) {
			if (it->baseAddress <= address && address < it->endAddress && it->loader->loadPage(address, it->note))
				return;
		}
		ASSERT_EXPL(0, "No page loader claims to be able to load the page");
	}
public:
	void insert(void *base, void *end, PageLoader *loader, void *note) {
#ifdef DEBUG_PAGESWAP
		fprintf(DEBUG_PAGESWAP, "DEBUG: Adding pages %p to %p, %zd blocks before adding, tag %p\n", base, end, blocks.size(), note);
#endif
		Block block;
		block.baseAddress = base;
		block.endAddress = end;
		block.loader = loader;
		block.note = note;
		blocks.push_back(block);
	}
	void remove(void *base, void *end) {
#ifdef DEBUG_PAGESWAP
		fprintf(DEBUG_PAGESWAP, "DEBUG: Removing pages %p to %p from list of %zd\n", base, end, blocks.size());
#endif
		for (vector<Block>::iterator it(blocks.begin()); it != blocks.end(); ++ it)
			if (it->baseAddress == base && it->endAddress == end) {
				blocks.erase(it);
				return;
			}
		ASSERT_EXPL(0, "DEBUG: No block found");
	}
};

SigMuxer *SigMuxer::singleton(NULL);

}

struct LRU::BlockIdInternal {
	void *base, *end;
	size_t loaded;
	int fd;
	PagePtr *pages;
};

struct LRU::Page {
	BlockIdInternal *id;
	unsigned char *basePtr;
	list<PagePtr>::iterator iter;
};

struct LRU::AccessList {
	list<PagePtr> pages;
};

LRU::LRU(size_t pageSize_, size_t pageCount_) :
	accessCount(0),
	freePages(pageCount_),
	pageSize(pageSize_),
	pageCount(pageCount_),
	accessList(new AccessList)
{ }

LRU::~ LRU() {
	delete accessList;
}

LRU::BlockId LRU::putMapping(void *pointer, size_t size, int fd) {
	ASSERT(size % pageSize == 0);
	BlockIdInternal *id(new BlockIdInternal);
	id->pages = new PagePtr[size / pageSize];
#ifdef DEBUG_PAGESWAP
	fprintf(DEBUG_PAGESWAP, "DEBUG: Put mapping, size %zu, pages %zu\n", size, size / pageSize);
#endif
	memset(id->pages, 0, size / pageSize * sizeof *id->pages);
	id->base = pointer;
	id->end = (void *)((unsigned char *) pointer + size);
	id->loaded = 0;
	id->fd = fd;
	SigMuxer::get()->insert(id->base, id->end, this, id);
	return BlockId(id);
}

LRU::BlockId LRU::migrateMapping(void *oldPointer, void *pointer, size_t oldSize, size_t size, int fd, const BlockId &id) {
	ASSERT(size % pageSize == 0);
	processAccesses(true);
	BlockIdInternal *idi(id.id);
	ASSERT(idi);
	ASSERT(oldPointer == idi->base);
	PagePtr *newPages = new PagePtr[size / pageSize];
#ifdef DEBUG_PAGESWAP
	fprintf(DEBUG_PAGESWAP, "DEBUG: Migrate mapping, oldsize %zu (%zu), new size %zu (%zu)\n", oldSize, oldSize / pageSize, size, size / pageSize);
#endif
	for (size_t i(0); i < min(size, oldSize) / pageSize; ++ i) {
		newPages[i] = idi->pages[i];
		if (newPages[i])
			newPages[i]->basePtr += (unsigned char *) pointer - (unsigned char *) oldPointer;
	}
	for (size_t i(size / pageSize); i < oldSize / pageSize; ++ i)
		if (idi->pages[i]) {
			idi->loaded --;
			freePages ++;
			accessList->pages.erase(idi->pages[i]->iter);
			delete idi->pages[i];
		}
	delete[] idi->pages;
	for (size_t i(oldSize / pageSize); i < size / pageSize; ++ i)
		newPages[i] = NULL;
	idi->pages = newPages;
	SigMuxer::get()->remove(idi->base, idi->end);
	idi->base = pointer;
	idi->end = (unsigned char *) pointer + size;
	SigMuxer::get()->insert(idi->base, idi->end, this, idi);
	return id;
}

void LRU::removeMapping(void *pointer, size_t size, int fd, const BlockId &id) {
	struct Cork : public AIO::Task {
		virtual void operator ()() {}
	};
	// Make sure all tasks related to this mapping are done
	Cork *cork(new Cork());
	AIO::submit(cork, true);
	BlockIdInternal *idi(id.id);
	ASSERT(idi);
	processAccesses(false);
	freePages += idi->loaded;
	SigMuxer::get()->remove(idi->base, idi->end);
	for (size_t i(0); i < size / pageSize; ++ i)
		if (idi->pages[i]) {
			accessList->pages.erase(idi->pages[i]->iter);
			delete idi->pages[i];
		}
	delete[] idi->pages;
	// Make sure all previous requests are processed
	AIO::wait(cork);
	// Close the FD (we take over the ownership)
	int result(close(idi->fd));
	ASSERT_EXPL(result != -1, "Could not close");
	delete idi;
}

void LRU::processAccesses(bool needFull) {
	// Nothing to do from last time
	if (!accessCount)
		return;
	// TODO: Some on-background processing in future
	for (size_t i(0); i < accessCount; ++ i) {
		BlockIdInternal *idi(accesses[i].id.id);
		size_t index(accesses[i].ptr - (unsigned char *) idi->base);
		index /= pageSize;
		PagePtr page(idi->pages[index]);
		if (page) {
			accessList->pages.erase(page->iter);
		} else {
			page = new Page;
			page->id = idi;
			page->basePtr = accesses[i].ptr;
			idi->pages[index] = page;
		}
		accessList->pages.push_front(page);
		page->iter = accessList->pages.begin();
	}
	accessCount = 0;
}

size_t LRU::listPages(BlockId id, size_t *buffer) {
	BlockIdInternal *idi(id.id);
	ASSERT(idi);
	if (!idi->loaded)
		return 0;
	processAccesses(true);
	size_t pageCount((unsigned char *) idi->end - (unsigned char *) idi->base);
	pageCount /= pageSize;
	size_t pos(0);
	for (size_t i(0); i < pageCount; ++ i)
		if (idi->pages[i])
			buffer[pos ++] = i;
	ASSERT(pos);
	return pos;
}

bool LRU::loadPage(void *address, void *note) {
#ifdef DEBUG_PAGESWAP
	fprintf(DEBUG_PAGESWAP, "DEBUG: The page belongs to LRU (page size = %zd, page count = %zd, free page count = %zd)\n", pageSize, pageCount, freePages);
#endif
	// that uses real pointers
	BlockIdInternal *idi((BlockIdInternal *) note);
	ASSERT(idi);
	size_t index((unsigned char *) address - (unsigned char *) NULL);
	index /= pageSize;
	address =((unsigned char *) NULL + index * pageSize);
	// Make sure it is touched in case it is some kind of memset, qsort or other
	access((unsigned char *) address, BlockId(idi), false);
#ifdef DEBUG_PAGESWAP
	fprintf(DEBUG_PAGESWAP, "DEBUG: Loading page %p, have %zd in list\n", address, accessList->pages.size());
#endif
	size_t retries(3);
	int ecode;
	do {
		// Now enable access to the page and load data into it
		ecode = mprotect(address, pageSize, PROT_READ | PROT_WRITE);
		if (ecode == -1) {
			usleep(100000);
		}
	// We are pretty sure this page is ours, so it must be kernel is low on its own resources.
	// Try again a small time later, maybe it managed to release something.
	} while (-- retries && ecode == -1 && errno == ENOMEM);
	ASSERT_EXPL(ecode != -1, "mprotect failed");
	struct ReadTask : public AIO::Task {
		virtual void operator ()() {
			off64_t seekResult(lseek64(fd, seekPos, SEEK_SET));
			ASSERT(seekPos == seekResult);
			size_t pos(0);
			ssize_t result;
			while (pos < pageSize && (result = read(fd, address + pos, pageSize - pos)) != -1 && result)
				pos += result;
			ASSERT_EXPL(result != -1, "DEBUG: Could not read");
		}
		unsigned char *address;
		size_t seekPos, pageSize;
		int fd;
	};
	ReadTask *rt(new ReadTask);
	rt->seekPos = (unsigned char *) address - (unsigned char *) idi->base;
	rt->fd = idi->fd;
	rt->pageSize = pageSize;
	rt->address = (unsigned char *) address;
	AIO::submit(rt, true);
	// We're full, need to evict a page
	if (!freePages) {
		void *evictAddress(evictPage());
		ASSERT(evictAddress != address);
	}
	idi->loaded ++;
	freePages --;
	AIO::wait(rt);
	return true;
}

void LRU::resize(size_t pageCount) {
	if (pageCount == this->pageCount)
		return;
	if (pageCount > this->pageCount)
		// Just let it populate
		freePages += pageCount - this->pageCount;
	else {
		size_t diff(this->pageCount - pageCount);
		if (diff < freePages)
			freePages -= diff;
		else {
			diff -= freePages;
			for (size_t i(0); i < diff; ++ i)
				evictPage();
			freePages = 0;
		}
	}
	this->pageCount = pageCount;
}

void *LRU::evictPage() {
	// Decide which one to evict
	processAccesses(true);
	PagePtr evict(accessList->pages.back());
	accessList->pages.pop_back();
	void *evictAddress(evict->basePtr);
	// Find the block it resides in
	BlockIdInternal *block(evict->id);
	ASSERT(block);
	size_t blockIndex(evict->basePtr - (unsigned char *) block->base);
	blockIndex /= pageSize;
	block->pages[blockIndex] = NULL;
	delete evict;
#ifdef DEBUG_PAGESWAP
	fprintf(DEBUG_PAGESWAP, "DEBUG: Evicting page at address %p\n", evictAddress);
#endif
	struct WriteTask : public AIO::Task {
		virtual void operator ()() {
			off64_t seekResult(lseek64(fd, seekPos, SEEK_SET));
			ASSERT(seekPos == seekResult);
			size_t pos(0);
			ssize_t result;
			while (pos < pageSize && (result = write(fd, data + pos, pageSize - pos)) != -1)
				pos += result;
			ASSERT_EXPL(result != -1, "Could not write");
			result = munmap(data, pageSize);
			ASSERT_EXPL(result != -1, "Could not munmap temporary buffer");
		}
		unsigned char *data;
		size_t seekPos, pageSize;
		int fd;
	};
	WriteTask *t(new WriteTask);
	t->data = (unsigned char *) mmap(NULL, pageSize, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
	ASSERT(t->data != MAP_FAILED);
	memcpy(t->data, evictAddress, pageSize);
	t->seekPos = (unsigned char *) evictAddress - (unsigned char *) block->base;
	t->pageSize = pageSize;
	t->fd = block->fd;
	// Write the data to disk and evict (unmap the page and map it again empty)
	int ecode(munmap(evictAddress, pageSize));
	ASSERT_EXPL(ecode != -1, "munmap failed");
	void *mmapResult(mmap(evictAddress, pageSize, PROT_NONE, MAP_ANONYMOUS | MAP_FIXED | MAP_PRIVATE, -1, 0));
	ASSERT(mmapResult != MAP_FAILED);
	freePages ++;
	block->loaded --;
	AIO::submit(t, false);
	return evictAddress;
}

}

int Logger::fd(-1);
size_t Logger::accessCount[2] = { 0, 0 };
Logger::Access Logger::accesses[2][Logger::maxAccesses];
size_t Logger::current;
AIO::Task *Logger::last(NULL);

void Logger::flush(bool block) {
	if (fd != -1) {
		// Write the data we have gathered
		struct Flush : public AIO::Task {
			virtual void operator ()() {
				size_t pos(0);
				ssize_t result;
				size_t size(count * sizeof(Access));
				while (pos < size && (result = write(fd, data + pos, size - pos)) != -1)
					pos += result;
				ASSERT_EXPL(result != -1, "Could not write");
			}
			unsigned char *data;
			size_t count;
			int fd;
		};
		Flush *flush(new Flush);
		flush->count = accessCount[current];
		flush->data = (unsigned char *) &accesses[current][0];
		flush->fd = fd;
		AIO::submit(flush, true, 1);
		if (last)
			AIO::wait(last, 1);
		if (block) {
			AIO::wait(flush, 1);
			last = NULL;
		} else
			last = flush;
	}
	accessCount[current = !current] = 0;
}

void Logger::setOutput(const char *filename) {
	flush(true);
	if (fd != -1) {
		int result(close(fd));
		ASSERT_EXPL(fd != -1, "Couldn't close");
	}
	fd = open(filename, O_WRONLY | O_CREAT | O_LARGEFILE, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
}

void Logger::setOutput(int fd_) {
	flush(true);
	if (fd != -1) {
		int result(close(fd));
		ASSERT_EXPL(fd != -1, "Couldn't close");
	}
	fd = fd_;
}

}
