#ifndef MEMORY_CACHE_H
#define MEMORY_CACHE_H

#include <cstdlib>
#include <vector>

namespace Memory {

namespace Cache {

class PageLoader {
public:
	virtual bool loadPage(void *address, void *note) = 0;
};

/*
 * FIXME: Currently we don't do any kind of thread support. We might want to do that in future.
 * In that case, we need to solve:
 * • Locking of data structures (that's the easy part)
 * • Make sure that when the page is readable/writable, the data are already there. Maybe some remapping or so.
 * • Make sure that we don't load a page twice if it is accessed by multiple threads.
 * • Make sure the signal handling works in all threads.
 */
class LRU : public PageLoader {
private:
	class BlockIdInternal;
public:
	class BlockId {
		friend class LRU;
		BlockIdInternal *id;
		BlockId(BlockIdInternal *id_) : id(id_) {}
	public:
		BlockId() : id(NULL) {}
	};
private:
	class Page;
	typedef Page *PagePtr;
	class AccessList;
	static const size_t bufferSize = 4096;
	struct Access {
		unsigned char *ptr;
		BlockId id;
	};
	Access accesses[4096];
	size_t accessCount;
	size_t freePages;
	size_t pageSize, pageCount;
	void processAccesses(bool needFull);
	void *evictPage();
	LRU(const LRU &other);
	LRU &operator =(const LRU &other);
	AccessList *accessList;
public:
	LRU(size_t pageSize, size_t pageCount);
	~ LRU();
	BlockId putMapping(void *pointer, size_t size, int fd);
	BlockId migrateMapping(void *oldPointer, void *pointer, size_t oldSize, size_t size, int fd, const BlockId &id);
	void removeMapping(void *pointer, size_t size, int fd, const BlockId &id);
	void access(unsigned char *ptr, const BlockId &id, bool doAccess = true) {
		if (accessCount == bufferSize)
			processAccesses(false);
		accesses[accessCount].ptr = ptr;
		accesses[accessCount ++].id = id;
		if (doAccess)
			last = *ptr;
	}
	virtual bool loadPage(void *address, void *note);
	size_t listPages(BlockId id, size_t *buffer);
	// Hack.
	volatile unsigned char last;
	void resize(size_t pageCount);
};

}

// pageSize MUST be multiple of OS level page size, or it won't work
template<size_t pageSize, size_t pageCount, class Policy = Cache::LRU> class DiskCache {
private:
	static Policy *policySingleton;
	static Policy &policy() {
		if (!policySingleton) {
			policySingleton = new Policy(pageSize, pageCount);
		}
		return *policySingleton;
	};
public:
	typedef typename Policy::BlockId Id;
	static size_t blocksize() { return pageSize; }
	static Id putMapping(void *pointer, size_t size, int fd) { return policy().putMapping(pointer, size, fd); }
	static Id migrateMapping(void *oldPointer, void *pointer, size_t oldSize, size_t size, int fd, const Id &id) { return policy().migrateMapping(oldPointer, pointer, oldSize, size, fd, id); }
	static void removeMapping(void *pointer, size_t size, int fd, const Id &id) { policy().removeMapping(pointer, size, fd, id); }
	static void touch(void *p, size_t size, const Id &id) {
		// Convert to size_t the allowed way
		unsigned char *up((unsigned char *) p);
		size_t start(up - (unsigned char *) NULL);
		size_t end(start + size - 1);
		// Compute indices of the pages
		start /= pageSize;
		end /= pageSize;
		for (size_t i(start); i <= end; ++ i)
			policy().access((unsigned char *) NULL + i * pageSize, id);
	}
	static size_t maxPageCount() { return pageCount; }
	static size_t listPages(const Id &id, size_t *buffer) {
		return policy().listPages(id, buffer);
	}
	static size_t resize(size_t pageCount_) {
		policy().resize(pageCount_);
	}
	static size_t restoreSize() {
		policy().resize(pageSize);
	}
};

template<size_t pageSize, size_t pageCount, class Policy> Policy *DiskCache<pageSize, pageCount, Policy>::policySingleton(NULL);

namespace {
namespace AIO {
class Task;
};
};

class Logger {
private:
	static int fd;
	const static size_t maxAccesses = 4096*64;
	static size_t accessCount[2];
	struct Access {
		void *access;
		size_t size;
		bool auxiliary;
	};
	static Access accesses[2][maxAccesses];
	static size_t current;
	static AIO::Task *last;
public:
	static void setOutput(const char *filename);
	static void setOutput(int fd);
	static void flush(bool blocking = true);
	static void log(void *p, size_t size, bool auxiliary) {
		accesses[current][accessCount[current]].access = p;
		accesses[current][accessCount[current]].auxiliary = auxiliary;
		accesses[current][accessCount[current] ++].size = size;
		if (accessCount[current] == maxAccesses)
			flush(false);
	}
};

template<class BaseType, class Type> struct IsAuxiliary { static bool isAuxiliary() { return true; } };
template<class BaseType> struct IsAuxiliary<BaseType, BaseType> { static bool isAuxiliary() { return false; } };

template<size_t pageSize, size_t pageCount, class BaseType, size_t logInaccuracy = 1, class Policy = Cache::LRU> class LogCache : public DiskCache<pageSize, pageCount, Policy> {
private:
	typedef DiskCache<pageSize, pageCount, Policy> Parent;
public:
	template<class Target> static void touch(Target *p, size_t size, const typename Parent::Id &id) {
		Parent::touch(p, size, id);
		unsigned char *pointer((unsigned char *) p);
		size_t index = pointer - (unsigned char *) NULL;
		pointer -= index % logInaccuracy;
		Logger::log(pointer, size, IsAuxiliary<BaseType, Target>::isAuxiliary());
	}
};

}

#endif
