#include "mapping.h"
#include <thread/lock.h>
#include <util/assert.h>

#include <sys/mman.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <cstring>
#include <cstdio>

using namespace Thread;

namespace Memory {

using namespace Impl;

namespace {

Lock lock;

const char *mapDir() {
	Scope sl(lock);
	static const char *dir(NULL);
	if (!dir)
		if (!(dir = getenv("MAP_DIR")))
			dir = get_current_dir_name();
	return dir;
}

}

Mapping<Core>::Mapping(size_t size, const char *name_) :
	name(name_),
	length(wholeBlocks(size))
{
#ifdef DEBUG_ALLOC
	fprintf(DEBUG_ALLOC, "DEBUG: Allocate %zd for %s\n", length, name);
#endif
#ifdef VALGRIND_SUPPORT
	data = malloc(length);
	ASSERT(data);
#else
	data = mmap(NULL, length, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
	ASSERT(data != MAP_FAILED);
#endif
}

Mapping<Core>::~ Mapping() {
#ifdef DEBUG_ALLOC
	fprintf(DEBUG_ALLOC, "DEBUG: Deallocate %zd from %s\n", length, name);
#endif
#ifdef VALGRIND_SUPPORT
	free(data);
#else
	int result(munmap(data, length));
	ASSERT_EXPL(!result, "munmap failed");
#endif
}

void Mapping<Core>::resize(size_t size) {
	size = wholeBlocks(size);
	if (size == length) {
		return;
	}
#ifdef DEBUG_ALLOC
	fprintf(DEBUG_ALLOC, "DEBUG: Resize from %zd to %zd on %s\n", length, size, name);
#endif
#ifdef VALGRIND_SUPPORT
	data = realloc(data, size);
	ASSERT(data);
#else
	data = mremap(data, length, size, MREMAP_MAYMOVE);
	ASSERT_EXPL(data != MAP_FAILED, "mremap failed");
#endif
	length = size;
}

namespace Impl {

size_t wholeBlocks(size_t size, size_t blocksize) {
	if (size == 0)
		return blocksize;
	return blocksize * (1 + (size - 1) / blocksize);
}

BaseDiskMapping::BaseDiskMapping(size_t size, const char *name_, size_t blocksize) :
	name(name_),
	length(wholeBlocks(size, blocksize))
{
	// Get the file to hold it
	const char *dirname = mapDir();
	{
		Scope sl(lock);
		static size_t number(0);
		char filename[strlen(dirname) + 1024];
		sprintf(filename, "%s/map-%zd", dirname, number ++);
		fd = open(filename, O_RDWR | O_CREAT | O_DIRECT | O_EXCL | O_LARGEFILE, S_IRUSR | S_IWUSR);
#ifdef DEBUG_ALLOC
		fprintf(DEBUG_ALLOC, "DEBUG: Allocate %zd for %s at file %s/%i\n", length, name, filename, fd);
#endif
		int result(unlink(filename));
		ASSERT_EXPL(result != -1, "unlink failed");
	}
	ASSERT(fd != -1);
	data = mmap(NULL, length + blocksize, PROT_NONE, MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
	ASSERT(data != MAP_FAILED);
	origPtr = data;
	data = (unsigned char *) NULL + wholeBlocks((unsigned char *) data - (unsigned char *) NULL, blocksize);
	origLength = length + blocksize;
}

BaseDiskMapping::~ BaseDiskMapping() {
#ifdef DEBUG_ALLOC
	fprintf(DEBUG_ALLOC, "DEBUG: Deallocate %zd from %s, closed in the cache (%i)\n", length, name, fd);
#endif
	int result(munmap(origPtr, origLength));
	ASSERT_EXPL(!result, "munmap failed");
}

void BaseDiskMapping::resize(size_t size, size_t blocksize, size_t *usedPages, size_t usedPageCount) {
	size = wholeBlocks(size, blocksize);
#ifdef DEBUG_ALLOC
	fprintf(DEBUG_ALLOC, "DEBUG: Resize from %zd to %zd on %s\n", length, size, name);
#endif
	// mremap doesn't seem to work, so migrate to new place manually
	void *newData(mmap(NULL, size + blocksize, PROT_NONE, MAP_PRIVATE | MAP_ANONYMOUS, -1, 0));
	ASSERT(newData != MAP_FAILED);
	void *newOrigPtr(newData);
	newData = (unsigned char *) NULL + wholeBlocks((unsigned char *) newData - (unsigned char *) NULL, blocksize);
	for (size_t i(0); i < usedPageCount; ++ i) {
		int result(mprotect((unsigned char *) newData + usedPages[i] * blocksize, blocksize, PROT_READ | PROT_WRITE));
		ASSERT_EXPL(result != -1, "Migration mprotect error");
		memcpy((unsigned char *) newData + usedPages[i] * blocksize, (unsigned char *) data + usedPages[i] * blocksize, blocksize);
	}
	int result(munmap(origPtr, origLength));
	ASSERT_EXPL(result != -1, "munmap failed");
	length = size;
	origLength = length + blocksize;
	data = newData;
	origPtr = newOrigPtr;

}

}

};
