#ifndef MAPPING_H
#define MAPPING_H

#include "memory.h"
#include "pointer.h"

#include <cstdlib>

namespace Memory {

namespace Impl {

class BaseMapping {
public:
	virtual ~ BaseMapping() { }
};

size_t wholeBlocks(size_t size, size_t blocksize = Core::blocksize());

}

template<class Backend> class Mapping;

template<> class Mapping<Core> : public Impl::BaseMapping {
public:
	Mapping(size_t size, const char *name = "");
	virtual ~ Mapping();

	void resize(size_t size);
	template<class Target> operator Pointer<Core, Target>() {
		return Pointer<Core, Target>(static_cast<Target *>(data));
	}
	size_t size() { return length; }
private:
	Mapping(const Mapping &other);
	Mapping &operator=(const Mapping &other);

	void *data;
	size_t length;
	const char *name;
};

namespace Impl {

class BaseDiskMapping : public BaseMapping {
public:
	size_t size() { return length; }
protected:
	BaseDiskMapping(size_t size, const char *name, size_t blocksize);
	virtual ~ BaseDiskMapping();
	void resize(size_t size, size_t blocksize, size_t *usedPages, size_t usedPageCount);
	void *data, *origPtr;
	size_t length, origLength;
	const char *name;
	int fd;
};

}

template<class Cache> class Mapping<Disk<Cache> > : public Impl::BaseDiskMapping {
public:
	Mapping(size_t size, const char *name = "") :
		Impl::BaseDiskMapping(size, name, Cache::blocksize()),
		id(Cache::putMapping(data, length, fd))
	{ }
	virtual ~ Mapping() {
		Cache::removeMapping(data, length, fd, id);
	}
	void resize(size_t size) {
		size = Impl::wholeBlocks(size, Cache::blocksize());
		if (size == length) {
			return;
		}
		void *oldData(data);
		size_t oldSize(length);
		size_t buffer[Cache::maxPageCount()];
		size_t pageCount(Cache::listPages(id, buffer));
		Impl::BaseDiskMapping::resize(size, Cache::blocksize(), buffer, pageCount);
		id = Cache::migrateMapping(oldData, data, oldSize, size, fd, id);
	}
	template<class Target> operator Pointer<Disk<Cache>, Target>() {
		return Pointer<Disk<Cache>, Target>(static_cast<Target *>(data), id);
	}
private:
	typename Cache::Id id;
};

}

#endif
