#ifndef SORT_FUNNEL_H
#define SORT_FUNNEL_H

#include "common.h"
#include "qsort.h"
#include <memory/allocator.h>
#include <util/stack.h>
#include <thread/thread.h>

#include <cstdio>
#include <cmath>
#include <vector>

namespace Algorithm {
namespace Sort {

namespace Funnel {

struct Split {
	std::vector<size_t> mergeSizes;
	size_t count;
	Split(double count);
};

template<class Storage, class T, size_t lowSize, size_t heapSize> void funnelInternal(Memory::Pointer<Storage, T> buffer, Memory::Pointer<Storage, T> dest, Memory::Pointer<Storage, T> tmp, size_t size, bool splitThreads, std::vector<Split>::const_iterator begin, std::vector<Split>::const_iterator end);

template<class Storage, class T, size_t lowSize, size_t heapSize> struct Splitter {
	std::vector<Split>::const_iterator split, end;
	typename Memory::Allocator::Pool<Storage>::State restore;
	Memory::Pointer<Storage, T> input, output, tmp;
	Memory::Pointer<Storage, size_t> splitPoints;
	size_t size;
	void operator ()() {
		funnelInternal<Storage, T, lowSize, heapSize>(input, output, tmp, size, false, split, end);
	}
};

template<class Storage, class T> struct Merger;

template<class Storage, class T> struct Buffer {
	Buffer() :
		size(0),
		pos(0)
	{ }
	Memory::Pointer<Storage, T> buffer;
	Memory::Pointer<Storage, Merger<Storage, T> > provider;
	size_t size, pos;
};

template<class Storage, class T> struct Merger {
	Memory::Pointer<Storage, Buffer<Storage, T > > inputs;
	// One grain of the heap-like structure
	struct Grain {
		T item;
		bool active: 1;
		bool bottom: 1;
		bool inited: 1;
		// We will have only small heaps, 13bits will be enough (expected is like 512 inputs)
		size_t input: 13;
	};
	Memory::Pointer<Storage, Grain> heap;
	// This is not needed to be large, the mergers are rather small
	size_t inputCount: 13;
	size_t restartInput: 14;
	bool needRestart: 1;
};

template<class Storage, class T> struct Builder {
	Memory::Pointer<Storage, Buffer<Storage, T> > output, inputs;
	std::vector<size_t>::const_iterator splits;
	size_t inputCount;
};

template<class Merger> int fillHeapMerger(Merger &merger, size_t position) {
#ifdef DEBUG_FUNNEL
	ASSERT(position);
#endif
	while (position < 2 * merger->inputCount && merger->heap[position].active) {
		if (merger->heap[position].bottom) {
			size_t input = merger->heap[position].input;
			if (merger->inputs[input].pos == merger->inputs[input].size) {
				return position;
			} else {
				merger->heap[position].item = merger->inputs[input].buffer[merger->inputs[input].pos ++];
				return -1;
			}
		} else {
			size_t left(position * 2);
			size_t right(left + 1);
			bool lActive(left < 2 * merger->inputCount && merger->heap[left].active);
			bool rActive(right < 2 * merger->inputCount && merger->heap[right].active);
			size_t newPos;
			if (lActive && rActive) {
				newPos = merger->heap[left].item < merger->heap[right].item ? left : right;
			} else if (lActive) {
				newPos = left;
			} else if (rActive) {
				newPos = right;
			} else {
				merger->heap[position].active = false;
				return -1; // We didn't run of any input buffer now, we stopped before it
			}
			merger->heap[position].item = merger->heap[newPos].item;
			position = newPos;
		}
	}
	return -1;
}

template<class Storage, class T, size_t heapSize> void buildMerger(Memory::Pointer<Storage, Buffer<Storage, T> > output, Memory::Pointer<Storage, Buffer<Storage, T> > inputs, size_t inputCount, std::vector<size_t>::const_iterator splits, Memory::Allocator::Pool<Storage> &pool, Util::SharedStack<Storage, Builder<Storage, T> > stack) {
	using namespace Memory;

	Builder<Storage, T> task;
	task.output = output;
	task.inputs = inputs;
	task.splits = splits;
	task.inputCount = inputCount;
#ifdef DEBUG_FUNNEL
	fprintf(DEBUG_FUNNEL, "DEBUG: Building merger out of %zu inputs\n", inputCount);
#endif
	stack.push(task);

	while (!stack.empty()) {
		Builder<Storage, T> task(stack.pop());
		if (*task.splits <= 2 || task.inputCount <= 3) {
#ifdef DEBUG_FUNNEL
			fprintf(DEBUG_FUNNEL, "DEBUG: Too small task (split to: %zu, input count: %zu)\n", *task.splits, task.inputCount);
#endif
			Pointer<Storage, Merger<Storage, T> > merger(pool);
			merger->inputs = task.inputs;
			merger->inputCount = task.inputCount;
			task.output->provider = merger;
		} else if (task.inputCount <= heapSize / sizeof(T)) {// Build a heap
#ifdef DEBUG_FUNNEL
			fprintf(DEBUG_FUNNEL, "DEBUG: Heaping %zu inputs together\n", task.inputCount);
#endif
			Pointer<Storage, Merger<Storage, T> > merger(pool);
			merger->inputs = task.inputs;
			merger->inputCount = task.inputCount;
			merger->heap = Pointer<Storage, typename Merger<Storage, T>::Grain>(pool[task.inputCount * 2]);
			merger->needRestart = false;
			task.output->provider = merger;
			for (size_t i(0); i < task.inputCount; ++ i) {
				merger->heap[i + task.inputCount].active = true;
				merger->heap[i + task.inputCount].bottom = true;
				merger->heap[i + task.inputCount].inited = false;
				merger->heap[i + task.inputCount].input = i;
				merger->heap[i].active = true;
				merger->heap[i].bottom = false;
				merger->heap[i].inited = false;
			}
		} else {
			// Splits to given number of parts of approx. same size and build mergers on them
			Pointer<Storage, Buffer<Storage, T> > middleBuffers(pool[*task.splits]);
			// If we are of size k, then *splits is sqrt(k) and we need internal buffers of k^(3/2)
			size_t bufferSize(task.inputCount * *task.splits);
#ifdef DEBUG_FUNNEL
			fprintf(DEBUG_FUNNEL, "DEBUG: Split %zu inputs to %zu submergers, using buffer size %zu\n", task.inputCount, *task.splits, bufferSize);
#endif
			for (size_t i(0); i < *task.splits; ++ i) {
				size_t left((i * task.inputCount) / *task.splits);
				size_t right(((i + 1) * task.inputCount) / *task.splits);
				size_t count(right - left);
				// We create the buffer as empty, it must be filled in first
				middleBuffers[i].pos = middleBuffers[i].size = bufferSize;
				middleBuffers[i].buffer = Pointer<Storage, T>(pool[bufferSize]);
				Builder<Storage, T> subtask;
				subtask.output = middleBuffers + i;
				subtask.inputs = task.inputs + left;
				subtask.splits = task.splits + 1;
				subtask.inputCount = count;
				stack.push(subtask);
			}
			task.inputs = middleBuffers;
			task.inputCount = *task.splits;
			task.splits ++;
			stack.push(task);
		}
	}
}

template<class Storage, class T> struct Position {
	Memory::Pointer<Storage, Merger<Storage, T> > ret;
	Memory::Pointer<Storage, Buffer<Storage, T> > output;
	size_t minpos, pos;
};

template<class Storage, class T, class Stack> void prepareSubproblem(Stack &stack, Memory::Pointer<Storage, Merger<Storage, T> > &merger, Memory::Pointer<Storage, Buffer<Storage, T> > &output, size_t &minpos, size_t &pos, size_t restartPos = 0) {
	// Run subproblem
	// Store this problem
	Position<Storage, T> current;
	merger->needRestart = true;
	merger->restartInput = restartPos;
	current.ret = merger;
	current.output = output;
	current.minpos = minpos;
	current.pos = pos;
	stack.push(current);
	// Switch to the subproblem
	output = merger->inputs + pos;
	merger = output->provider;
	output->pos = 0;
	minpos = pos = 0;
}

template<class Storage, class T, size_t heapSize> void runMerger(Memory::Pointer<Storage, Buffer<Storage, T> > dest, Memory::Pointer<Storage, Merger<Storage, T> > merger, Util::SharedStack<Storage, Position<Storage, T> > mergeStack) {
	size_t pos(0), minpos(0);
	while (!mergeStack.empty() || dest->pos < dest->size) {
		SUBPROBLEM_RESTART:
		if (dest->pos == dest->size) {
			// We filled the buffer. Return to the parent problem.
			// We start the buffer filled buffer from the beginning
			dest->pos = 0;
			// Switch the problems
			Position<Storage, T> parent(mergeStack.pop());
			dest = parent.output;
			merger = parent.ret;
			minpos = parent.minpos;
			pos = parent.pos;
		}
		if (heapSize && merger->heap) {
			if (merger->needRestart) {
				size_t input(merger->heap[merger->restartInput].input);
				if (merger->inputs[input].pos == merger->inputs[input].size)
					merger->heap[merger->restartInput].active = false;
				else {
					merger->heap[merger->restartInput].item = merger->inputs[input].buffer[merger->inputs[input].pos ++];
					merger->heap[merger->restartInput].inited = true;
				}
				merger->needRestart = false;
			}
			if (!merger->heap[1].inited) {
				for (size_t i(merger->inputCount * 2 - 1); i; -- i) {
					if (merger->heap[i].inited)
						continue;
					int result(fillHeapMerger(merger, i));
					if (result != i)
						merger->heap[i].inited = true;
					if (result != -1) {
#ifdef DEBUG_FUNNEL
						ASSERT(merger->heap[result].bottom);
#endif
						size_t input(merger->heap[result].input);
						if (merger->inputs[input].provider) {
							prepareSubproblem(mergeStack, merger, dest, minpos, input, result);
							goto SUBPROBLEM_RESTART;
						} else {
							merger->heap[result].inited = true;
							merger->heap[result].active = false;
						}
					}
				}
			}
			if (!merger->heap[1].active) {
				// We run out of data here, remove us
				dest->size = dest->pos;
				dest->provider = Memory::Pointer<Storage, Merger<Storage, T> >();
				continue;
			}
			dest->buffer[dest->pos ++] = merger->heap[1].item;
			int result(fillHeapMerger(merger, 1));
			if (result != -1) {
#ifdef DEBUG_FUNNEL
				ASSERT(merger->heap[result].bottom);
#endif
				size_t input(merger->heap[result].input);
				if (merger->inputs[input].provider) {
					prepareSubproblem(mergeStack, merger, dest, minpos, input, result);
					goto SUBPROBLEM_RESTART;
				} else {
					merger->heap[result].active = false;
				}
			}
		} else {
			// Look for the smallest one to output
			while (pos < merger->inputCount) {
				if (merger->inputs[pos].pos == merger->inputs[pos].size) {
					// We run out of a buffer.
					if (merger->inputs[pos].provider.null()) {
						// This one can not be filled, drop it
						for (size_t i(pos + 1); i < merger->inputCount; ++ i)
							merger->inputs[i - 1] = merger->inputs[i];
						-- merger->inputCount;
						// Try next buffer.
						continue;
					} else {
						prepareSubproblem(mergeStack, merger, dest, minpos, pos);
						// Run it
						goto SUBPROBLEM_RESTART;
					}
				} else if (merger->inputs[pos].buffer[merger->inputs[pos].pos] < merger->inputs[minpos].buffer[merger->inputs[minpos].pos])
					minpos = pos;
				++ pos;
			}
			if (!merger->inputCount) {
				// Run out of input, make the output buffer the correct size and return
				dest->size = dest->pos;
				dest->provider = Memory::Pointer<Storage, Merger<Storage, T> >(); // We are of no more use
				continue; // It will just get back to original problem now
			} else {
				dest->buffer[dest->pos ++] = merger->inputs[minpos].buffer[merger->inputs[minpos].pos ++];
				minpos = pos = 0;
			}
		}
	}
}

void buildSplitHierarchy(std::vector<Split> &hierarchy, size_t size);

template<class Storage, class T, size_t lowSize, size_t heapSize> void funnelInternal(Memory::Pointer<Storage, T> buffer, Memory::Pointer<Storage, T> dest, Memory::Pointer<Storage, T> tmp, size_t size, bool splitThreads, std::vector<Split>::const_iterator begin, std::vector<Split>::const_iterator end) {
	using namespace Memory;
	using namespace Memory::Allocator;
	Memory::Allocator::Pool<Storage> pool;
	if (begin == end)
		splitThreads = false;
	Util::Stack<Storage, Splitter<Storage, T, lowSize, heapSize> > stack;
	Splitter<Storage, T, lowSize, heapSize> splitter;
	splitter.input = buffer;
	splitter.output = dest;
	splitter.size = size;
	splitter.split = begin;
	splitter.tmp = tmp;
	stack.push(splitter);
	while(!stack.empty()) {
		splitter = stack.top();
		// We need to run it recursively
		if (splitter.splitPoints.null()) {
			if (splitter.split == end || splitter.split->count <= 1) {
#ifdef DEBUG_FUNNEL
				fprintf(DEBUG_FUNNEL, "DEBUG: Not splitting out array of %zu elements\n", splitter.size);
#endif
				// It's too small for us to know how to split it, just sort it
				for (size_t i(0); i < splitter.size; ++ i)
					splitter.output[i] = splitter.input[i];
				// Who should care about the small case?
				Common::bubble(splitter.output, splitter.size);
				// And we're done with this one, so finish off the task
				stack.pop();
			} else {
				if (splitter.size <= lowSize / sizeof(T)) {
#ifdef DEBUG_FUNNEL
					fprintf(DEBUG_FUNNEL, "DEBUG: Calling qsort on array of %zu elements\n", splitter.size);
#endif
					// It's small, use qsort (we are in hybrid mode)

					// It should reuse our stack, not use a new one
					Util::SharedStack<Storage, QSort::Range> qsortStack(stack);
					// In case it's not the same buffer, copy it
					if (splitter.input != splitter.output)
						for (size_t i(0); i < splitter.size; ++ i)
							splitter.output[i] = splitter.input[i];
					QSort::qsort(splitter.output, splitter.size, 1, qsortStack);
					// We're done with this part, drop it
					stack.pop();
				} else {
					// Mark the pool before we use anything so we can restore it
					splitter.restore = typename Pool<Storage>::State(pool);
					size_t splitCount(splitter.split->count);
#ifdef DEBUG_FUNNEL
					fprintf(DEBUG_FUNNEL, "Debug: Splitting array of %zu elements to %zu subtasks\n", splitter.size, splitCount);
#endif
					// Define where we split
					Pointer<Storage, size_t> splitPoints(pool[splitCount + 1]);
					for (size_t i(0); i < splitCount; ++ i)
						splitPoints[i] = (i * splitter.size) / splitCount;
					splitPoints[splitCount] = splitter.size;
					splitter.splitPoints = splitPoints;
					// Update the version on stack
					stack.top() = splitter;
					// And create the subtasks
					for (size_t i(0); i < splitCount; ++ i) {
						Splitter<Storage, T, lowSize, heapSize> subTask;
						subTask.input = splitter.input + splitPoints[i];
						subTask.output = splitter.tmp + splitPoints[i];
						subTask.tmp = splitter.output + splitPoints[i];
						subTask.size = splitPoints[i + 1] - splitPoints[i];
						subTask.split = splitter.split + 1;
						if (splitThreads) {
							subTask.end = end;
							Thread::enqueue(subTask);
						} else
							stack.push(subTask);
					}
					if (splitThreads) {
						splitThreads = false;
						Thread::waitAll();
					}
				}
			}
		} else {
			// Good, we have inputs, merge them
			size_t splitCount(splitter.split->count);
			// Prepare the buffer wrappers
			Pointer<Storage, Buffer<Storage, T> > buffers(pool[splitCount]);
			for (size_t i(0); i < splitCount; ++ i) {
				buffers[i].buffer = splitter.tmp + splitter.splitPoints[i];
				buffers[i].size = splitter.splitPoints[i + 1] - splitter.splitPoints[i];
			}
			Pointer<Storage, Buffer<Storage, T> > outputBuff(pool);
			outputBuff->buffer = splitter.output;
			outputBuff->size = splitter.size;
			// Run it
			buildMerger<Storage, T, heapSize>(outputBuff, buffers, splitCount, splitter.split->mergeSizes.begin(), pool, Util::SharedStack<Storage, Builder<Storage, T> >(stack));
			runMerger<Storage, T, heapSize>(outputBuff, outputBuff->provider, Util::SharedStack<Storage, Position<Storage, T> >(stack));
			// Free some memory and finish off the task
			pool.restore(splitter.restore);
			stack.pop();
		}
	}
}

template<class Storage, class T, size_t lowSize, size_t heapSize> void funnel(Memory::Pointer<Storage, T> buffer, size_t size, bool splitThreads) {
	using namespace Memory;
	std::vector<Split> splits;
	buildSplitHierarchy(splits, size);
	Mapping<Storage> destMap(size * sizeof(T), "Top level dest buffer");
	Pointer<Storage, T> dest(destMap);
	Mapping<Storage> tmpMap(size * sizeof(T), "Temporary buffer");
	Pointer<Storage, T> tmp(tmpMap);
	funnelInternal<Storage, T, lowSize, heapSize>(buffer, dest, tmp, size, splitThreads, splits.begin(), splits.end());
	for (size_t i = 0; i < size; ++ i)
		buffer[i] = dest[i];
}

}

template<class Storage, class T> void funnel(Memory::Pointer<Storage, T> buffer, size_t size) {
	Funnel::funnel<Storage, T, 0, 0>(buffer, size, false);
}

template<class Storage, class T> void funnelThreaded(Memory::Pointer<Storage, T> buffer, size_t size) {
	Funnel::funnel<Storage, T, 0, 0>(buffer, size, true);
}

template<class Storage, class T> void funnelHybrid(Memory::Pointer<Storage, T> buffer, size_t size) {
	Funnel::funnel<Storage, T, 65536, 0>(buffer, size, false);
}

template<class Storage, class T> void funnelHybridThreaded(Memory::Pointer<Storage, T> buffer, size_t size) {
	Funnel::funnel<Storage, T, 65536, 0>(buffer, size, true);
}

#ifdef DISK_SIZES
// With the disk sizes, we use large items and can set this to something little bit more, as we are limited by disk, not memory
template<class Storage, class T> void funnelFullHybrid(Memory::Pointer<Storage, T> buffer, size_t size) {
	Funnel::funnel<Storage, T, 65536, 8192*16>(buffer, size, false);
}

template<class Storage, class T> void funnelFullHybridThreaded(Memory::Pointer<Storage, T> buffer, size_t size) {
	Funnel::funnel<Storage, T, 65536, 8192*16>(buffer, size, true);
}
#else
template<class Storage, class T> void funnelFullHybrid(Memory::Pointer<Storage, T> buffer, size_t size) {
	Funnel::funnel<Storage, T, 65536, 8192>(buffer, size, false);
}

template<class Storage, class T> void funnelFullHybridThreaded(Memory::Pointer<Storage, T> buffer, size_t size) {
	Funnel::funnel<Storage, T, 65536, 8192>(buffer, size, true);
}
#endif

}
}

#endif
