#ifndef SORT_MERGE_H
#define SORT_MERGE_H

#include <memory/pointer.h>
#include <memory/allocator.h>
#include <util/stack.h>
#include <thread/thread.h>

namespace Algorithm {
namespace Sort {

namespace Merge {

template<class Storage, class T> struct Level {
	Memory::Pointer<Storage, T> input;
	Memory::Pointer<Storage, T> output;
	Memory::Pointer<Storage, T> temporary;
	size_t sizes[2];
};

template<class Output, class Input> void merge(Output output, Input inputs, size_t sizes[2]) {
	while (sizes[0] && sizes[1]) {
		size_t input(*inputs[0] > *inputs[1]);
		*output = *inputs[input];
		++ output;
		++ inputs[input];
		-- sizes[input];
	}
	for (size_t i(0); i < 2; ++ i)
		while (sizes[i] --) {
			*output = *inputs[i];
			++ output;
			++ inputs[i];
		}
}

template<class Storage, class T> struct Merger {
	typedef Memory::Pointer<Storage, T> Ptr;
	typedef void (* SubMerger)(Ptr, size_t);
	Ptr input;
	SubMerger merger;
	size_t size;
	Merger(Ptr input_, size_t size_, SubMerger merger_) :
		input(input_),
		merger(merger_),
		size(size_)
	{ }
	void operator ()() { merger(input, size); }
};

template<class Storage, class T> struct MergeRun {
	Memory::Pointer<Storage, T> inputs[2];
	Memory::Pointer<Storage, T> output;
	size_t sizes[2];
	void operator ()() { merge(output, inputs, sizes); }
};

template<class Storage, class T> void mergeThreaded(void (* subMerger)(Memory::Pointer<Storage, T>, size_t), Memory::Pointer<Storage, T> input, size_t size) {
	// First, split it into parts and merge them separately, using threads
	MergeRun<Storage, T> runs[THREAD_COUNT];
	size_t runCount(0);
	for (size_t i(0); i < THREAD_COUNT; ++ i) {
		size_t start(i * size / THREAD_COUNT);
		size_t end((i + 1) * size / THREAD_COUNT);
		if (end != start) {
			Thread::enqueue(Merger<Storage, T>(input + start, end - start, subMerger));
			runs[runCount].inputs[0] = input + start;
			runs[runCount ++].sizes[0] = end - start;
		}
	}
	Thread::waitAll();
	// Now, start merging the runs together, each merge in separate thread
	Memory::Mapping<Storage> mapping(size * sizeof(T), "Post-split merge buffer");
	Memory::Pointer<Storage, T> tmp(mapping);
	size_t pos(0);
	for (size_t i(0); i < runCount; ++ i) {
		runs[i].output = tmp + pos;
		pos += runs[i].sizes[0];
	}
	bool needsSwap(false);
	while (runCount > 1) {
		for (size_t i(0); i < runCount - 1; i += 2) {
			runs[i / 2] = runs[i];
			runs[i / 2].inputs[1] = runs[i + 1].inputs[0];
			runs[i / 2].sizes[1] = runs[i + 1].sizes[0];
			Thread::enqueue(runs[i / 2]);
		}
		if (runCount % 2) {
			runs[runCount / 2] = runs[runCount - 1];
			for (size_t i(0); i < runs[runCount / 2].sizes[0]; ++ i)
				runs[runCount / 2].output[i] = runs[runCount / 2].inputs[0][i];
			runs[runCount / 2].sizes[1] = 0;
		}
		runCount = (runCount + 1) / 2;
		needsSwap = !needsSwap;
		// We can modify them now, the threads have their own copy
		for (size_t i(0); i < runCount; ++ i) {
			Memory::Pointer<Storage, T> tmp(runs[i].inputs[0]);
			runs[i].inputs[0] = runs[i].output;
			runs[i].output = tmp;
			runs[i].sizes[0] += runs[i].sizes[1];
		}
		// Wait for them to finish before using their outputs
		Thread::waitAll();
	}
	if (needsSwap)
		// TODO In threads too?
		for (size_t i(0); i < size; ++ i)
			runs[0].output[i] = runs[0].inputs[0][i];
}

}

template<class Storage, class T> void mergeStack(Memory::Pointer<Storage, T> input, size_t size) {
	using namespace Memory;
	using namespace Memory::Allocator;
	using namespace Util;
	using namespace Merge;
	Stack<Storage, Level<Storage, T> > stack;
	Mapping<Storage> temporaryMapping(size * sizeof(T), "Temporary buffer");
	Pointer<Storage, T> temporary(temporaryMapping);
	Level<Storage, T> current;
	current.output = input;
	current.input = input;
	current.sizes[0] = size;
	current.sizes[1] = 0;
	current.temporary = temporary;
	stack.push(current);
	while (!stack.empty()) {
		current = stack.pop();
		if (!current.sizes[1]) {
			if (current.sizes[0] <= 1) {
				if (current.sizes[0])
					*current.output = *current.input;
			} else {
				// Split
				current.sizes[1] = current.sizes[0] / 2;
				current.sizes[0] -= current.sizes[1];
				// Output buffers for subproblems
				Pointer<Storage, T> leftB(current.temporary);
				Pointer<Storage, T> rightB(current.temporary + current.sizes[0]);
				// Subproblems
				Level<Storage, T> left;
				left.sizes[0] = current.sizes[0];
				left.sizes[1] = 0;
				left.output = leftB;
				left.temporary = current.output;
				left.input = current.input;
				Level<Storage, T> right;
				right.sizes[0] = current.sizes[1];
				right.sizes[1] = 0;
				right.output = rightB;
				right.temporary = current.output + current.sizes[0];
				right.input = current.input + current.sizes[0];
				// Put it all into the stack
				stack.push(current);
				stack.push(right);
				stack.push(left);
			}
		} else {
			Pointer<Storage, T> inputs[2] = { current.temporary, current.temporary + current.sizes[0] };
			merge(current.output, inputs, current.sizes);
		}
	}
}

template<class Storage, class T> void mergeStackThreaded(Memory::Pointer<Storage, T> input, size_t size) {
	Merge::mergeThreaded(mergeStack, input, size);
}

template<class Storage, class T> void mergeFifo(Memory::Pointer<Storage, T> input, size_t size) {
	bool needCopy(false);
	using namespace Memory;
	using namespace Merge;
	Mapping<Storage> tempMap(size * sizeof(T), "temp merge buffer");
	Pointer<Storage, T> temp(tempMap);
	for (size_t len(1); len < size; len *= 2) {
		size_t pos(0);
		while (pos < size) {
			if (pos + len >= size) {
				for (size_t i(pos); i < size; ++ i)
					temp[i] = input[i];
				pos = size;
			} else {
				size_t sizes[2] = {len, len};
				if (pos + 2 * len >= size)
					sizes[1] = size - pos - len;
				Pointer<Storage, T> starts[2] = {input + pos, input + pos + len};
				merge(temp + pos, starts, sizes);
				pos += 2 * len;
			}
		}
		needCopy = !needCopy;
		Pointer<Storage, T> tmp(temp);
		temp = input;
		input = tmp;
	}
	if (needCopy)
		for (size_t i(0); i < size; ++ i)
			temp[i] = input[i];
}

template<class Storage, class T> void mergeFifoThreaded(Memory::Pointer<Storage, T> input, size_t size) {
	Merge::mergeThreaded(mergeFifo, input, size);
}

}
}

#endif
