#include "thread.h"
#include "lock.h"
#include <list>
#include <set>
#include <pthread.h>

// TODO: Some error checking

using std::list;
using std::set;

namespace Thread {

using namespace Impl;

namespace {

// Storage of the tasks
list<BaseTask *> tasks;
pthread_mutex_t taskMutex = PTHREAD_MUTEX_INITIALIZER;

void lock() { pthread_mutex_lock(&taskMutex); }
void unlock() { pthread_mutex_unlock(&taskMutex); }

pthread_cond_t taskCondition = PTHREAD_COND_INITIALIZER;
// Waiting for completition of tasks
pthread_cond_t waitCondition = PTHREAD_COND_INITIALIZER;
set<size_t> waitable, completed;

// Thread statistics
// Number of free threads to run now (one is the main thread running)
size_t slots(THREAD_COUNT - 1),
// Number of tasks running now
       tasksRunning(0),
       done(0);
size_t requests(0);
size_t threadsReady(0);

size_t query(size_t *variable) {
	lock();
	size_t result(*variable);
	unlock();
	return result;
}

void *runThread(void *voidCond) {
	pthread_cond_t *cond = (pthread_cond_t *) voidCond;
	lock();
	for (;;) {
		if (tasks.empty() || !slots) {
			// We're waiting for a task and a slot, so we are ready
			++ threadsReady;
			if (cond) {
#ifdef DEBUG_THREADS
				fprintf(DEBUG_THREADS, "DEBUG: Thread started and waiting\n");
#endif
				pthread_cond_signal(cond);
				cond = NULL;
			}
#ifdef DEBUG_THREADS
	fprintf(DEBUG_THREADS, "DEBUG: Thread waiting (size: %zd, slots: %zd, threads ready: %zd)\n", tasks.size(), slots, threadsReady);
#endif
			pthread_cond_wait(&taskCondition, &taskMutex);
#ifdef DEBUG_THREADS
	fprintf(DEBUG_THREADS, "DEBUG: Thread retry\n");
#endif
			-- threadsReady;
		} else {
			// Acquire the task
			-- slots;
			++ tasksRunning;
			BaseTask *task(tasks.front());
			tasks.pop_front();
			if (cond) {
#ifdef DEBUG_THREADS
				fprintf(DEBUG_THREADS, "DEBUG: Thread started and working\n");
#endif
				pthread_cond_signal(cond);
				cond = NULL;
			}
			// and run it, outside the mutex
			unlock();
			task->run();
			// The task is completed, so signal it
			lock();
			++ slots;
			-- tasksRunning;
			++ done;
			// Anybody waiting for this task should be woken up
			completed.insert(task->tid);
			delete task;
			pthread_cond_broadcast(&waitCondition);
		}
	}
}

void startThread() {
#ifdef DEBUG_THREADS
	fprintf(DEBUG_THREADS, "DEBUG: Starting thread (size: %zd, slots: %zd, threads ready: %zd)\n", tasks.size(), slots, threadsReady);
#endif
	pthread_t thread;
	pthread_cond_t startCond;
	pthread_cond_init(&startCond, NULL);
	pthread_create(&thread, NULL, runThread, &startCond);
	// Wait for the thread to start properly
#ifdef DEBUG_THREADS
	fprintf(DEBUG_THREADS, "DEBUG: Waiting for thread to start\n");
#endif
	pthread_cond_wait(&startCond, &taskMutex);
	pthread_cond_destroy(&startCond);
}

void waitInternal(size_t tid) {
	if (!threadsReady && !tasks.empty())
		startThread();
	// We don't ocupy the slot while waiting
	++ slots;
	pthread_cond_signal(&taskCondition);
	while (completed.find(tid) == completed.end())
		pthread_cond_wait(&waitCondition, &taskMutex);
	completed.erase(tid);
	waitable.erase(tid);
	// Acquire the slot back
	while (!slots) {
		pthread_cond_wait(&waitCondition, &taskMutex);
	}
	-- slots;
}

}

namespace Impl {

size_t enqueue(BaseTask *task) {
	lock();
	size_t id = task->tid = requests ++;
	waitable.insert(id);
	tasks.push_back(task);
	// If we don't have any work threads ready to run, but can run more
	if (!threadsReady) {
		startThread();
	}
	unlock();
	pthread_cond_signal(&taskCondition);
	return id;
}

}

size_t running() {
	return query(&tasksRunning);
}

size_t waiting() {
	lock();
	size_t result(tasks.size());
	unlock();
	return result;
}

size_t unused() {
	return query(&slots);
}

void wait(size_t tid) {
#ifdef DEBUG_THREADS
	fprintf(DEBUG_THREADS, "DEBUG: Waiting for completion of %zd\n", tid);
#endif
	lock();
	waitInternal(tid);
	unlock();
#ifdef DEBUG_THREADS
	fprintf(DEBUG_THREADS, "DEBUG: Task %zd completed\n", tid);
#endif
}

void waitAll() {
#ifdef DEBUG_THREADS
	fprintf(DEBUG_THREADS, "DEBUG: Waiting for completion of all tasks\n");
#endif
	lock();
	while (!waitable.empty() || tasksRunning || !tasks.empty()) {
		if (waitable.empty()) {
			pthread_cond_wait(&waitCondition, &taskMutex);
		} else {
			size_t tid(*waitable.begin());
			waitInternal(tid);
		}
	}
	unlock();
#ifdef DEBUG_THREADS
	fprintf(DEBUG_THREADS, "DEBUG: All tasks completed\n");
#endif
}

}
