#include <cstdio>
#include <cassert>
#include <algorithm>
#include <CL/cl.h>
#include <util/stopwatch.h>
#include <matrix/classic.h>
#include <matrix/divide.h>

using Util::Stopwatch;

namespace {

void die(const char *text) {
	fprintf(stderr, "%s\n", text);
	exit(1);
}

cl_context context;
cl_command_queue commandQueue;
cl_program program;
cl_kernel kern;
const char *progSrc(
"kernel void multiply(unsigned size, global const unsigned *left, global const unsigned *right, global unsigned *result, local unsigned *buffer) {\n"
	"unsigned row = get_global_id(0);\n"
	"unsigned column = get_global_id(1);\n"
	"event_t read;\n"
	"read = async_work_group_strided_copy(buffer, right + column, size, size, 0);\n"
	"wait_group_events(1, &read);\n"
	"unsigned smallerSize = size / 16;\n"
	"global uint16 *leftV = left + size * row;\n"
	"local uint16 *rightV = buffer;\n"
	"uint16 cell = (0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0);\n"
	"for (uint i = 0; i < smallerSize; ++ i)\n"
		"cell += leftV[i] * rightV[i];\n"
	"private uint *c = &cell;\n"
	"uint r = 0;\n"
	"for (uint i = 0; i < 16; ++ i)\n"
		"r += c[i];\n"
	"result[row * size + column] = r;\n"
"}\n"
);

#define CHECK assert(status == CL_SUCCESS)

void errorReport(const char *error, const void *, size_t, void *) {
	printf("Error: %s\n", error);
}

void initialize() {
	cl_int status = 0;
	cl_uint numPlatforms;
	status = clGetPlatformIDs(0, NULL, &numPlatforms);
	if (status != CL_SUCCESS)
		die("Didn't get platform count");

	cl_platform_id* platforms = new cl_platform_id[numPlatforms];
	status = clGetPlatformIDs(numPlatforms, platforms, NULL);
	if (status != CL_SUCCESS)
		die("Didn't get platforms");
	cl_context_properties cps[3] = { CL_CONTEXT_PLATFORM, (cl_context_properties) platforms[0], 0 };

	context = clCreateContextFromType(cps, CL_DEVICE_TYPE_GPU, errorReport, NULL, &status);
	if(status != CL_SUCCESS)
	{
		fprintf(stderr, "Try with CPU, no GPU found\n");
		context = clCreateContextFromType(cps, CL_DEVICE_TYPE_CPU, errorReport, NULL, &status);
		if(status != CL_SUCCESS)
			die("Couldn't create context");
	}
	delete platforms;
	size_t deviceListSize;
	status = clGetContextInfo(context, CL_CONTEXT_DEVICES, 0, NULL, &deviceListSize);
	if(status != CL_SUCCESS)
		die("Could not get device count");
	cl_device_id *devices = new cl_device_id[deviceListSize];
	status = clGetContextInfo(context, CL_CONTEXT_DEVICES, deviceListSize, devices, NULL);
	if (status != CL_SUCCESS)
		die("Could not get device list");
	commandQueue = clCreateCommandQueue(context, devices[0], CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE, &status);
	if (status != CL_SUCCESS)
		die("Could not create queue");
	program = clCreateProgramWithSource(context, 1, &progSrc, NULL, &status);
	if (status != CL_SUCCESS)
		die("Unable to load CL source code");
	status = clBuildProgram(program, 1, devices, NULL, NULL, NULL);
	if (status != CL_SUCCESS)
		die("Unable to build CL source code");
	clUnloadCompiler();
	kern = clCreateKernel(program, "multiply", &status);
	CHECK;
}

void runTest(size_t size) {
	Stopwatch watch("classic", "matrix", "gpu", "unsigned", size);
	cl_uint *left = (cl_uint *) malloc(size * size * sizeof *left);
	cl_uint *right = (cl_uint *) malloc(size * size * sizeof *right);
	cl_uint *result = (cl_uint *) malloc(size * size * sizeof *result);
	for (size_t i(0); i < size; ++ i)
		for (size_t j(0); j < size; ++ j) {
			left[j + size * i] = random();
			right[j + size * i] = random();
		}
	cl_int status;
	watch.start();
	cl_mem leftB = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, size * size * sizeof *left, left, &status);
	CHECK;
	cl_mem rightB = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, size * size * sizeof *right, right, &status);
	CHECK;
	cl_mem resultB = clCreateBuffer(context, CL_MEM_WRITE_ONLY, size * size * sizeof *right, NULL, &status);
	cl_uint sizeC(size);
	status = clSetKernelArg(kern, 0, sizeof sizeC, &sizeC);
	CHECK;
	status = clSetKernelArg(kern, 1, sizeof leftB, &leftB);
	CHECK;
	status = clSetKernelArg(kern, 2, sizeof rightB, &rightB);
	CHECK;
	status = clSetKernelArg(kern, 3, sizeof resultB, &resultB);
	CHECK;
	status = clSetKernelArg(kern, 4, size * sizeof(cl_uint), NULL);
	CHECK;
	size_t ws[2] = { size, size };
	// TODO: The 128 should be auto-detected
	size_t lws[2] = { std::min((size_t) 128, size), 1 };
	cl_event event;
	status = clEnqueueNDRangeKernel(commandQueue, kern, 2, NULL, ws, lws, 0, NULL, &event);
	CHECK;
	status = clEnqueueReadBuffer(commandQueue, resultB, true, 0, size * size * sizeof *result, result, 1, &event, NULL);
	CHECK;
	status = clReleaseMemObject(leftB);
	CHECK;
	status = clReleaseMemObject(rightB);
	CHECK;
	status = clReleaseMemObject(resultB);
	CHECK;
	watch.stop();
	Matrix::Classic<Memory::Core, unsigned> leftM(size, size);
	Matrix::Classic<Memory::Core, unsigned> rightM(size, size);
	Matrix::Classic<Memory::Core, unsigned> resultM(size, size);
	for (size_t i(0); i < size; ++ i)
		for (size_t j(0); j < size; ++ j) {
			leftM[i][j] = left[i * size + j];
			rightM[i][j] = right[i * size + j];
		}
	Matrix::divideSquareMultiplyHybrid(resultM, leftM, rightM, 0);
	for (size_t i(0); i < size; ++ i)
		for (size_t j(0); j < size; ++ j)
			assert(result[i * size + j] == resultM[i][j]);
}

}

int main() {
	initialize();
	for (size_t i(16); i <= 4096; i *= 2) {
		runTest(i);
	}
}
