#include <cstdio>
#include <unistd.h>
#include <pthread.h>
#include <cerrno>
#include <cstring>
#include <cstdlib>
#include <stdint.h>

namespace {

const size_t blockSize(4096*64);
const size_t blockCount(16);
const size_t hashSize = 10017011;

struct AccessRaw {
	unsigned char *pointer;
	size_t size;
	bool auxiliary;
};

struct DataBlock {
	size_t count;
	struct Access {
		size_t index;
		size_t size;
		bool auxiliary;
	} accesses[blockSize];
	bool last;
	size_t round, unprocessed;
	DataBlock() :
		round(0),
		unprocessed(0)
	{}
} dataBlocks[blockCount];

pthread_mutex_t blockLock = PTHREAD_MUTEX_INITIALIZER;
pthread_cond_t blockEmpty = PTHREAD_COND_INITIALIZER, blockReady = PTHREAD_COND_INITIALIZER;

struct Cache {
	size_t cacheline;
	size_t linecount;
	size_t emptyLineCount;
	size_t hitCount, missCount;
	size_t auxHits, auxMisses;
	pthread_t thread;
	unsigned char lineshift;
	void start() {
		pthread_create(&thread, NULL, &runThread, this);
	}
	void wait() {
		pthread_join(thread, NULL);
	}
	static void *runThread(void *cacheV) {
		Cache *cache = (Cache *) cacheV;
		cache->run();
	}
	virtual void run() = 0;
} **caches;

template<class IT> struct ICache : public Cache {
	struct CacheLine {
		size_t index;
		IT next, prev;
		IT hashNext;
	};
	CacheLine *lines;
	IT *hashTable;
	IT head, tail;
	void touch(size_t index, bool aux, size_t hashPos) {
		// Find the cache line if it is in cache
		IT found(hashTable[hashPos]);
		while (found && lines[found].index != index)
			found = lines[found].hashNext;
		if (found) {
			hitCount ++;
			if (aux)
				auxHits ++;
			lines[lines[found].prev].next = lines[found].next;
			lines[lines[found].next].prev = lines[found].prev;
		} else {
			// We didn't find it. So get a new cache line somewhere
			missCount ++;
			if (aux)
				auxMisses ++;
			if (emptyLineCount) { // We are not yet full
				found = emptyLineCount --;
			} else { // Evict the last one in cache
				// Rip it out of the LRU
				found = lines[tail].prev;
				lines[tail].prev = lines[found].prev;
				lines[lines[found].prev].next = tail;
				// Find it in the hash table and remove
				IT *position(&hashTable[lines[found].index % hashSize]);
				while (*position != found)
					position = &lines[*position].hashNext;
				*position = lines[found].hashNext;
			}
			// Put it into the hash table at the new place
			lines[found].index = index;
			lines[found].hashNext = hashTable[hashPos];
			hashTable[hashPos] = found;
		}
		// Now we have a free floating cache line, put it in the front of LRU
		lines[found].prev = head;
		lines[found].next = lines[head].next;
		lines[lines[head].next].prev = found;
		lines[head].next = found;
	}
	virtual void run() {
		emptyLineCount = linecount;
		lines = new CacheLine[linecount + 3];
		head = linecount + 2;
		tail = linecount + 1;
		// Get a nice prime number
		hashTable = new IT[hashSize];
		memset(hashTable, 0, hashSize * sizeof *hashTable);
		hitCount = missCount = auxMisses = auxHits = 0;
		lines[head].prev = lines[tail].next = 0;
		lines[head].next = tail;
		lines[tail].prev = head;
		size_t blockIndex(0);
		size_t lastRound(0);
		bool finished(false);
		pthread_mutex_lock(&blockLock);
		do {
			while (dataBlocks[blockIndex].round <= lastRound)
				pthread_cond_wait(&blockReady, &blockLock);
			pthread_mutex_unlock(&blockLock);
			lastRound = dataBlocks[blockIndex].round;
			if (dataBlocks[blockIndex].count) {
				for (size_t i(0); i < blockSize; ++ i) {
					size_t index(dataBlocks[blockIndex].accesses[i].index);
					size_t end(index + dataBlocks[blockIndex].accesses[i].size - 1);
					index >>= lineshift;
					end >>= lineshift;
					size_t hash = index % hashSize;
					for (size_t j(index); j <= end; ++ j) {
						touch(j, dataBlocks[blockIndex].accesses[i].auxiliary, hash);
						hash ++;
						if (hash == hashSize)
							hash = 0;
					}
				}
			}
			finished = dataBlocks[blockIndex].last;
			pthread_mutex_lock(&blockLock);
			dataBlocks[blockIndex].unprocessed --;
			pthread_cond_signal(&blockEmpty);
			blockIndex ++;
			blockIndex %= blockCount;
		} while (!finished);
		pthread_mutex_unlock(&blockLock);
		delete[] lines;
		delete[] hashTable;
	}
};

size_t cacheCount(0);

}

typedef Cache *PCache;

int main(int argc, char *argv[]) {
	caches = new PCache[argc];
	for (char **cacheDef(argv + 1); *cacheDef; ++ cacheDef) {
		long long int cacheline(strtoll(strtok(*cacheDef, ": "), NULL, 0));
		long long int linecount(strtoll(strtok(NULL, ": "), NULL, 0));
		// We have lines only of power of two sizes anyway, this should speed it up
		int shift(0);
		while ((1 << shift) < cacheline)
			shift ++;
		if (linecount <= (1L << 16) - 3)
			caches[cacheCount] = new ICache<uint16_t>;
		else
			caches[cacheCount] = new ICache<uint32_t>;
		caches[cacheCount]->cacheline = cacheline;
		caches[cacheCount]->lineshift = shift;
		caches[cacheCount ++]->linecount = linecount;
	}
	for (size_t i(0); i < cacheCount; ++ i)
		caches[i]->start();
	AccessRaw accesses[blockSize];
	bool eof(false);
	size_t round(1);
	size_t blockIndex(0);
	size_t total(0);
	while (!eof) {
		size_t got(0);
		do {
			if (eof) {
				fprintf(stderr, "Wrong file size\n");
				return 1;
			}
			ssize_t result = read(0, (unsigned char *) accesses + got, sizeof(accesses) - got);
			if (result < 0) {
				perror("Could not read input");
				return 1;
			}
			eof = !result;
			got += result;
		} while ((got < blockSize * sizeof(*accesses)) && !eof);
		got /= sizeof(*accesses);
		pthread_mutex_lock(&blockLock);
		while (dataBlocks[blockIndex].unprocessed)
			pthread_cond_wait(&blockEmpty, &blockLock);
		pthread_mutex_unlock(&blockLock);
		DataBlock *block(&dataBlocks[blockIndex]);
		block->count = got;
		block->last = eof;
		for (size_t j(0); j < got; ++ j) {
			block->accesses[j].index = (accesses[j].pointer - (unsigned char *) NULL);
			block->accesses[j].size = accesses[j].size;
			block->accesses[j].auxiliary = accesses[j].auxiliary;
		}
		for (size_t j(got); j < blockSize; ++ j) {
			block->accesses[j] = block->accesses[j - 1];
			block->accesses[j].auxiliary = false;
		}
		total += got;
		pthread_mutex_lock(&blockLock);
		block->round = round ++;
		block->unprocessed = cacheCount;
		pthread_mutex_unlock(&blockLock);
		pthread_cond_broadcast(&blockReady);
		blockIndex ++;
		blockIndex %= blockCount;
	}
	for (size_t i(0); i < cacheCount; ++ i) {
		caches[i]->wait();
		printf("Cache %zu*%zu: miss %zu, hit %zu, aux miss %zu, aux hit %zu, total %zu\n", caches[i]->linecount, caches[i]->cacheline, caches[i]->missCount, total - caches[i]->missCount, caches[i]->auxMisses, caches[i]->auxHits, total);
	}
	return 0;
}
