#!/usr/bin/perl
use common::sense;
use Getopt::Long qw(:config auto_abbrev permute);
use List::Util qw(&min &max);

my $src = 'caches';
my @limits;
my $multigrap;
my @axis;
my $output = 'graph.png';
my $log;
my $xlog;
my $open;
my $tex;
my $suffix;
my @axname;

my @names;

GetOptions(
	tex => \$tex,
	log => \$log,
	xlog => \$xlog,
	open => \$open,
	'axis=s' => \@axis,
	'axname=s' => \@axname,
	'multigraph=s' => \$multigrap,
	'src=s' => \$src,
	'output=s' => \$output,
	'eq=s%' => sub {
		my ($option, $name, $value) = @_;
		push @limits, sub { $_[0]->{$name} eq $value; };
	},
	'le=s%' => sub {
		my ($option, $name, $value) = @_;
		push @limits, sub { $_[0]->{$name} <= $value; };
	},
	'ge=s%' => sub {
		my ($option, $name, $value) = @_;
		push @limits, sub { $_[0]->{$name} >= $value; };
	},
	'group=s' => sub {
		my ($option, $value) = @_;
		my $algos = {
			sort => [qw(mergeFifo mergeStack qsort heap funnel funnelHybrid funnelFullHybrid)],
			matrix => [qw(classic inverse divide divideHybrid square squareHybrid squareZ squareZHybrid strassen strassenHybrid strassenZ strassenZHybrid)],
			dfs => [qw(bfsSort dfsBRT bfsBRT dfsClassic bfsClassic)],
			componentsFull => [qw(componentsBfs componentsDfs componentsBfsBRT componentsDfsBRT componentsBfsSort componentsDC)],
			components => [qw(componentsBfsSortThreaded componentsDCThreaded componentsDfs componentsDfsBRT componentsBfsSort componentsDC)],
			maxmatch => [qw(maxMatchClassic maxMatchDC maxMatchBRT)],
		}->{$value} // die "Unknow algorithm group";
		@names = @$algos;
		$algos = {map { $_ => 1 } @$algos};
		push @limits, sub { $algos->{$_[0]->{algo}}; };
	},
	'suffix=s' => sub {
		my ($option, $value) = @_;
		push @limits, sub { $_[0]->{suffix} eq $value; }
	},
) or die "Could not get options\n";

my @input;
open my $input, '<', $src or die "Could not read $src ($!)\n";
my ($algo, $size, $suffix);
while (<$input>) {
	chomp;
	if (/^(.*)-tr-(\d+)(.*)$/) {
		$algo = $1;
		$size = $2;
		$suffix = $3;
	} elsif (/^Cache (\d+)\*(\d+): miss (\d+), hit (\d+), aux miss (\d+), aux hit (\d+), total (\d+)$/) {
		push @input, {
			algo => $algo,
			size => $size,
			suffix => $suffix,
			linecount => $1,
			linesize => $2,
			cachesize => $1 * $2,
			miss => $3,
			trans => $3 * $2,
			hit => $4,
			amiss => $5,
			ahit => $6,
			total => $7,
			atotal => $5 + $6,
			atrans => $5 * $2,
			dmiss => $3 - $5,
			dhit => $4 - $6,
			dtotal => $7 - $5 - $6,
			dtrans => ($3 - $5) * $2,
		};
	} else {
		warn "Bad line: $_";
	}
}
close $input;

sub check($) {
	my ($result) = @_;
	return min (1, map $_->($result), @limits);
}

my %outputs;
my @xr;

for my $result (grep check $_, @input) {
	my $oname = $multigrap ? $result->{$multigrap} : 'single';
	$outputs{$oname}->{"@{$result}{@axis}"} = 1;
	push @xr, $result->{$axis[0]};
}

my $min = min @xr;
my $max = max @xr;
my $diff = $max / $min;
$diff **= 0.0625;
my $range = "[".max($min / $diff).":".($max * $diff)."]";

for my $oname (keys %outputs) {
	open my $data, '>', $oname.".data.$$" or die "Could not output data $oname.data.$$ ($!)\n";
	print $data $_, "\n" for sort { $a <=> $b } keys %{$outputs{$oname}};
	close $data;
}

open my $plot, '|gnuplot' or die "Could not run gnuplot ($!)\n";
select $plot;
$\ = ";\n";
if ($tex) {
	print "set terminal pslatex color 10 size 16cm,9cm";
} else {
	print "set terminal png size 1024, 768";
	print "set output '$output'";
}
print "set key left top";
print "set log xyz" if $log;
print 'set format xy "$10^{%L}$"' if $log;
#print 'set format xy "\\\\gformat{%L}{%f}"' if $log; -- This, unfortunately, does not work, as it computes the length of the string and makes space for it
print "set log x" if $xlog;
print "set xlabel '$axname[0]'";
print "set ylabel '$axname[1]'";
print "set xrange $range";
my $s = 's'x(@axis - 2);
my @colors = qw(red blue black orchid green brown purple olivegreen orange #83ffd5 #007f00 #8a0000);
my $cnum;
print "${s}plot ".(join ', ', (map "'$_.data.$$' title '$_' with linespoints lt 1 lc rgb \"".$colors[$cnum ++ % scalar @colors]."\"", sort @names));
close $plot;
unlink map "$_.data.$$", keys %outputs;
exec ('geeqie', $output) if $open;
