#include <cstdio>
#include <unistd.h>
#include <pthread.h>
#include <cerrno>
#include <cstring>
#include <cstdlib>
#include <CL/cl.h>

namespace {

void die(const char *text) {
	fprintf(stderr, "%s\n", text);
	exit(1);
}

cl_context context;
cl_command_queue commandQueue;
cl_program program;
const char *progSrc =
"struct Line;\n"
"typedef global struct Line *PLine;\n"
"struct Line {\n"
"	PLine next, prev, hashNext;\n"
"	unsigned long index;\n"
"};\n"
"struct State {\n"
"	PLine head, tail;\n"
"	unsigned long hits, misses;\n"
"};\n"
"struct Hash {\n"
"	PLine head;\n"
"};\n"
"kernel void listify(global struct Line *lines) {\n"
"	unsigned id = get_global_id(0);\n"
"	lines[id].next = &lines[id + 1];\n"
"	lines[id + 1].hashNext = &lines[id];\n"
"	lines[id + 1].prev = &lines[id];\n"
"	lines[id].index = 0;\n"
"}\n"
"kernel void initialize(global struct Hash *hash, global struct Line *lines, global struct State *state, constant unsigned *definitions) {\n"
"	unsigned cacheId = get_global_id(0);\n"
"	unsigned lineStart = definitions[cacheId * 5 + 3];\n"
"	unsigned lineCount = definitions[cacheId * 5 + 1];\n"
"	state[cacheId].head = &lines[lineStart + lineCount + 1];\n"
"	state[cacheId].tail = &lines[lineStart + lineCount];\n"
"	state[cacheId].head->next = &lines[lineStart];\n"
"	state[cacheId].tail->prev = &lines[lineStart + lineCount - 1];\n"
"	lines[lineStart].prev = state[cacheId].head;\n"
"	lines[lineStart + lineCount - 1].next = state[cacheId].tail;\n"
"	state[cacheId].hits = state[cacheId].misses = 0;\n"
"	unsigned hashStart = definitions[cacheId * 5 + 4];\n"
"	unsigned hashLength = definitions[cacheId * 5 + 2];\n"
"	for (unsigned i = hashStart; i < hashStart + hashLength; ++ i)\n"
"		hash[i].head = 0;\n"
"	hash[hashStart].head = &lines[lineStart + lineCount - 1];\n"
"	lines[lineStart].hashNext = 0;\n"
"}\n"
"kernel void block(global struct Hash *hash, global struct Line *lines, global struct State *state, constant unsigned *definitions, global unsigned long *input) {\n"
	// Load definitions and state
"	unsigned cacheId = get_global_id(0);\n"
"	struct State lstate = state[cacheId];\n"
"	unsigned lineSize = definitions[cacheId * 5];\n"
"	unsigned hashLength = definitions[cacheId * 5 + 2];\n"
"	unsigned lineStart = definitions[cacheId * 5 + 3];\n"
"	unsigned hashStart = definitions[cacheId * 5 + 4];\n"
"	unsigned count = input[0];\n"
"	for (unsigned i = 1; i <= count; ++ i) {\n"
		// Load the pointer
"		unsigned long index = input[i] / lineSize;\n"
"		unsigned hashPos = index % hashLength;\n"
		// Try to find it in the hash table
"		PLine found = hash[hashStart + hashPos].head;\n"
"		while (found && found->index != index)\n"
"			found = found->hashNext;\n"
"		if (found) {\n"
"			lstate.hits ++;\n"
"			found->prev->next = found->next;\n"
"			found->next->prev = found->prev;\n"
"		} else {\n"
"			lstate.misses ++;\n"
"			found = lstate.tail->prev;\n"
"			lstate.tail->prev = found->prev;\n"
"			lstate.tail->prev->next = lstate.tail;\n"
"			global PLine *position = &hash[hashStart + (found->index % hashLength)].head;\n"
"			while (*position != found)\n"
"				position = &(*position)->hashNext;\n"
"			*position = found->hashNext;\n"
"			found->index = index;\n"
"			found->hashNext = hash[hashStart + hashPos].head;\n"
"			hash[hashStart + hashPos].head = found;\n"
"		}\n"
"		found->prev = lstate.head;\n"
"		found->next = lstate.head->next;\n"
"		lstate.head->next->prev = found;\n"
"		lstate.head->next = found;\n"
"	}\n"
"	state[cacheId] = lstate;\n"
"}\n"
"kernel void answer(global struct State *state, global unsigned long *output) {\n"
"	unsigned cacheId = get_global_id(0);\n"
"	output[cacheId * 2] = state[cacheId].misses;\n"
"	output[cacheId * 2 + 1] = state[cacheId].hits;\n"
"}\n"
"kernel void getSizes(global unsigned int *sizes) {\n"
"	sizes[0] = sizeof(struct Hash);\n"
"	sizes[1] = sizeof(struct Line);\n"
"	sizes[2] = sizeof(struct State);\n"
"}\n"
;

void initializeCL(void)
{
	cl_int status = 0;
	cl_uint numPlatforms;
	status = clGetPlatformIDs(0, NULL, &numPlatforms);
	if (status != CL_SUCCESS)
		die("Didn't get platform count");

	cl_platform_id* platforms = new cl_platform_id[numPlatforms];
	status = clGetPlatformIDs(numPlatforms, platforms, NULL);
	if (status != CL_SUCCESS)
		die("Didn't get platforms");
	cl_context_properties cps[3] = { CL_CONTEXT_PLATFORM, (cl_context_properties) platforms[0], 0 };

	context = clCreateContextFromType(cps, CL_DEVICE_TYPE_GPU, NULL, NULL, &status);
	if(status != CL_SUCCESS)
	{
		fprintf(stderr, "Try with CPU, no GPU found\n");
		context = clCreateContextFromType(cps, CL_DEVICE_TYPE_CPU, NULL, NULL, &status);
		if(status != CL_SUCCESS)
			die("Couldn't create context");
	}
	delete platforms;
	size_t deviceListSize;
	status = clGetContextInfo(context, CL_CONTEXT_DEVICES, 0, NULL, &deviceListSize);
	if(status != CL_SUCCESS)
		die("Could not get device count");
	cl_device_id *devices = new cl_device_id[deviceListSize];
	status = clGetContextInfo(context, CL_CONTEXT_DEVICES, deviceListSize, devices, NULL);
	if (status != CL_SUCCESS)
		die("Could not get device list");
	commandQueue = clCreateCommandQueue(context, devices[0], CL_QUEUE_OUT_OF_ORDER_EXEC_MODE_ENABLE, &status);
	if (status != CL_SUCCESS)
		die("Could not create queue");
	program = clCreateProgramWithSource(context, 1, &progSrc, NULL, &status);
	if (status != CL_SUCCESS)
		die("Unable to load CL source code");
	status = clBuildProgram(program, 1, devices, NULL, NULL, NULL);
	if (status != CL_SUCCESS)
		die("Unable to build CL source code");
	clUnloadCompiler();
}

const size_t blockSize(1024*8);

struct AccessRaw {
	unsigned char *pointer;
	size_t size;
	bool auxiliary;
};

cl_uint *cache_defs;

struct Cache {
	size_t cacheline;
	size_t linecount;
} *caches;

size_t cacheCount(0);

size_t lineByteSize, hashByteSize, stateByteSize;

void getCLSizes() {
	cl_int status;
	cl_mem buffer = clCreateBuffer(context, CL_MEM_WRITE_ONLY, 3 * sizeof(cl_uint), NULL, &status);
	if (status != CL_SUCCESS)
		die("Could not allocate size buffer");
	cl_kernel function = clCreateKernel(program, "getSizes", &status);
	if (status != CL_SUCCESS)
		die("Could not get getSizes kernel");
	status = clSetKernelArg(function, 0, sizeof buffer, &buffer);
	if (status != CL_SUCCESS)
		die("Could not set the size buffer");
	cl_event execEvent;
	status = clEnqueueTask(commandQueue, function, 0, NULL, &execEvent);
	if (status != CL_SUCCESS)
		die("Could not run the getSizes kernel");
	cl_uint results[3];
	status = clEnqueueReadBuffer(commandQueue, buffer, CL_TRUE, 0, 3 * sizeof(cl_uint), results, 1, &execEvent, NULL);
	if (status != CL_SUCCESS)
		die("Could not read sizes");
	hashByteSize = results[0];
	lineByteSize = results[1];
	stateByteSize = results[2];
	status = clReleaseMemObject(buffer);
	if (status != CL_SUCCESS)
		die("Could not release size buffer");
	status = clReleaseKernel(function);
	if (status != CL_SUCCESS)
		die("Could not release the getSizes kernel");
#ifdef TRACE
	fprintf(stderr, "sizes: %zu %zu %zu\n", lineByteSize, hashByteSize, stateByteSize);
#endif
}

}

int main(int argc, char *argv[]) {
	// Prepare everything
	initializeCL();
	getCLSizes();
	caches = new Cache[argc];
	cache_defs = new cl_uint[argc * 5];
	cl_uint offsetHashes(0), offsetLines(0);
	for (char **cacheDef(argv + 1); *cacheDef; ++ cacheDef) {
		long long int cacheline(strtoll(strtok(*cacheDef, ": "), NULL, 0));
		long long int linecount(strtoll(strtok(NULL, ": "), NULL, 0));
		cache_defs[cacheCount * 5] = cacheline;
		cache_defs[cacheCount * 5 + 1] = linecount;
		cl_uint hashSize = 2 * linecount - 1;
		bool found;
		do {
			hashSize += 2;
			found = true;
			for (size_t i(3); i * i <= hashSize; ++ i)
				if (!(hashSize % i)) {
					found = false;
					break;
				}
		} while (!found);
		cache_defs[cacheCount * 5 + 2] = hashSize;
		offsetLines = (cache_defs[cacheCount * 5 + 3] = offsetLines) + linecount + 2;
		offsetHashes = (cache_defs[cacheCount * 5 + 4] = offsetHashes) + hashSize;
		caches[cacheCount].cacheline = cacheline;
		caches[cacheCount ++].linecount = linecount;
	}
	// Initialize the computation buffers and run initialization inside the GPU
	cl_int status;
	cl_mem hashBuffer = clCreateBuffer(context, CL_MEM_READ_WRITE, offsetHashes * hashByteSize, NULL, &status);
	if (status != CL_SUCCESS)
		die("Could not create hash buffer");
	cl_mem lineBuffer = clCreateBuffer(context, CL_MEM_READ_WRITE, offsetLines * lineByteSize, NULL, &status);
	if (status != CL_SUCCESS)
		die("Could not create line buffer");
	cl_mem stateBuffer = clCreateBuffer(context, CL_MEM_READ_WRITE, cacheCount * stateByteSize, NULL, &status);
	if (status != CL_SUCCESS)
		die("Could not create status buffer");
	cl_mem defBuffer = clCreateBuffer(context, CL_MEM_READ_ONLY | CL_MEM_COPY_HOST_PTR, cacheCount * 5 * sizeof(cl_uint), cache_defs, &status);
	if (status != CL_SUCCESS)
		die("Could not create defs buffer");
	cl_kernel listify = clCreateKernel(program, "listify", &status);
	if (status != CL_SUCCESS)
		die("Could not create the listify kernel");
	clSetKernelArg(listify, 0, sizeof lineBuffer, &lineBuffer);
	if (status != CL_SUCCESS)
		die("Could not provide list to listify");
	cl_event listified;
	size_t lineCount = offsetLines;
	status = clEnqueueNDRangeKernel(commandQueue, listify, 1, NULL, &lineCount, NULL, 0, NULL, &listified);
	if (status != CL_SUCCESS)
		die("Could not listify");
	cl_kernel initFunc = clCreateKernel(program, "initialize", &status);
	if (status != CL_SUCCESS)
		die("Could not create the initialize kernel");
	status = clSetKernelArg(initFunc, 0, sizeof hashBuffer, &hashBuffer);
	if (status != CL_SUCCESS)
		die("Could not assign init parameter 0");
	status = clSetKernelArg(initFunc, 1, sizeof lineBuffer, &lineBuffer);
	if (status != CL_SUCCESS)
		die("Could not assign init parameter 1");
	status = clSetKernelArg(initFunc, 2, sizeof stateBuffer, &stateBuffer);
	if (status != CL_SUCCESS)
		die("Could not assign init parameter 2");
	status = clSetKernelArg(initFunc, 3, sizeof defBuffer, &defBuffer);
	if (status != CL_SUCCESS)
		die("Could not assign init parameter 3");
	cl_event lastWork;
	status = clEnqueueNDRangeKernel(commandQueue, initFunc, 1, NULL, &cacheCount, NULL, 1, &listified, &lastWork);
	if (status != CL_SUCCESS)
		die("Could not run initialization");
	// Prepare the computation kernel
	cl_kernel block = clCreateKernel(program, "block", &status);
	if (status != CL_SUCCESS)
		die("Could not load the block kernel");
	status = clSetKernelArg(block, 0, sizeof hashBuffer, &hashBuffer);
	if (status != CL_SUCCESS)
		die("Could not assign block parameter 0");
	status = clSetKernelArg(block, 1, sizeof lineBuffer, &lineBuffer);
	if (status != CL_SUCCESS)
		die("Could not assign block parameter 1");
	status = clSetKernelArg(block, 2, sizeof stateBuffer, &stateBuffer);
	if (status != CL_SUCCESS)
		die("Could not assign block parameter 2");
	status = clSetKernelArg(block, 3, sizeof defBuffer, &defBuffer);
	if (status != CL_SUCCESS)
		die("Could not assign block parameter 3");
	// A tripple buffer to transfer data
	struct Round {
		cl_mem buffer;
		cl_ulong data[blockSize + 1];
		cl_event transfered;
		cl_event executed;
		cl_event waits[2];
		bool active;
	};
	Round *rounds = new Round[3];
	for (size_t i(0); i < 3; ++ i) {
		rounds[i].buffer = clCreateBuffer(context, CL_MEM_READ_ONLY, (blockSize + 1) * sizeof(cl_ulong), NULL, &status);
		if (status != CL_SUCCESS)
			die("Could not create data buffer");
		rounds[i].active = false;
	}
	// Go on and process the data
	AccessRaw *accesses = new AccessRaw[blockSize];
	bool eof(false);
	size_t round(0);
#ifdef TRACE
	fprintf(stderr, "Initialized\n");
#endif
	size_t completed(0);
	while (!eof) {
		// Read it
		size_t got(0);
		do {
			if (eof) {
				fprintf(stderr, "Wrong file size\n");
				return 1;
			}
			ssize_t result = read(0, (unsigned char *) accesses + got, blockSize * sizeof(*accesses) - got);
			if (result < 0) {
				perror("Could not read input");
				return 1;
			}
			eof = !result;
			got += result;
		} while ((got < blockSize * sizeof(*accesses)) && !eof);
		if (got) {
			got /= sizeof(*accesses);
			// Make sure we can put data into the buffer, that they are transferred
			if (rounds[round].active) {
#ifdef TRACE
				fprintf(stderr, "Waited for %zu\n", round);
#endif
				status = clWaitForEvents(1, &rounds[round].transfered);
				if (status != CL_SUCCESS)
					die("Could not wait for buffer transfer");
			}
			// Prepare the data for transfer
			for (size_t j(0); j < got; ++ j)
				rounds[round].data[j + 1] = (accesses[j].pointer - (unsigned char *) NULL);
			rounds[round].data[0] = got;
			// Transfer the data and let them be executed
#ifdef TRACE
			fprintf(stderr, "Prepared\n");
#endif
			status = clEnqueueWriteBuffer(commandQueue, rounds[round].buffer, CL_FALSE, 0, (got + 1) * sizeof(cl_ulong), rounds[round].data, rounds[round].active ? 1 : 0, rounds[round].active ? &rounds[round].executed : NULL, &rounds[round].transfered);
			if (status != CL_SUCCESS)
				die("Could not transfer data");
			rounds[round].waits[0] = rounds[round].transfered;
			rounds[round].waits[1] = lastWork;
			status = clSetKernelArg(block, 4, sizeof rounds[round].buffer, &rounds[round].buffer);
			if (status != CL_SUCCESS)
				die("Could not set block kernel data param");
			status = clEnqueueNDRangeKernel(commandQueue, block, 1, NULL, &cacheCount, NULL, 2, rounds[round].waits, &rounds[round].executed);
			if (status != CL_SUCCESS)
				die("Could not execute data");
			lastWork = rounds[round].executed;
			completed += got;
#ifdef TRACE
			fprintf(stderr, "Enqueued %zu\n", completed);
#endif
			rounds[round ++].active = true;
			round %= 3;
		}
	}
#ifdef TRACE
	fprintf(stderr, "Done feeding\n");
#endif
	// Release temporary computation buffers and the block kernel
	/*
	status = clReleaseMemObject(hashBuffer);
	if (status != CL_SUCCESS)
		die("Could not release hash buffer");
	status = clReleaseMemObject(lineBuffer);
	if (status != CL_SUCCESS)
		die("Could not release line buffer");
	status = clReleaseMemObject(defBuffer);
	if (status != CL_SUCCESS)
		die("Could not release definition buffer");
	for (size_t i(0); i < 3; ++ i) {
		clReleaseMemObject(rounds[i].buffer);
		if (status != CL_SUCCESS)
			die("Could not release data buffer");
	}
	status = clReleaseKernel(block);
	if (status != CL_SUCCESS)
		die("Could not release the block kernel");
		*/
#ifdef TRACE
	fprintf(stderr, "Done releasing\n");
#endif
	// Get the answer
	cl_mem output = clCreateBuffer(context, CL_MEM_WRITE_ONLY, cacheCount * 2 * sizeof(cl_ulong), NULL, &status);
	if (status != CL_SUCCESS)
		die("Could not create output buffer");
	cl_kernel answer = clCreateKernel(program, "answer", &status);
	if (status != CL_SUCCESS)
		die("Could not load the answer kernel");
	status = clSetKernelArg(answer, 0, sizeof stateBuffer, &stateBuffer);
	if (status != CL_SUCCESS)
		die("Could not set state for the answer");
	status = clSetKernelArg(answer, 1, sizeof output, &output);
	if (status != CL_SUCCESS)
		die("Could not set the output buffer");
	cl_event result;
	status = clEnqueueNDRangeKernel(commandQueue, answer, 1, NULL, &cacheCount, NULL, 1, &lastWork, &result);
	if (status != CL_SUCCESS)
		die("Could not run answer kernel");
#ifdef TRACE
	fprintf(stderr, "Got answers\n");
#endif
	cl_ulong answerData[cacheCount * 2];
	status = clEnqueueReadBuffer(commandQueue, output, CL_TRUE, 0, cacheCount * 2 * sizeof(cl_ulong), answerData, 1, &result, NULL);
	if (status != CL_SUCCESS)
		die("Could not read answer");
	for (size_t i(0); i < cacheCount; ++ i) {
		size_t missCount = answerData[i * 2];
		size_t hitCount = answerData[i * 2 + 1];
		printf("Cache %zu*%zu: miss %zu, hit %zu, total %zu\n", caches[i].linecount, caches[i].cacheline, missCount, hitCount, missCount + hitCount);
	}
	return 0;
}
