#!/usr/bin/perl
use common::sense;
use AnyEvent;
use AnyEvent::Util;
use Getopt::Long qw(:config auto_abbrev permute);;
use List::Util qw(shuffle);
use Data::Dumper;

my $templateName;

my %vars;

sub generate($$$$) {
	my ($algorithm, $header, $size, $output) = @_;
	@vars{qw(algorithm size header)} = ($algorithm, $size, $header);
	open my $file, '>', "src/bench/$output.cpp" or die "Could not write output file $output ($!)\n";
	open my $template, '<', "templates/$templateName" or die "Could not read template $templateName ($!)\n";
	while (my $line = <$template>) {
		$line =~ s/\$$_/$vars{$_}/g for keys %vars;
		print $file $line;
	}
	close $file or die "Could not close $output ($!)\n";
	close $template;
}

sub compile($$$$) {
	my ($algorithm, $header, $size, $output) = @_;
	generate $algorithm, $header, $size, $output;
	print "Compiling $output\n";
	return run_cmd ['make', "bin/$output.opt"],
		close_all => 1;
}

my $slots = 1;
my $running;
#my $dir_avail = 1;
my $dir_avail;
my $doCache;

my @cacheSizes;
open CACHE_SIZES, 'cache-sizes' or die "No cache sizes\n";
@cacheSizes = grep $_, (<CACHE_SIZES>);
chomp @cacheSizes;
close CACHE_SIZES;

my $done = 0;

sub run($$$$) {
	my ($algorithm, $header, $size, $output) = @_;
	$running ++;
	$slots --;
	my $use_dir = $dir_avail;
	undef $dir_avail;
	my $cv = AnyEvent->condvar;
	compile($algorithm, $header, $size, $output)->cb(sub {
		die "Couldn't compile $output\n" if shift->recv;
		my ($read, $write) = portable_pipe;
		print "Starting $output\n";
		my $out;
		my $result;
		(run_cmd ['nice', '-n4', "./bin/$output.opt"],
			'3>' => $write,
			'>' => \$out,
			on_prepare => sub {
				$ENV{'MAP_DIR'} = '/mnt/ata/tmp/' if $use_dir;
			},
			close_all => 1)->cb(sub
		{
			$result = shift->recv;
			if (!$result) {
				open my $file, '>', "logs/$output.done";
				print $file $out;
				close $file;
			}
			$dir_avail = 1 if $use_dir;
			print "Terminated $output ($result)\n";
			print $out;
		});
		my $cmd = ['nice', '-n2', "xz", "--lzma2=preset=3,dict=48MiB", '--threads=4'];
		my $outputRedirect = "logs/$output.xz";
		my $outputData;
		if ($doCache) {
			$outputRedirect = \$outputData;
			$cmd = ['nice', '-n19', './tools/cache', @cacheSizes];
		}
		(run_cmd $cmd,
			'<' => $read,
			'>' => $outputRedirect,
			close_all => 1)->cb(sub {
				$done ++;
				if ($doCache) {
					open CACHE, '>>', 'caches' or die "Could not write output";
					print CACHE "$output\n";
					print CACHE $outputData;
					close CACHE;
				}
				$slots ++;
				$running --;
				$cv->send(shift->recv);
			});
		close $read;
		close $write;
		sleep 2; # Avoid collisions on opening mapping files
	});
	return $cv;
}

my $term = AnyEvent->condvar;
my @tasks;

sub start() {
	if (!@tasks && !$running) {
		$term->send;
		undef;
	} elsif (my ($task) = pop @tasks) {
		return undef unless $task;
		return &run(@$task);
	}
}

sub startTask() {
	my $cv = start;
	print "Done $done, ", scalar @tasks, " to go\n";
	return unless $cv;
	$cv->cb(sub {
		print "Compression error\n" if shift->recv;
		&startTask;
	});
}

my @sizes;
my %algos;
my $suffix;

GetOptions
	'size=i' => \@sizes,
	'jobs=i' => \$slots,
	'binary=i' => sub {
		@sizes = ();
		for (my $i = 2; $i <= @_[1]; $i *= 2) {
			push @sizes, $i;
		}
	},
	'direct-cache' => \$doCache,
	'template=s' => \$templateName,
	'var=s' => \%vars,
	'suffix=s' => \$suffix, # To modify the name
	'algorithm=s' => \%algos or die "Parsing failed\n";

foreach my $alname (keys %algos) {
	foreach my $size (@sizes) {
		next if -e "logs/$alname-tr-$size$suffix.done";
		push @tasks, [$alname, $algos{$alname}, $size, "$alname-tr-$size$suffix"];
	}
}
@tasks = shuffle @tasks;

$| = 1;

print "There are ", scalar @tasks, " tasks\n";

exit unless @tasks;

startTask while ($slots && @tasks);

$term->recv;
